﻿/* Pump.cs
 * 
 * Copyright (C) 2014 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

namespace Popolo.HVAC.Circuit
{
  /// <summary>遠心ポンプ</summary>
  [Serializable]
  public class CentrifugalPump : FluidMachinery, ImmutableCentrifugalPump
  {

    #region インスタンス変数・プロパティ

    /// <summary>バイパス量[m3/s]を取得する</summary>
    public double BypassFlowRate { get; private set; }

    /// <summary>制御方式を取得する</summary>
    public ControlMethod Control { get; private set; }

    /// <summary>圧力設定値[kPa]を設定・取得する</summary>
    public double PressureSetpoint { get; set; }

    #endregion

    #region 列挙型定義

    /// <summary>流量制御方式</summary>
    public enum ControlMethod
    {
      /// <summary>吐出圧一定制御（バイパス）</summary>
      ConstantPressureWithBypass,
      /// <summary>吐出圧一定制御（INV）</summary>
      ConstantPressureWithInverter,
      /// <summary>最小吐出圧制御</summary>
      MinimumPressure
    }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="nomPressure">定格揚程[kPa]</param>
    /// <param name="nomFlowRate">定格水量[m3/s]</param>
    /// <param name="designPressure">設計揚程[kPa]</param>
    /// <param name="designFlowRate">設計水量[m3/s]</param>
    /// <param name="control">制御方式</param>
    /// <param name="actualHead">実揚程[kPa]</param>
    public CentrifugalPump(double nomPressure, double nomFlowRate, double designPressure,
      double designFlowRate, ControlMethod control, double actualHead) :
       base(nomPressure, nomFlowRate, designPressure, designFlowRate, actualHead,
         control != ControlMethod.ConstantPressureWithBypass)
    {
      //制御方式を保存
      Control = control;

      //効率・圧力特性係数を計算
      efficiencyCoefficient = new double[3];
      pressureCoefficient = new double[3];
      getGeneralParameters(nomFlowRate, nomPressure, ref efficiencyCoefficient, ref pressureCoefficient);
      PressureSetpoint = designPressure;
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>状態を更新する</summary>
    /// <param name="flowRate">体積流量[m3/s]</param>
    public void UpdateState(double flowRate)
    {
      //停止判定
      VolumetricFlowRate = flowRate;
      if (VolumetricFlowRate <= 0)
      {
        ShutOff();
        return;
      }

      //バイパスによる吐出圧一定制御の場合：最大回転数・設計圧力で状態更新
      if (Control == ControlMethod.ConstantPressureWithBypass)
      {
        updateWithRotationRatioAndPressure(1.0, PressureSetpoint);
        BypassFlowRate = VolumetricFlowRate - flowRate;
      }
      //INVによる吐出圧一定制御の場合：必要流量・圧力を満たすように回転数を調整
      else if (Control == ControlMethod.ConstantPressureWithInverter)
      {
        updateWithFlowRateAndPressure(flowRate, PressureSetpoint);
        //最小回転数比[-]未満の場合は最小回転数にしてバイパス流量を計算
        if (RotationRatio < MinimumRotationRatio)
        {
          updateWithRotationRatioAndPressure(MinimumRotationRatio, PressureSetpoint);
          BypassFlowRate = VolumetricFlowRate - flowRate;
        }
      }
      //INVによる最小吐出圧制御の場合：抵抗曲線に合う圧力と流量で計算
      else
      {
        double ps = flowRate * flowRate * resistanceCoefficient + ActualHead;
        updateWithFlowRateAndPressure(flowRate, ps);
        //最小回転数比[-]未満の場合は最小回転数にしてバイパス流量を計算
        if (RotationRatio < MinimumRotationRatio)
        {
          updateWithResistanceAndRotationRatio(MinimumRotationRatio);
          BypassFlowRate = VolumetricFlowRate - flowRate;
        }
      }
    }

    /// <summary>回転数から流量を計算する</summary>
    /// <param name="rotationRatio">回転数比[-]</param>
    public void UpdateStateWithRotationRatio(double rotationRatio)
    {
      updateWithResistanceAndRotationRatio
        (Math.Min(1, Math.Max(MinimumRotationRatio, rotationRatio)));
    }

    /// <summary>停止させる</summary>
    public override void ShutOff()
    {
      base.ShutOff();
      BypassFlowRate = 0;
    }

    #endregion

    #region staticメソッド

    /// <summary>汎用ポンプの特性係数を計算する</summary>
    /// <param name="flowRate">設計流量[m3/s]</param>
    /// <param name="pressure">設計揚程[kPa]</param>
    /// <param name="efficiencyCoef">出力：効率特性係数</param>
    /// <param name="pressureCoef">出力：揚程特性係数</param>
    private static void getGeneralParameters(double flowRate, double pressure,
      ref double[] efficiencyCoef, ref double[] pressureCoef)
    {
      double vf2 = flowRate * flowRate;

      //効率特性
      double lnm = Math.Log(flowRate);
      double maxEff = -0.01618 * lnm * lnm - 0.05662 * lnm + 0.78179;
      efficiencyCoef[0] = -0.929 * maxEff / vf2;
      efficiencyCoef[1] = 1.795 * maxEff / flowRate;
      efficiencyCoef[2] = 0.109 * maxEff;

      //圧力特性
      pressureCoef[0] = -0.395 * pressure / vf2;
      pressureCoef[1] = 0.096 * pressure / flowRate;
      pressureCoef[2] = 1.294 * pressure;
    }

    #endregion

  }

  /// <summary>読み取り専用ポンプ</summary>
  public interface ImmutableCentrifugalPump : ImmutableFluidMachinery
  {
    /// <summary>バイパス量[m3/s]を取得する</summary>
    double BypassFlowRate { get; }

    /// <summary>制御方式を取得する</summary>
    CentrifugalPump.ControlMethod Control { get; }
  }

}
