﻿/* Conduit.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.Numerics;

namespace Popolo.HVAC.Circuit
{
  /// <summary>管内流れに関する静的メソッドを提供する</summary>
  /// <remarks>配管・ダクトに共通する処理</remarks>
  public static class Conduit
  {

    #region 列挙型定義

    /// <summary>管材</summary>
    public enum Material
    {
      /// <summary>炭素鋼</summary>
      UncoatedCarbonSteel_Clean,
      /// <summary>ポリ塩化ビニル</summary>
      PVC_PlasticPipe,
      /// <summary>アルミニウム</summary>
      Aluminum,
      /// <summary>ガルバリウム</summary>
      GalvanizedSteel,
      /// <summary>ステンレス</summary>
      StainlessSteel,
      /// <summary>グラスウール吹付</summary>
      FibrousGlassSpray,
      /// <summary>フレキシブルダクト</summary>
      FlexibleDuct,
      /// <summary>グラスウールダクト</summary>
      FibrousGlassDuct,
    }

    #endregion

    #region 等価直径関連の処理

    /// <summary>等価直径[m]を計算する</summary>
    /// <param name="flowArea">流路面積[m2]</param>
    /// <param name="perimeterLength">周長[m]</param>
    /// <returns>等価直径[m]</returns>
    public static double GetEquivalentDiameter(double flowArea, double perimeterLength)
    { return 4 * flowArea / perimeterLength; }

    /// <summary>角ダクトの等価直径[m]を計算する</summary>
    /// <param name="side1Length">1側の長さ[m]</param>
    /// <param name="side2Length">2側の長さ[m]</param>
    /// <returns>角ダクトの等価直径[m]</returns>
    /// <remarks>Huebscher(1948)による</remarks>
    public static double GetEquivalentDiameterOfRectangularDuct(double side1Length, double side2Length)
    {
      double a = side1Length / 1000d;
      double b = side2Length / 1000d;
      return (1.30 * Math.Pow(a * b, 0.625) / Math.Pow(a + b, 0.250)) / 1000d;
    }

    /// <summary>オーバルダクトの等価直径[m]を計算する</summary>
    /// <param name="majorLength">長径[m]</param>
    /// <param name="minorLength">短径[m]</param>
    /// <returns>オーバルダクトの等価直径[m]</returns>
    /// <remarks>Heyt and Diaz(1975)による</remarks>
    public static double GetEquivalentDiameterOfOvalDuct(double majorLength, double minorLength)
    {
      double a = majorLength / 1000d;
      double b = minorLength / 1000d;
      double aa = (Math.PI * b * b / 4d) + b * (a - b);
      double p = Math.PI * b + 2d * (a - b);
      return 1.55 * Math.Pow(aa, 0.625) / Math.Pow(p, 0.250);
    }

    #endregion

    #region ダルシーワイスバッハ式関連の処理

    /// <summary>ダルシーワイスバッハの式で圧力損失[Pa]を計算する</summary>
    /// <param name="frictionFactor">管摩擦係数[-]</param>
    /// <param name="density">密度[kg/m3]</param>
    /// <param name="length">管長[m]</param>
    /// <param name="diameter">管内径[m]</param>
    /// <param name="velocity">管内平均流速[m/s]</param>
    /// <returns>圧力損失[Pa]</returns>
    public static double GetPressureDrop
      (double frictionFactor, double density, double length, double diameter, double velocity)
    { return frictionFactor * density * length * velocity * velocity / (2 * diameter); }

    /// <summary>ダルシーワイスバッハの式で管内平均流速[m/s]を計算する</summary>
    /// <param name="frictionFactor">管摩擦係数[-]</param>
    /// <param name="density">密度[kg/m3]</param>
    /// <param name="length">管長[m]</param>
    /// <param name="diameter">管内径[m]</param>
    /// <param name="pressureDrop">圧力損失[Pa]</param>
    /// <returns>管内平均流速[m/s]</returns>
    public static double GetVelocity
      (double frictionFactor, double density, double length, double diameter, double pressureDrop)
    {
      double pda = Math.Abs(pressureDrop);
      double vel = (2 * pda * diameter) / (frictionFactor * density * length);
      return Math.Sign(pressureDrop) * Math.Sqrt(vel);
    }

    #endregion

    #region 管摩擦係数関連の処理

    /// <summary>素材の粗度[m]を取得する</summary>
    /// <param name="mat">素材</param>
    /// <returns>粗度[m]</returns>
    /// <remarks>管摩擦係数の計算では内径[m]で除した相対粗度[-]を用いる点に注意</remarks>
    public static double GetRoughness(Material mat)
    {
      switch (mat)
      {
        case Material.Aluminum:
          return 0.03e-3;
        case Material.FibrousGlassSpray:
          return 3.0e-3;
        case Material.FibrousGlassDuct:
          return 0.9e-3;
        case Material.FlexibleDuct:
          return 3.0e-3;
        case Material.GalvanizedSteel:
          return 0.09e-3;
        case Material.StainlessSteel:
          return 0;
        case Material.UncoatedCarbonSteel_Clean:
          return 0.03e-3;
        case Material.PVC_PlasticPipe:
          return 0.03e-3;
        default:
          return 0;
      }
    }

    /// <summary>管摩擦係数[-]を計算する</summary>
    /// <param name="reynoldsNumber">レイノルズ数[-]</param>
    /// <param name="relRoughness">相対粗度[-]（粗度[m]÷内径[m]）</param>
    /// <returns>管摩擦係数[-]</returns>
    public static double GetFrictionFactor(double reynoldsNumber, double relRoughness)
    {
      //層流の場合はレイノルズ数のみに依存
      if (reynoldsNumber < 4000) return 64d / reynoldsNumber;
      else
      {
        //遷移領域と仮定してCole brookの式を解く
        Roots.ErrorFunction eFnc = delegate (double fc)
        {
          double fcc = -2.0 * Math.Log10(relRoughness + 9.34 / (reynoldsNumber * Math.Sqrt(fc))) + 1.14;
          return fc - 1d / (fcc * fcc);
        };
        double fCoef = Roots.Newton(eFnc, 0.02, 0.0001, 0.000001, 0.0001, 10);
        double bnd = reynoldsNumber * Math.Sqrt(fCoef) * relRoughness;
        if (bnd <= 200) return fCoef;
        else
        {
          //完全に粗い場合（fully rough）にはNikuradseの式を解く
          fCoef = -2.0 * Math.Log10(relRoughness) + 1.14;
          return 1d / (fCoef * fCoef);
        }
      }
    }

    #endregion

  }
}
