﻿/* ControllableSingleFlow.cs
 * 
 * Copyright (C) 2018 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

namespace Popolo.HVAC.Circuit.ControllableFlowSolver
{
  /// <summary>抵抗調整によって流量制御可能な直列回路</summary>
  public class ControllableSeriesFlow : IFlowControllableBranch
  {

    #region インスタンス変数・プロパティ

    /// <summary>流量設定値[m3/s]を設定・取得する</summary>
    public double FlowRateSetPoint { get; set; }

    /// <summary>最小抵抗係数[kPa/(m3/s)^2]を取得する</summary>
    public double MinimumResistance { get; private set; }

    /// <summary>直列に並んだ固定抵抗係数[kPa/(m3/s)^2]を取得する</summary>
    public double FixedResistance { get; set; }

    /// <summary>抵抗係数[kPa/(m3/s)^2]を設定・取得する</summary>
    public double Resistance
    {
      get { return resist; }
      set
      {
        resist = Math.Max(MinimumResistance, value);
        HasTotalResistanceChanged = true;
      }
    }

    /// <summary>抵抗係数[kPa/(m3/s)^2]</summary>
    private double resist;

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="minimumResistance">最小抵抗[kPa/(m3/s)^2]</param>
    /// <param name="fixedResistance">直列に並んだ固定抵抗[kPa/(m3/s)^2]</param>
    public ControllableSeriesFlow(double minimumResistance, double fixedResistance)
    {
      MinimumResistance = Resistance = minimumResistance;
      FixedResistance = fixedResistance;
    }

    #endregion

    #region IFlowControllableBranch実装

    /// <summary>合成抵抗が変化したか否か</summary>
    public bool HasTotalResistanceChanged { get; private set; }

    /// <summary>回路の合計流量[m3/s]を取得する</summary>
    public double TotalFlowRate { get; private set; }

    /// <summary>回路の合計設定流量[m3/s]を取得する</summary>
    /// <returns>回路の合計設定流量[m3/s]</returns>
    public double GetTotalFlowSetPoint()
    { return FlowRateSetPoint; }

    /// <summary>必要最低差圧[kPa]を取得する</summary>
    /// <returns>必要最低差圧[kPa]</returns>
    public double GetMinimumPressure()
    { return FlowRateSetPoint * FlowRateSetPoint * (MinimumResistance + FixedResistance); }

    /// <summary>回路全体の合成抵抗[kPa/(m3/s)^2]を取得する</summary>
    /// <returns>回路全体の合成抵抗[kPa/(m3/s)^2]</returns>
    public double GetTotalResistance()
    {
      HasTotalResistanceChanged = false;
      return MinimumResistance + FixedResistance;
    }

    /// <summary>差圧にもとづいて流量を調整する</summary>
    /// <param name="pressure">差圧[kPa]</param>
    public void ControlFlowRate(double pressure)
    { TotalFlowRate = Math.Sqrt(pressure / GetTotalResistance()); }

    #endregion

  }
}
