﻿/* ParallelFluidMachineries.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;
using System.Collections.Generic;

using Popolo.Numerics;

namespace Popolo.HVAC.Circuit
{
  /// <summary>並列設置の流体機械群</summary>
  public class ParallelFluidMachineries: ICircuitBranch
  {

    #region インスタンス変数・プロパティ

    /// <summary>流体機械リスト</summary>
    private List<FluidMachinery> allFlds = new List<FluidMachinery>();

    /// <summary>各段数の流体機械リスト</summary>
    private List<FluidMachinery>[] flds;

    /// <summary>流体機械リストを取得する</summary>
    public ImmutableFluidMachinery[] FluidMachineries
    { get { return allFlds.ToArray(); } }
    
    /// <summary>回転数比[-]を設定・取得する</summary>
    public double RotationRatio { get; set; }

    /// <summary>最小回転数比[-]を取得する</summary>
    public double MinimumRotationRatio
    {
      get
      {
        double min = 0.001;
        foreach (FluidMachinery fm in flds[Stage - 1])
          if (min < fm.MinimumRotationRatio) min = fm.MinimumRotationRatio;
        return min;
      }
    }

    /// <summary>バイパス量[m3/s]を取得する</summary>
    public double BypassFlowRate { get; private set; }

    /// <summary>設計圧力[kPa]を取得する</summary>
    public double DesignPressure { get; private set; }

    /// <summary>最大段数を取得する</summary>
    public int MaximumStageNumber { get; private set; }

    /// <summary>最低差圧[kPa]を設定・取得する</summary>
    /// <remarks>全量バイパス時のバイパス回路差圧</remarks>
    public double MinimumPressure { get; set; }

    /// <summary>段数を設定・取得する</summary>
    public int Stage { get; set; }

    #endregion

    #region コンストラクタ

    /// <summary>コンストラクタ</summary>
    /// <param name="numberOfStages">段数制御数</param>
    /// <param name="designPressure">設計圧力[kPa]</param>
    public ParallelFluidMachineries(int numberOfStages, double designPressure)
    {
      if (numberOfStages <= 0) throw new Exception("StageNumberError");
      MaximumStageNumber = numberOfStages;
      RotationRatio = 1.0;
      DesignPressure = designPressure;
      MinimumPressure = 20;

      flds = new List<FluidMachinery>[numberOfStages];
      for (int i = 0; i < flds.Length; i++) flds[i] = new List<FluidMachinery>();
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>流体を流す</summary>
    /// <param name="flowRate">流量[m3/s]</param>
    /// <param name="pressure">差圧[kPa]</param>
    /// <returns>流すことができるか否か</returns>
    public bool TryToRunFluid(double flowRate, double pressure)
    {
      //一旦、全台停止
      foreach (FluidMachinery fm in allFlds) fm.ShutOff();
      pressure = Math.Max(MinimumPressure, pressure);

      for (Stage = 1; Stage <= MaximumStageNumber; Stage++)
      {
        VolumetricFlowRate = 0;
        foreach (FluidMachinery fm in flds[Stage - 1])
        {
          fm.updateWithRotationRatioAndPressure(1.0, pressure);
          VolumetricFlowRate += fm.VolumetricFlowRate;
        }
        if (flowRate < VolumetricFlowRate)
        {
          Roots.ErrorFunction eFnc = delegate (double rr)
          {
            VolumetricFlowRate = 0;
            foreach (FluidMachinery fm in flds[Stage - 1])
            {
              fm.updateWithRotationRatioAndPressure(rr, pressure);
              VolumetricFlowRate += fm.VolumetricFlowRate;
            }
            return VolumetricFlowRate - flowRate;
          };
          //収束計算//0.0001の加算は解がある側へシフトさせる保険
          //PQ特性切片付近の極微小流量で差圧不足となる場合があるため
          RotationRatio = Roots.NewtonBisection(eFnc, 1.0, 0.0001, 0.0001, 0.0001, 20);
          RotationRatio = Math.Max(0.0001 + RotationRatio, MinimumRotationRatio);
          BypassFlowRate = eFnc(RotationRatio);
          return true;
        }
      }
      Stage = MaximumStageNumber;
      return false;
    }

    /// <summary>流体機械を登録する</summary>
    /// <param name="stage">登録する段数(1～)</param>
    /// <param name="fm">登録する流体機械</param>
    public void AddFluidMachinery(int stage, FluidMachinery fm)
    {
      if (!allFlds.Contains(fm)) allFlds.Add(fm);
      if (!flds[stage - 1].Contains(fm)) flds[stage - 1].Add(fm);
    }

    /// <summary>流体機械を解除する</summary>
    /// <param name="stage">解除する段数(1～)</param>
    /// <param name="fm">解除する流体機械</param>
    public void RemoveFluidMachinery(int stage, FluidMachinery fm)
    {
      if (flds[stage - 1].Contains(fm)) flds[stage - 1].Remove(fm);
      for (int i = 0; i < flds.Length; i++)
        if (flds[i].Contains(fm)) return;
      allFlds.Remove(fm);
    }

    /// <summary>消費電力[kW]を取得する</summary>
    public double GetElectricConsumption()
    {
      double se = 0;
      foreach (FluidMachinery fm in allFlds) se += fm.GetElectricConsumption();
      return se;
    }

    #endregion

    #region ICircuitBranch実装

    /// <summary>流量[m3/s]を取得する</summary>
    public double VolumetricFlowRate { get; private set; }

    /// <summary>上流節点を設定・取得する</summary>
    public CircuitNode UpStreamNode { get; set; }

    /// <summary>下流節点を設定・取得する</summary>
    public CircuitNode DownStreamNode { get; set; }

    /// <summary>前後差圧から体積流量[m3/s]を計算する</summary>
    /// <returns>体積流量[m3/s]</returns>
    public void UpdateFlowRateFromNodePressureDifference()
    {
      VolumetricFlowRate = 0;
      if (Stage == 0) return;
      foreach (FluidMachinery fm in flds[Stage - 1])
      {
        fm.UpStreamNode = UpStreamNode;
        fm.DownStreamNode = DownStreamNode;
        fm.UpdateFlowRateFromNodePressureDifference();
        VolumetricFlowRate += fm.VolumetricFlowRate;
      }
    }

    #endregion

  }
}