﻿/* Regulator.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.Numerics;

namespace Popolo.HVAC.Circuit
{
  /// <summary>流量制御バルブ・ダンパ</summary>
  [Serializable]
  public class Regulator: ICircuitBranch
  {

    #region インスタンス変数・プロパティ

    /// <summary>全開時抵抗係数[kPa/(m3/s)^2]</summary>
    private double minResistance;

    /// <summary>開度[-]</summary>
    private double lift = 1;

    /// <summary>リニア特性重み係数[-]</summary>
    private double lWeight;

    /// <summary>レンジアビリティ[-]</summary>
    private double rangeAbility = 100;

    /// <summary>開度[-]を設定・取得する</summary>
    public double Lift
    {
      get { return lift; }
      set { lift = Math.Max(0, Math.Min(1, value)); }
    }

    /// <summary>リニア特性の重み係数[-]を設定・取得する</summary>
    public double LinearCharactaristicWeight
    {
      get { return lWeight; }
      set { lWeight = Math.Max(0, Math.Min(1, value)); }
    }

    /// <summary>レンジアビリティ[-]を設定・取得する</summary>
    public double RangeAbility
    {
      get { return rangeAbility; }
      set { if (0 < value) rangeAbility = value; }
    }
    
    /// <summary>流量設定値[m3/s]を設定・取得する</summary>
    public double VolumetricFlowRateSetPoint { get; set; }

    /// <summary>全閉可能か否かを設定・取得する</summary>
    public bool IsTotallyClosable { get; set; } = false;

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="flowRate">全開時流量[m3/s]</param>
    /// <param name="pressureDrop">全開時圧力損失[kPa]</param>
    /// <param name="rangeAbility">レンジアビリティ[-]</param>
    /// <param name="linearWeight">リニア特性重み係数[-]</param>
    public Regulator(double flowRate, double pressureDrop, double rangeAbility, double linearWeight)
    {
      DesignFlowRate = flowRate;
      VolumetricFlowRateSetPoint = flowRate;
      minResistance = pressureDrop / (flowRate * flowRate);
      RangeAbility = rangeAbility;
      LinearCharactaristicWeight = linearWeight;
    }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="cvValue">CV値[USgal/min]</param>
    /// <param name="rangeAbility">レンジアビリティ[-]</param>
    /// <param name="linearWeight">リニア特性重み係数[-]</param>
    public Regulator(double cvValue, double rangeAbility, double linearWeight)
    {
      DesignFlowRate = cvValue * 6.31e-5;
      VolumetricFlowRateSetPoint = DesignFlowRate;
      minResistance = 6.89 / (DesignFlowRate * DesignFlowRate);
      RangeAbility = rangeAbility;
      LinearCharactaristicWeight = linearWeight;
    }

    #endregion

    #region ICircuitBranch実装

    /// <summary>流量[m3/s]を設定・取得する</summary>
    public double VolumetricFlowRate { get; set; }

    /// <summary>設計流量[m3/s]を取得する</summary>
    public double DesignFlowRate { get; private set; }

    /// <summary>上流節点を設定・取得する</summary>
    public CircuitNode UpStreamNode { get; set; }

    /// <summary>下流節点を設定・取得する</summary>
    public CircuitNode DownStreamNode { get; set; }

    /// <summary>前後差圧から体積流量[m3/s]を計算する</summary>
    /// <returns>体積流量[m3/s]</returns>
    public void UpdateFlowRateFromNodePressureDifference()
    {
      if (IsTotallyClosable && Lift == 0) VolumetricFlowRate = 0;
      else
      {
        double dp = UpStreamNode.Pressure - DownStreamNode.Pressure;
        VolumetricFlowRate = Math.Sign(dp) * Math.Sqrt(Math.Abs(dp) / GetResistance());
      }
    }

    /// <summary>流動抵抗係数[kPa/(m3/s)^2]を取得する</summary>
    /// <returns>流動抵抗係数[kPa/(m3/s)^2]</returns>
    public double GetResistance()
    {
      if (IsTotallyClosable && Lift == 0) return double.PositiveInfinity;
      double wf = LinearCharactaristicWeight;
      double lam = 1d / RangeAbility;
      return wf* minResistance / Math.Pow((1 - lam) * Lift + lam, 2)
       + (1d - wf) * minResistance * Math.Pow(lam, 2 * Lift - 2);
    }

    #endregion

    #region 開度計算処理

    /// <summary>現在の前後差圧[kPa]を前提に開度を調整する</summary>
    public void UpdateLift()
    {
      if (VolumetricFlowRateSetPoint == 0) Lift = 0;
      double dp = Math.Abs(UpStreamNode.Pressure - DownStreamNode.Pressure);
      UpdateLift(dp);
    }

    /// <summary>前後差圧[kPa]にもとづいて開度を調整する</summary>
    /// <param name="pressure">前後差圧[kPa]</param>
    public void UpdateLift(double pressure)
    {
      double res = pressure / (VolumetricFlowRateSetPoint * VolumetricFlowRateSetPoint);
      if (res < minResistance) Lift = 1.0;
      else
      {
        double wf = LinearCharactaristicWeight;
        double lam = 1d / RangeAbility;
        Roots.ErrorFunction eFnc = delegate (double c)
        {
          Lift = c;
          return wf * minResistance / Math.Pow((1 - lam) * Lift + lam, 2)
          + (1d - wf) * minResistance * Math.Pow(lam, 2 * Lift - 2) - res;
        };
        Lift = Roots.Newton(eFnc, 0.5, 1e-4, 1e-4, 1e-4, 20);
      }
    }

    #endregion

  }
}
