﻿/* CoolingTower.cs
 * 
 * Copyright (C) 2014 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.ThermophysicalProperty;
using Popolo.Numerics;

namespace Popolo.HVAC.HeatExchanger
{
  /// <summary>冷却塔</summary>
  [Serializable]
  public class CoolingTower : ImmutableCoolingTower
  {

    #region 定数宣言

    /// <summary>大気圧[kPa]</summary>
    private const double ATMOSPHERIC_PRESSURE = 101.325;

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    /// <summary>特性係数</summary>
    private const double EXP_N = -0.6;

    #endregion

    #region 列挙型定義

    /// <summary>流れのタイプ</summary>
    public enum AirFlowDirection
    {
      /// <summary>向流型</summary>
      CounterFlow,
      /// <summary>直交流型</summary>
      CrossFlow
    }

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>特性係数c[-]</summary>
    private double coefC;

    /// <summary>飛散水量比[-]</summary>
    private double driftWaterRate = 0.002;

    /// <summary>濃縮倍率[-]</summary>
    private double concentrationRatio = 4;

    /// <summary>水量[kg/s]を設定・取得する</summary>
    public double WaterFlowRate { get; set; }

    /// <summary>上限水量[kg/s]を取得する</summary>
    public double MaxWaterFlowRate { get; private set; }

    /// <summary>風量[kg/s]を取得する</summary>
    public double AirFlowRate { get; private set; }

    /// <summary>最大風量[kg/s]を取得する</summary>
    public double MaxAirFlowRate { get; private set; }

    /// <summary>外気湿球温度[C]を取得する</summary>
    public double OutdoorWetbulbTemperature { get; private set; } = 27;

    /// <summary>外気絶対湿度[kg/kg]を取得する</summary>
    public double OutdoorHumidityRatio { get; private set; } = 0.0195;

    /// <summary>入口水温[C]を取得する</summary>
    public double InletWaterTemperature { get; private set; }

    /// <summary>出口水温[C]を取得する</summary>
    public double OutletWaterTemperature { get; private set; }

    /// <summary>出口水温設定値[C]を設定・取得する</summary>
    public double OutletWaterSetPointTemperature { get; set; }

    /// <summary>除去熱量[kW]を取得する</summary>
    public double HeatRejection { get; private set; }

    /// <summary>濃縮倍率[-]を設定・取得する</summary>
    public double ConcentrationRatio
    {
      get { return concentrationRatio; }
      set { concentrationRatio = Math.Max(0, value); }
    }

    /// <summary>飛散水量比[-]を設定・取得する</summary>
    public double DriftWaterRate
    {
      get { return driftWaterRate; }
      set { driftWaterRate = Math.Min(1, Math.Max(0, value)); }
    }

    /// <summary>インバータ制御か否か</summary>
    public bool HasInverter { get; private set; }

    /// <summary>空気の流れのタイプを取得する</summary>
    public AirFlowDirection AirFlowType { get; private set; }

    /// <summary>定格消費電力[kW]を取得する</summary>
    public double NominalPowerConsumption { get; private set; }

    /// <summary>消費電力[kW]を取得する</summary>
    public double ElectricConsumption { get; private set; }

    /// <summary>蒸発による水消費量[kg/s]を取得する</summary>
    public double EvaporationWater { get; private set; }

    /// <summary>飛散による水消費量[kg/s]を取得する</summary>
    public double DriftWater { get; private set; }

    /// <summary>ブローによる水消費量[kg/s]を取得する</summary>
    public double BlowDownWater { get; private set; }

    /// <summary>水消費量（蒸発+飛散+ブロー）[kg/s]を取得する</summary>
    public double WaterConsumption
    { get { return EvaporationWater + DriftWater + BlowDownWater; } }

    /// <summary>過負荷か否か</summary>
    public bool IsOverLoad { get; private set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="inletWaterTemperature">入口水温[C]</param>
    /// <param name="outletWaterTemperature">出口水温[C]</param>
    /// <param name="wetbulbTemperature">外気湿球温度[C]</param>
    /// <param name="waterFlowRate">冷水流量[kg/s]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    /// <param name="airFlowType">流れのタイプ</param>
    /// <param name="powerConsumption">消費電力[kW]</param>
    /// <param name="hasInverter">インバータ制御か否か</param>
    public CoolingTower
      (double inletWaterTemperature, double outletWaterTemperature, double wetbulbTemperature, double waterFlowRate, 
      double airFlowRate, AirFlowDirection airFlowType, double powerConsumption, bool hasInverter)
    {
      //定格条件を保存
      OutletWaterSetPointTemperature = outletWaterTemperature;
      MaxAirFlowRate = AirFlowRate = airFlowRate;
      MaxWaterFlowRate = WaterFlowRate = waterFlowRate;
      HasInverter = hasInverter;
      AirFlowType = airFlowType;
      NominalPowerConsumption = powerConsumption;

      //定格条件にもとづき特性係数c[-]を初期化
      coefC = GetCoolingTowerCoefficient
        (inletWaterTemperature, outletWaterTemperature, wetbulbTemperature, waterFlowRate, airFlowRate, airFlowType);

      //出口状態を初期化
      Update(inletWaterTemperature, false);
    }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="inletWaterTemperature">入口水温[C]</param>
    /// <param name="outletWaterTemperature">出口水温[C]</param>
    /// <param name="wetbulbTemperature">外気湿球温度[C]</param>
    /// <param name="waterFlowRate">冷水流量[kg/s]</param>
    /// <param name="airFlowType">流れのタイプ</param>
    /// <param name="hasInverter">インバータ制御か否か</param>
    public CoolingTower
      (double inletWaterTemperature, double outletWaterTemperature, double wetbulbTemperature,
      double waterFlowRate, AirFlowDirection airFlowType, bool hasInverter)
      : this(inletWaterTemperature, outletWaterTemperature, wetbulbTemperature,
          waterFlowRate, waterFlowRate * 0.8, airFlowType,
         getFunPower((inletWaterTemperature - outletWaterTemperature) * waterFlowRate * WATER_SPECIFIC_HEAT),
         hasInverter)
    { }

    /// <summary>定格能力[kW]から定格ファン動力[kW]を推定する</summary>
    /// <param name="load">定格能力[kW]</param>
    /// <returns>定格ファン動力[kW]</returns>
    private static double getFunPower(double load)
    {
      if (load < 30) return 0.2;
      else if (load < 35) return 0.4;
      else if (load < 100) return 0.75;
      else if (load < 210) return 1.5;
      else if (load < 350) return 2.2;
      else if (load < 680) return 3.7;
      else if (load < 1020) return 5.5;
      else if (load < 1130) return 7.5;
      else throw new Exception("Out of range");
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>外気条件を設定する</summary>
    /// <param name="wetbulbTemperature">湿球温度[C]</param>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    public void SetOutdoorAirState(double wetbulbTemperature, double humidityRatio)
    {
      OutdoorWetbulbTemperature = wetbulbTemperature;
      OutdoorHumidityRatio = humidityRatio;
    }

    /// <summary>熱交換を更新する</summary>
    /// <param name="inletWaterTemperature">入口水温[C]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    private void updateHeatExchange(double inletWaterTemperature, double airFlowRate)
    {
      //入力条件設定
      InletWaterTemperature = inletWaterTemperature;
      AirFlowRate = airFlowRate;

      //除去熱量から冷却水出口温度を計算//2017.03.01 DEBUG
      if (AirFlowRate <= 0 || WaterFlowRate <= 0)
      {
        HeatRejection = 0;
        OutletWaterTemperature = InletWaterTemperature;
      }
      else
      {
        HeatRejection = GetHeatRejection
          (InletWaterTemperature, OutdoorWetbulbTemperature, WaterFlowRate, AirFlowRate, coefC, AirFlowType);
        OutletWaterTemperature = InletWaterTemperature - HeatRejection / (WATER_SPECIFIC_HEAT * WaterFlowRate);
      }
    }

    /// <summary>消費電力と水消費を更新する</summary>
    private void updateElectricyAndWater()
    {
      ElectricConsumption = GetPowerConsumption(AirFlowRate, MaxAirFlowRate, NominalPowerConsumption, HasInverter);
      double ew, dw, bw;
      double hOA = MoistAir.GetEnthalpyFromHumidityRatioAndWetBulbTemperature
        (OutdoorHumidityRatio, OutdoorWetbulbTemperature, ATMOSPHERIC_PRESSURE);
      GetMakeupWater(HeatRejection, hOA, OutdoorHumidityRatio, WaterFlowRate,
        AirFlowRate, MaxAirFlowRate, DriftWaterRate, ConcentrationRatio, out ew, out dw, out bw);
      EvaporationWater = ew;
      DriftWater = dw;
      BlowDownWater = bw;
    }

    /// <summary>状態を更新する</summary>
    /// <param name="inletWaterTemperature">入口水温[C]</param>
    /// <param name="airFlowRate">風量[kg/s]</param>
    public void Update(double inletWaterTemperature, double airFlowRate)
    {
      IsOverLoad = false;
      updateHeatExchange(inletWaterTemperature, airFlowRate);
      updateElectricyAndWater();
    }

    /// <summary>状態を更新する</summary>
    /// <param name="inletWaterTemperature">入口水温[C]</param>
    /// <param name="controlOutletWaterTemperature">出口水温を制御するか</param>
    public void Update(double inletWaterTemperature, bool controlOutletWaterTemperature)
    {
      //入力条件設定
      InletWaterTemperature = inletWaterTemperature;

      //出口温度を制御する場合には風量を調整
      if (controlOutletWaterTemperature)
      {
        if (InletWaterTemperature <= OutletWaterSetPointTemperature) ShutOff();
        else
        {
          bool oload;
          double af = GetAirFlowRate
            (inletWaterTemperature, OutletWaterSetPointTemperature, OutdoorWetbulbTemperature,
            WaterFlowRate, MaxAirFlowRate, coefC, AirFlowType, out oload);
          IsOverLoad = oload;
          updateHeatExchange(inletWaterTemperature, af);
        }
      }
      else Update(inletWaterTemperature, MaxAirFlowRate);

      updateElectricyAndWater();
    }

    /// <summary>除去熱量[kW]を満たす冷却水出入口温度を計算する</summary>
    /// <param name="heatRejection">[kW]</param>
    public void UpdateFromHeatRejection(double heatRejection)
    {
      const double MIN_TEMP = 10;
      const double MAX_TEMP = 50;

      //最小水温で熱交換が過剰ならば終了
      updateHeatExchange(MIN_TEMP, AirFlowRate);
      if (heatRejection < HeatRejection) return;

      //最高水温で熱交換が不足ならば終了
      updateHeatExchange(MAX_TEMP, AirFlowRate);
      if (HeatRejection < heatRejection) return;

      Roots.ErrorFunction eFnc = delegate (double tIn) 
      {
        updateHeatExchange(tIn, AirFlowRate);
        return Math.Abs(heatRejection - HeatRejection);
      };
      Roots.NewtonBisection(eFnc, 37, 1e-4, 1e-3, 1e-3, 20);
    }

    /// <summary>停止させる</summary>
    public void ShutOff()
    {
      OutletWaterTemperature = InletWaterTemperature;
      HeatRejection = ElectricConsumption = 0;
      EvaporationWater = DriftWater = BlowDownWater = 0;
    }

    #endregion

    #region staticメソッド

    /// <summary>冷却塔の特性係数c[-]を求める</summary>
    /// <param name="inletWaterTemperature">冷却水入口温度[C]</param>
    /// <param name="outletWaterTemperature">冷却水出口温度[C]</param>
    /// <param name="wetbulbTemperature">空気入口湿球温度[C]</param>
    /// <param name="waterFlowRate">冷却水の質量流量[kg/s]</param>
    /// <param name="airFlowRate">空気の質量流量[kg/s]</param>
    /// <param name="airFlowType">流れのタイプ</param>
    /// <returns>冷却塔の特性係数c[-]</returns>
    public static double GetCoolingTowerCoefficient
      (double inletWaterTemperature, double outletWaterTemperature, double wetbulbTemperature,
      double waterFlowRate, double airFlowRate, AirFlowDirection airFlowType)
    {
      //熱交換量[kW]を計算
      double capacity = (inletWaterTemperature - outletWaterTemperature) * waterFlowRate * WATER_SPECIFIC_HEAT;

      //冷却水温度に相当する空気の飽和エンタルピーを計算
      double hswi = MoistAir.GetEnthalpyFromDryBulbTemperatureAndRelativeHumidity
        (inletWaterTemperature, 100, ATMOSPHERIC_PRESSURE);
      double hswo = MoistAir.GetEnthalpyFromDryBulbTemperatureAndRelativeHumidity
        (outletWaterTemperature, 100, ATMOSPHERIC_PRESSURE);
      //入口空気の飽和エンタルピーを計算
      double hai = MoistAir.GetEnthalpyFromWetBulbTemperatureAndRelativeHumidity
        (wetbulbTemperature, 100, ATMOSPHERIC_PRESSURE);

      //平均的な比熱で熱容量流量比を計算
      double cs = (hswi - hswo) / (inletWaterTemperature - outletWaterTemperature);
      double mwc = waterFlowRate * WATER_SPECIFIC_HEAT / cs;
      double mMin = Math.Min(mwc, airFlowRate);
      double mMax = Math.Max(mwc, airFlowRate);
      double rmc = mMin / mMax;

      //熱通過有効度とNTUの計算
      double epsilon = capacity / mMin / (hswi - hai);
      double ntuMin;
      HeatExchange.FlowType aft;
      if (airFlowType == AirFlowDirection.CrossFlow) aft = HeatExchange.FlowType.CrossFlow_BothFluidsUnmixed;
      else aft = HeatExchange.FlowType.CounterFlow;
      ntuMin = HeatExchange.GetNTU(epsilon, rmc, aft);

      //NTUwの計算
      double ntuW;
      if (mwc == mMin) ntuW = ntuMin;
      else ntuW = ntuMin * mwc / airFlowRate;

      //冷却塔の特性係数c[-]を出力
      return ntuW / Math.Pow(waterFlowRate / airFlowRate, EXP_N);
    }

    /// <summary>除去熱量[kW]を計算する</summary>
    /// <param name="inletWaterTemperature">冷却水入口温度[C]</param>
    /// <param name="wetbulbTemperature">空気入口湿球温度[C]</param>
    /// <param name="waterFlowRate">冷却水の質量流量[kg/s]</param>
    /// <param name="airFlowRate">空気の質量流量[kg/s]</param>
    /// <param name="coolingTowerCoefficient">冷却塔の特性係数[-]</param>
    /// <param name="airFlowType">流れのタイプ</param>
    /// <returns>除去熱量[kW]</returns>
    public static double GetHeatRejection
      (double inletWaterTemperature, double wetbulbTemperature, double waterFlowRate, double airFlowRate,
      double coolingTowerCoefficient, AirFlowDirection airFlowType)
    {
      //冷却水温度に相当する空気の飽和エンタルピーを計算
      double hswi = MoistAir.GetEnthalpyFromDryBulbTemperatureAndRelativeHumidity
        (inletWaterTemperature, 100, ATMOSPHERIC_PRESSURE);
      //入口空気の飽和エンタルピーを計算
      double hai = MoistAir.GetEnthalpyFromWetBulbTemperatureAndRelativeHumidity
        (wetbulbTemperature, 100, ATMOSPHERIC_PRESSURE);
      //NTUwを計算
      double ntuw = coolingTowerCoefficient * Math.Pow(waterFlowRate / airFlowRate, EXP_N);

      //誤差関数を定義
      Roots.ErrorFunction eFnc = delegate (double waterOutletTemperature)
      {
        //平均的な比熱を計算
        double hswo = MoistAir.GetEnthalpyFromDryBulbTemperatureAndRelativeHumidity
        (waterOutletTemperature, 100, ATMOSPHERIC_PRESSURE);
        double cs = (hswi - hswo) / (inletWaterTemperature - waterOutletTemperature);

        //冷却水の換算流量を計算
        double mwc = waterFlowRate * WATER_SPECIFIC_HEAT / cs;

        //熱通過有効度を計算
        double ntu;
        if (airFlowRate < mwc) ntu = ntuw * (mwc / airFlowRate);
        else ntu = ntuw;
        HeatExchange.FlowType ft;
        if (airFlowType == AirFlowDirection.CounterFlow) ft = HeatExchange.FlowType.CounterFlow;
        else ft = HeatExchange.FlowType.CrossFlow_BothFluidsUnmixed;
        double epsilon = HeatExchange.GetEffectiveness
        (ntu, Math.Min(airFlowRate, mwc) / Math.Max(airFlowRate, mwc), ft);

        //除去熱量を計算して誤差を評価
        double hReject = epsilon * (hswi - hai) * Math.Min(airFlowRate, mwc);
        return hReject - (inletWaterTemperature - waterOutletTemperature) * waterFlowRate * WATER_SPECIFIC_HEAT;
      };

      //冷却水出口温度を収束計算
      double wOutT = inletWaterTemperature - 1;
      wOutT = Roots.Newton(eFnc, wOutT, 0.0001, 0.01, 0.001, 10);

      return (inletWaterTemperature - wOutT) * waterFlowRate * WATER_SPECIFIC_HEAT;
    }

    /// <summary>必要風量[kg/s]を計算する</summary>
    /// <param name="inletWaterTemperature">冷却水入口温度[C]</param>
    /// <param name="outletWaterTemperature">冷却水出口温度[C]</param>
    /// <param name="wetbulbTemperature">空気入口湿球温度[C]</param>
    /// <param name="waterFlowRate">冷却水の質量流量[kg/s]</param>
    /// <param name="maxAirFlowRate">空気の最大の質量流量[kg/s]</param>
    /// <param name="coolingTowerCoefficient">冷却塔の特性係数c[-]</param>
    /// <param name="airFlow">流れのタイプ</param>
    /// <param name="isOverLoad">過負荷か否か</param>
    /// <returns>除去熱量[kW]</returns>
    public static double GetAirFlowRate
      (double inletWaterTemperature, double outletWaterTemperature, double wetbulbTemperature,
      double waterFlowRate, double maxAirFlowRate, double coolingTowerCoefficient,
      AirFlowDirection airFlow, out bool isOverLoad)
    {
      //冷却水温度に相当する空気の飽和エンタルピーを計算
      double hswi = MoistAir.GetEnthalpyFromDryBulbTemperatureAndRelativeHumidity
        (inletWaterTemperature, 100, ATMOSPHERIC_PRESSURE);
      double hswo = MoistAir.GetEnthalpyFromDryBulbTemperatureAndRelativeHumidity
        (outletWaterTemperature, 100, ATMOSPHERIC_PRESSURE);
      //入口空気の飽和エンタルピー
      double hai = MoistAir.GetEnthalpyFromWetBulbTemperatureAndRelativeHumidity
        (wetbulbTemperature, 100, ATMOSPHERIC_PRESSURE);

      //加熱してしまう条件の場合には風量0とする
      if (hswi < hai)
      {
        isOverLoad = true;
        return 0;
      }

      //熱通過有効度×質量流量：εmの計算
      double emmi = waterFlowRate * WATER_SPECIFIC_HEAT
        * (inletWaterTemperature - outletWaterTemperature) / (hswi - hai);

      //平均的な比熱で冷却水の換算流量を計算
      double cs = (hswi - hswo) / (inletWaterTemperature - outletWaterTemperature);
      double mwc = waterFlowRate * WATER_SPECIFIC_HEAT / cs;

      //誤差関数を定義
      Roots.ErrorFunction eFnc = delegate (double afRate)
      {
        //NTUを計算
        double ntuw = coolingTowerCoefficient * Math.Pow(waterFlowRate / afRate, EXP_N);
        double ntu;
        if (afRate < mwc) ntu = ntuw * (mwc / afRate);
        else ntu = ntuw;

        //熱通過有効度を計算
        double mmin = Math.Min(mwc, afRate);
        double mmax = Math.Max(mwc, afRate);
        HeatExchange.FlowType ft;
        if (airFlow == AirFlowDirection.CounterFlow) ft = HeatExchange.FlowType.CounterFlow;
        else ft = HeatExchange.FlowType.CrossFlow_BothFluidsUnmixed;
        double epsilon = HeatExchange.GetEffectiveness(ntu, mmin / mmax, ft);

        return emmi - epsilon * mmin;
      };

      //風量不足の場合には最大風量を出力する
      isOverLoad = 0 < eFnc(maxAirFlowRate);
      if (isOverLoad) return maxAirFlowRate;

      //風量比0.1%で過剰処理の場合には0.1%とする
      isOverLoad = 0 < eFnc(0.001 * maxAirFlowRate);
      if (!isOverLoad) return 0.001 * maxAirFlowRate;

      //二分法で収束計算//誤差率0.1%未満
      isOverLoad = false;
      return Roots.Bisection(eFnc, 0.001 * maxAirFlowRate, maxAirFlowRate, 0.001, 0.001 * maxAirFlowRate, 20);
    }

    /// <summary>水消費量[kg/s]を計算する</summary>
    /// <param name="heatRejection">除去熱量[kW]</param>
    /// <param name="airEnthalpy">入口空気のエンタルピー[kJ/kg]</param>
    /// <param name="airHumidityRatio">入口空気の絶対湿度[kg/kg]</param>
    /// <param name="waterFlowRate">冷却水の質量流量[kg/s]</param>
    /// <param name="airFlowRate">空気の質量流量[kg/s]</param>
    /// <param name="nominalAirFlowRate">定格点での空気の質量流量[kg/s]</param>
    /// <param name="driftWaterRate">飛散水量比[-]</param>
    /// <param name="concentrationRatio">濃縮倍率[-]</param>
    /// <param name="evaporationWater">出力：蒸発による水消費量[kg/s]</param>
    /// <param name="driftWater">出力：飛散による水消費量[kg/s]</param>
    /// <param name="blowDownWater">出力：ブローによる水消費量[kg/s]</param>
    public static void GetMakeupWater
      (double heatRejection, double airEnthalpy, double airHumidityRatio, double waterFlowRate,
      double airFlowRate, double nominalAirFlowRate, double driftWaterRate, double concentrationRatio,
      out double evaporationWater, out double driftWater, out double blowDownWater)
    {
      //処理熱量から蒸発水を計算
      double outletHRatio;
      if (heatRejection <= 0 || airFlowRate <= 0) outletHRatio = 0;
      else outletHRatio = MoistAir.GetHumidityRatioFromEnthalpyAndRelativeHumidity
        (airEnthalpy + heatRejection / airFlowRate, 100, ATMOSPHERIC_PRESSURE);
      evaporationWater = airFlowRate * (outletHRatio - airHumidityRatio);

      //循環量から飛散水を計算
      driftWater = waterFlowRate * driftWaterRate * (airFlowRate / nominalAirFlowRate);

      //0kg/s以上になるようにブロー水量を調整
      blowDownWater = Math.Max(0, evaporationWater / (concentrationRatio - 1) - driftWater);
    }

    /// <summary>ファン消費電力[kW]を計算する</summary>
    /// <param name="airFlowRate">空気の質量流量[kg/s]</param>
    /// <param name="nominalAirFlowRate">定格風量[kg/s]</param>
    /// <param name="nominalPowerConsumption">定格消費電力[kW]</param>
    /// <param name="hasInverter">インバータ制御か否か</param>
    /// <returns>ファン消費電力[kW]</returns>
    public static double GetPowerConsumption
      (double airFlowRate, double nominalAirFlowRate, double nominalPowerConsumption, bool hasInverter)
    {
      if (hasInverter) return nominalPowerConsumption * Math.Pow(airFlowRate / nominalAirFlowRate, 3);
      else return nominalPowerConsumption * airFlowRate / nominalAirFlowRate;
    }

    #endregion

  }

  #region 読み取り専用冷却塔インターフェース

  /// <summary>読み取り専用冷却塔</summary>
  public interface ImmutableCoolingTower
  {
    /// <summary>水量[kg/s]を取得する</summary>
    double WaterFlowRate { get; }

    /// <summary>上限水量[kg/s]を取得する</summary>
    double MaxWaterFlowRate { get; }

    /// <summary>風量[kg/s]を取得する</summary>
    double AirFlowRate { get; }

    /// <summary>定格風量[kg/s]を取得する</summary>
    double MaxAirFlowRate { get; }

    /// <summary>外気湿球温度[C]を取得する</summary>
    double OutdoorWetbulbTemperature { get; }

    /// <summary>外気絶対湿度[kg/kg]を取得する</summary>
    double OutdoorHumidityRatio { get; }

    /// <summary>入口水温[C]を取得する</summary>
    double InletWaterTemperature { get; }

    /// <summary>出口水温[C]を取得する</summary>
    double OutletWaterTemperature { get; }

    /// <summary>出口水温設定値[C]を取得する</summary>
    double OutletWaterSetPointTemperature { get; }

    /// <summary>除去熱量[kW]を取得する</summary>
    double HeatRejection { get; }

    /// <summary>濃縮倍率[-]を取得する</summary>
    double ConcentrationRatio { get; }

    /// <summary>飛散水量比[-]を取得する</summary>
    double DriftWaterRate { get; }

    /// <summary>インバータ制御か否か</summary>
    bool HasInverter { get; }

    /// <summary>空気の流れのタイプを取得する</summary>
    CoolingTower.AirFlowDirection AirFlowType { get; }

    /// <summary>定格消費電力[kW]を取得する</summary>
    double NominalPowerConsumption { get; }

    /// <summary>消費電力[kW]を取得する</summary>
    double ElectricConsumption { get; }

    /// <summary>蒸発による水消費量[kg/s]を取得する</summary>
    double EvaporationWater { get; }

    /// <summary>飛散による水消費量[kg/s]を取得する</summary>
    double DriftWater { get; }

    /// <summary>ブローによる水消費量[kg/s]を取得する</summary>
    double BlowDownWater { get; }

    /// <summary>水消費量（蒸発+飛散+ブロー）[kg/s]を取得する</summary>
    double WaterConsumption { get; }
  }

  #endregion

}
