﻿/* SimpleGroundHeatExchanger.cs
 * 
 * Copyright (C) 2018 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;

namespace Popolo.HVAC.HeatExchanger
{
  /// <summary>簡易地中熱交換器モデル(https://doi.org/10.3130/aije.83.491)</summary>
  [Serializable]
  public class SimpleGroundHeatExchanger : ImmutableSimpleGroundHeatExchanger
  {

    #region 列挙型

    /// <summary>種類</summary>
    public enum Type
    {
      /// <summary>垂直型</summary>
      Vertical,
      /// <summary>水平型</summary>
      Horizontal
    }

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>前回の近傍土壌温度[C]</summary>
    private double prevNearTemp;

    /// <summary>前回の遠方土壌温度[C]</summary>
    private double prevDistTemp;

    /// <summary>不易層温度[C]を設定・取得する</summary>
    public double ConstantGroundTemperature { get; set; } = 20;

    /// <summary>近傍の土壌温度[C]を取得する</summary>
    public double NearGroundTemperature { get; private set; } = 20;

    /// <summary>遠方の土壌温度[C]を取得する</summary>
    public double DistantGroundTemperature { get; private set; } = 20;

    /// <summary>地中熱交換器の熱通過有効度(0～1.0)を設定・取得する</summary>
    public double Effectiveness { get; set; } = 0.7;

    /// <summary>熱源水の比熱[kJ/(kgK)]を設定・取得する</summary>
    public double FluidSpecificHeat { get; set; } = 4.186;

    /// <summary>近傍～遠方の土壌の伝熱係数[kW/K]を設定・取得する</summary>
    public double NearGroundHeatConductance { get; set; }

    /// <summary>遠方～不易層の土壌の伝熱係数[kW/K]を設定・取得する</summary>
    public double DistantGroundHeatConductance { get; set; }

    /// <summary>近傍の土壌の熱容量[kJ/K]を設定・取得する</summary>
    public double NearGroundHeatCapacity { get; set; }

    /// <summary>遠方の土壌の熱容量[kJ/K]を設定・取得する</summary>
    public double DistantGroundHeatCapacity { get; set; }

    /// <summary>計算タイムステップ[sec]を設定・取得する</summary>
    public double TimeStep { get; set; } = 3600d;

    /// <summary>熱源水入口温度[C]を取得する</summary>
    public double FluidInletTemperature { get; private set; }

    /// <summary>熱源水出口温度[C]を取得する</summary>
    public double FluidOutletTemperature { get; private set; }

    /// <summary>熱源水流量[kg/s]を取得する</summary>
    public double FluidMassFlowRate { get; private set; }

    /// <summary>熱交換量[kW]を取得する（採熱が正、放熱が負）</summary>
    public double HeatExchange { get; private set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="nominalFlowRate">定格水量[kg/s]</param>
    /// <param name="fluidSpecificHeat">熱源水比熱[kJ/(kgK)]</param>
    /// <param name="effectiveness">地中熱交換器の熱通過有効度[-]</param>
    /// <param name="type">地中熱交換器の種類</param>
    public SimpleGroundHeatExchanger(double nominalFlowRate, double fluidSpecificHeat, double effectiveness, Type type)
    {
      InitTemperature(20);

      Effectiveness = effectiveness;
      FluidSpecificHeat = fluidSpecificHeat;
      double emc = effectiveness * nominalFlowRate * fluidSpecificHeat;

      //デフォルトパラメータ設定。詳細は文献参照
      if (type == Type.Horizontal)
      {
        NearGroundHeatConductance = 0.217 * emc;
        DistantGroundHeatConductance = 0.031 * emc;
        NearGroundHeatCapacity = 155.4e3 * emc;
        DistantGroundHeatCapacity = 840.4e3 * emc;
      }
      else
      {
        NearGroundHeatConductance = 0.373 * emc;
        DistantGroundHeatConductance = 0.153 * emc;
        NearGroundHeatCapacity = 5.8e3 * emc;
        DistantGroundHeatCapacity = 921.4e3 * emc;
      }
    }

    /// <summary>地中温度を初期化する</summary>
    /// <param name="groundTemperature">地中温度[C]</param>
    public void InitTemperature(double groundTemperature)
    {
      NearGroundTemperature = DistantGroundTemperature =
        prevNearTemp = prevDistTemp = groundTemperature;
    }

    #endregion

    #region 状態更新処理

    /// <summary>状態を予想する</summary>
    public void ForecastSate(double fluidInletTemperature, double fluidMassFlowRate)
    {
      DistantGroundTemperature = prevDistTemp;
      NearGroundTemperature = prevNearTemp;

      double emc = Effectiveness * fluidMassFlowRate * FluidSpecificHeat;
      double aCnt = (emc * fluidInletTemperature + NearGroundHeatConductance * DistantGroundTemperature) / (emc + NearGroundHeatConductance);
      double bCnt = Math.Exp(-(emc + NearGroundHeatConductance) / NearGroundHeatCapacity * TimeStep);
      double aFar = (NearGroundHeatConductance * NearGroundTemperature + DistantGroundHeatConductance * ConstantGroundTemperature) / (NearGroundHeatConductance + DistantGroundHeatConductance);
      double bFar = Math.Exp(-(NearGroundHeatConductance + DistantGroundHeatConductance) / DistantGroundHeatCapacity * TimeStep);

      NearGroundTemperature = aCnt + (NearGroundTemperature - aCnt) * bCnt;
      DistantGroundTemperature = aFar + (DistantGroundTemperature - aFar) * bFar;

      FluidInletTemperature = fluidInletTemperature;
      FluidMassFlowRate = fluidMassFlowRate;
      if (FluidMassFlowRate <= 0)
      {
        HeatExchange = 0;
        FluidOutletTemperature = FluidInletTemperature;
      }
      else
      {
        HeatExchange = emc * (NearGroundTemperature - FluidInletTemperature);
        FluidOutletTemperature = FluidInletTemperature + HeatExchange / (FluidMassFlowRate * FluidSpecificHeat);
      }
    }

    /// <summary>予想した状態を確定する</summary>
    public void FixState()
    {
      prevDistTemp = DistantGroundTemperature;
      prevNearTemp = NearGroundTemperature;
    }

    /// <summary>地中温度を更新する</summary>
    /// <param name="fluidInletTemperature"></param>
    /// <param name="fluidMassFlowRate"></param>
    public void Update(double fluidInletTemperature, double fluidMassFlowRate)
    {
      ForecastSate(fluidInletTemperature, fluidMassFlowRate);
      FixState();
    }

    #endregion

  }

  #region 読み取り専用の簡易地中熱交換器インターフェース

  /// <summary>読み取り専用の簡易地中熱交換器</summary>
  public interface ImmutableSimpleGroundHeatExchanger
  {
    /// <summary>不易層温度[C]を取得する</summary>
    double ConstantGroundTemperature { get; }

    /// <summary>近傍の土壌温度[C]を取得する</summary>
    double NearGroundTemperature { get; }

    /// <summary>遠方の土壌温度[C]を取得する</summary>
    double DistantGroundTemperature { get; }

    /// <summary>地中熱交換器の熱通過有効度(0～1.0)を取得する</summary>
    double Effectiveness { get; }

    /// <summary>熱源水の比熱[kJ/(kgK)]を取得する</summary>
    double FluidSpecificHeat { get; }

    /// <summary>近傍～遠方の土壌の伝熱係数[kW/K]を取得する</summary>
    double NearGroundHeatConductance { get; }

    /// <summary>遠方～不易層の土壌の伝熱係数[kW/K]を取得する</summary>
    double DistantGroundHeatConductance { get; }

    /// <summary>近傍の土壌の熱容量[kJ/K]を取得する</summary>
    double NearGroundHeatCapacity { get; }

    /// <summary>遠方の土壌の熱容量[kJ/K]を取得する</summary>
    double DistantGroundHeatCapacity { get; }

    /// <summary>計算タイムステップ[sec]を取得する</summary>
    double TimeStep { get; }

    /// <summary>熱源水入口温度[C]を取得する</summary>
    double FluidInletTemperature { get; }

    /// <summary>熱源水出口温度[C]を取得する</summary>
    double FluidOutletTemperature { get; }

    /// <summary>熱源水流量[kg/s]を取得する</summary>
    double FluidMassFlowRate { get; }

    /// <summary>熱交換量[kW]を取得する（採熱が正、放熱が負）</summary>
    double HeatExchange { get; }
  }

  #endregion

}
