﻿/* Boiler.cs
 * 
 * Copyright (C) 2014 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using Popolo.ThermophysicalProperty;

namespace Popolo.HVAC.HeatSource
{
  /// <summary>温水・蒸気ボイラに関する静的メソッドを提供する</summary>
  public static class Boiler
  {

    #region 定数宣言

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    /// <summary>空気の比熱[kJ/(kgK)]</summary>
    private const double AIR_SPECIFIC_HEAT = 1.006;

    /// <summary>空気の比重[kg/m3]</summary>
    private const double AIR_DENSITY = 1.2;

    /// <summary>排ガスの比熱[kJ/(kgK)]</summary>
    private const double SMOKE_SPECIFIC_HEAT = 1.38;

    /// <summary>排ガス温度近似係数1</summary>
    private const double A_ST = 1.17;

    /// <summary>排ガス温度近似係数2</summary>
    private const double B_ST = -0.17;

    #endregion

    #region 列挙型定義

    /// <summary>燃料種別</summary>
    public enum Fuel
    {
      /// <summary>都市ガス13A</summary>
      Gas13A,
      /// <summary>LNG</summary>
      LNG,
      /// <summary>LPG</summary>
      LPG,
      /// <summary>石炭</summary>
      Coal
    }

    #endregion

    #region 成り行き状態の計算

    /// <summary>出口状態を計算する</summary>
    /// <param name="inletWaterTemperature">入口水温[C]</param>
    /// <param name="waterFlowRate">水量[kg/s]</param>
    /// <param name="fuelConsumption">燃料消費量[Nm3/s, kg/s]</param>
    /// <param name="inletAirTemperature">入口空気温度[C]</param>
    /// <param name="fuel">燃料の種類</param>
    /// <param name="nominalSmokeTemperature">定格排ガス温度[C]</param>
    /// <param name="airRatio">空気比[-]</param>
    /// <param name="heatLossCoefficient">熱損失係数[kW/K]</param>
    /// <param name="ambientTemperature">周囲の温度[C]</param>
    /// <param name="nominalOutletWaterTemperature">定格出口水温[C]</param>
    /// <param name="outletWaterTemperature">出口水温[C]</param>
    /// <param name="heatLoad">加熱量[kW]</param>
    public static void GetOutletWaterTemperature
      (double inletWaterTemperature, double waterFlowRate, double fuelConsumption, double inletAirTemperature, Fuel fuel,
      double nominalSmokeTemperature, double airRatio, double heatLossCoefficient, double ambientTemperature, 
      double nominalOutletWaterTemperature, out double outletWaterTemperature, out double heatLoad)
    {
      double rf = getSmokeCoefficient(airRatio, fuel);
      double bf1 = AIR_DENSITY * GetTheoreticalAir(fuel) * airRatio * AIR_SPECIFIC_HEAT * inletAirTemperature;
      bf1 += 1000d * GetHeatingValue(fuel, true);
      bf1 -= SMOKE_SPECIFIC_HEAT * nominalSmokeTemperature * rf * B_ST;
      bf1 *= fuelConsumption;
      bf1 += heatLossCoefficient * ambientTemperature + waterFlowRate * WATER_SPECIFIC_HEAT * inletWaterTemperature;
      double bf2 = (waterFlowRate * WATER_SPECIFIC_HEAT + heatLossCoefficient);
      bf2 += A_ST * SMOKE_SPECIFIC_HEAT * rf * fuelConsumption * nominalSmokeTemperature / nominalOutletWaterTemperature;
      outletWaterTemperature = bf1 / bf2;
      heatLoad = WATER_SPECIFIC_HEAT * waterFlowRate * (outletWaterTemperature - inletWaterTemperature);
    }

    /// <summary>蒸気流量[kg/s]を計算する</summary>
    /// <param name="inletWaterTemperature">入口水温[C]</param>
    /// <param name="steamPressure">蒸気圧力[kPa]</param>
    /// <param name="fuelConsumption">燃料消費量[Nm3/s, kg/s]</param>
    /// <param name="inletAirTemperature">入口空気温度[C]</param>
    /// <param name="fuel">燃料の種類</param>
    /// <param name="nominalSmokeTemperature">定格排ガス温度[C]</param>
    /// <param name="airRatio">空気比[-]</param>
    /// <param name="heatLossCoefficient">熱損失係数[kW/K]</param>
    /// <param name="ambientTemperature">周囲の温度[C]</param>
    /// <param name="nominalSteamTemperature">定格蒸気温度[C]</param>
    /// <param name="steamFlowRate">出力：蒸気流量[kg/s]</param>
    /// <param name="heatLoad">出力：加熱量[kW]</param>
    public static void GetSteamFlowRate
      (double inletWaterTemperature, double steamPressure, double fuelConsumption, double inletAirTemperature, 
      Fuel fuel, double nominalSmokeTemperature, double airRatio, double heatLossCoefficient, 
      double ambientTemperature, double nominalSteamTemperature, out double steamFlowRate, out double heatLoad)
    {
      double rf = getSmokeCoefficient(airRatio, fuel);
      double twsv = Water.GetSaturationTemperature(steamPressure);
      double hwsv = Water.GetSaturatedVaporEnthalpy(twsv);
      double hwi = inletWaterTemperature * WATER_SPECIFIC_HEAT;
      double qa = AIR_DENSITY * GetTheoreticalAir(fuel) * airRatio * AIR_SPECIFIC_HEAT * inletAirTemperature;
      double qf = 1000d * GetHeatingValue(fuel, true);
      double qs = SMOKE_SPECIFIC_HEAT * rf * nominalSmokeTemperature * (A_ST * twsv / nominalSteamTemperature + B_ST);
      double ql = heatLossCoefficient * (twsv - ambientTemperature);

      heatLoad = (qa + qf - qs) * fuelConsumption - ql;
      steamFlowRate = heatLoad / (hwsv - hwi);
    }

    #endregion

    #region 燃料消費量の計算

    /// <summary>燃料消費量[kg/s, Nm3/s]を計算する</summary>
    /// <param name="heatLoad">加熱負荷[kW]</param>
    /// <param name="outletWaterTemperature">水（蒸気）の出口温度[C]</param>
    /// <param name="inletAirTemperature">入口空気温度[C]</param>
    /// <param name="fuel">燃料の種類</param>
    /// <param name="nominalSmokeTemperature">定格排ガス温度[C]</param>
    /// <param name="airRatio">空気比[-]</param>
    /// <param name="heatLossCoefficient">熱損失係数[kW/K]</param>
    /// <param name="ambientTemperature">周囲の温度[C]</param>
    /// <param name="nominalOutletWaterTemperature">定格出口水温[C]</param>
    /// <returns>燃料消費量[kg/s, Nm3/s]</returns>
    public static double GetFuelConsumption
      (double heatLoad, double outletWaterTemperature, double inletAirTemperature, Fuel fuel,
      double nominalSmokeTemperature, double airRatio, double heatLossCoefficient, 
      double ambientTemperature, double nominalOutletWaterTemperature)
    {
      double smokeTemp = nominalSmokeTemperature 
        * (A_ST * outletWaterTemperature / nominalOutletWaterTemperature + B_ST);
      double m1 = heatLoad + heatLossCoefficient * (outletWaterTemperature - ambientTemperature);
      double m2 = AIR_DENSITY * GetTheoreticalAir(fuel) * airRatio * AIR_SPECIFIC_HEAT * inletAirTemperature 
        + 1000d * GetHeatingValue(fuel, true);
      m2 -= SMOKE_SPECIFIC_HEAT * smokeTemp * getSmokeCoefficient(airRatio, fuel);

      return m1 / m2;
    }
    
    #endregion

    #region 熱損失係数の計算

    /// <summary>熱損失係数[kW/K]を計算する</summary>
    /// <param name="heatLoad">加熱量[kW]</param>
    /// <param name="outletWaterTemperature">水（蒸気）の出口温度[C]</param>
    /// <param name="inletAirTemperature">入口空気温度[C]</param>
    /// <param name="fuel">燃料の種類</param>
    /// <param name="smokeTemperature">排ガス温度[C]</param>
    /// <param name="airRatio">空気比[-]</param>
    /// <param name="fuelConsumption">料消費量[kg/s, Nm3/s]</param>
    /// <param name="ambientTemperature">周囲の温度[C]</param>
    /// <returns>熱損失係数[kW/K]</returns>
    public static double GetHeatLossCoefficient
      (double heatLoad, double outletWaterTemperature, double inletAirTemperature, Fuel fuel,
      double smokeTemperature, double airRatio, double fuelConsumption, double ambientTemperature)
    {
      double airHeat = AIR_DENSITY * fuelConsumption * GetTheoreticalAir(fuel) * airRatio
        * AIR_SPECIFIC_HEAT * inletAirTemperature;
      double fuelHeat = fuelConsumption * GetHeatingValue(fuel, true) * 1000;
      double smokeHeat = SMOKE_SPECIFIC_HEAT * fuelConsumption
        * getSmokeCoefficient(airRatio, fuel) * smokeTemperature;
      double heatLoss = airHeat + fuelHeat - heatLoad - smokeHeat;

      return heatLoss / (outletWaterTemperature - ambientTemperature);
    }

    #endregion

    #region その他の計算

    /// <summary>排ガス熱量計算のための係数を計算する</summary>
    /// <param name="airRatio">空気比[-]</param>
    /// <param name="fuel">燃料の種類</param>
    /// <returns>排ガス熱量計算のための係数</returns>
    private static double getSmokeCoefficient(double airRatio, Fuel fuel)
    {
      double am1 = airRatio - 1d;
      switch (fuel)
      {
        case Fuel.Gas13A: return (0.293 + 0.268 * am1) * 41d;
        case Fuel.LNG: return (0.376 + 0.296 * am1) * 49.1d - 3.91 - 1.36 * am1;
        case Fuel.LPG: return (0.376 + 0.296 * am1) * 45.7d - 3.91 - 1.36 * am1;
        default: return (0.0216 + 0.0241 * am1) * 27.6d + 1.67 + 0.56 * am1;  //石炭
      }
    }

    /// <summary>燃料の発熱量[MJ/Nm3, MJ/kg]を取得する</summary>
    /// <param name="fuel">燃料の種類</param>
    /// <param name="isHighValue">高位発熱量か否か</param>
    /// <returns>燃料の発熱量[MJ/Nm3, MJ/kg]</returns>
    public static double GetHeatingValue(Fuel fuel, bool isHighValue)
    {
      switch (fuel)
      {
        case Fuel.Gas13A:
          if (isHighValue) return 45.6;
          else return 41.0;
        case Fuel.LNG:
          if (isHighValue) return 54.6;
          else return 49.1;
        case Fuel.LPG:
          if (isHighValue) return 50.8;
          else return 45.7;
        default:  //石炭
          if (isHighValue) return 29.0;
          else return 27.6;
      }
    }

    /// <summary>理論空気量[Nm3/Nm3,Nm3/kg]を取得する</summary>
    /// <param name="fuel">燃料の種類</param>
    /// <returns>理論空気量[Nm3/Nm3,Nm3/kg]</returns>
    public static double GetTheoreticalAir(Fuel fuel)
    {
      switch (fuel)
      {
        case Fuel.Gas13A: return 10.949;
        case Fuel.LNG: return 13.093;
        case Fuel.LPG: return 12.045;
        default: return 7.8;  //石炭
      }
    }

    #endregion

  }
}
