﻿/* SimpleCompressionRefrigerator.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;

namespace Popolo.HVAC.HeatSource
{
  /// <summary>特性式による簡易ターボ冷凍機</summary>
  [Serializable]
  public class SimpleCentrifugalChiller : ICentrifugalChiller, ImmutableCentrifugalChiller
  {

    #region 定数宣言

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    #endregion

    #region プロパティ・インスタンス変数

    /// <summary>機器特性係数</summary>
    private readonly double[] coef;

    /// <summary>理論COP[-]</summary>
    private readonly double theoreticalCOP;

    /// <summary>インバータ制御か否か</summary>
    public bool HasInverter { get; private set; }

    /// <summary>運転中か否か</summary>
    public bool IsOperating { get; set; }

    /// <summary>冷水出口温度[C]を取得する</summary>
    public double ChilledWaterOutletTemperature { get; private set; }

    /// <summary>冷水出口温度設定値[C]を設定・取得する</summary>
    public double ChilledWaterOutletSetPointTemperature { get; set; }

    /// <summary>冷水入口温度[C]を取得する</summary>
    public double ChilledWaterInletTemperature { get; private set; }

    /// <summary>冷却水出口温度[C]を取得する</summary>
    public double CoolingWaterOutletTemperature { get; private set; }

    /// <summary>冷却水入口温度[C]を取得する</summary>
    public double CoolingWaterInletTemperature { get; private set; }

    /// <summary>冷水流量[kg/s]を取得する</summary>
    public double ChilledWaterFlowRate { get; private set; }

    /// <summary>冷却水流量[kg/s]を取得する</summary>
    public double CoolingWaterFlowRate { get; private set; }

    /// <summary>定格冷凍能力[kW]を取得する</summary>
    public double NominalCapacity { get; private set; }

    /// <summary>定格入力[kW]を取得する</summary>
    public double NominalInput { get; private set; }

    /// <summary>定格COP[-]を取得する</summary>
    public double NominalCOP { get; private set; }

    /// <summary>容量制御下限値[-]を取得する</summary>
    public double MinimumPartialLoadRatio { get; private set; }

    /// <summary>消費電力[kW]を取得する</summary>
    public double ElectricConsumption { get; private set; }

    /// <summary>冷却負荷[kW]を取得する</summary>
    public double CoolingLoad { get; private set; }

    /// <summary>COP[-]を取得する</summary>
    public double COP { get { return (ElectricConsumption == 0) ? 0 : CoolingLoad / ElectricConsumption; } }

    /// <summary>冷水上限流量[kg/s]を取得する</summary>
    public double MaxChilledWaterFlowRate { get; private set; }

    /// <summary>冷水下限流量比[-]を設定・取得する</summary>
    public double MinChilledWaterFlowRatio { get; set; } = 0.4;

    /// <summary>過負荷か否か</summary>
    public bool IsOverLoad { get; private set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="nominalInput">定格入力[kW]</param>
    /// <param name="minimumPartialLoadRatio">容量制御下限値[-]</param>
    /// <param name="chilledWaterInletTemperature">冷水入口温度[C]</param>
    /// <param name="chilledWaterOutletTemperature">冷水出口水温[C]</param>
    /// <param name="coolingWaterOutletTemperature">冷却水出口水温[C]</param>
    /// <param name="chilledWaterFlowRate">冷水質量流量[kg/s]</param>
    /// <param name="hasInverter">インバータ付か否か</param>
    public SimpleCentrifugalChiller
      (double nominalInput, double minimumPartialLoadRatio, double chilledWaterInletTemperature, 
      double chilledWaterOutletTemperature, double coolingWaterOutletTemperature,
      double chilledWaterFlowRate, bool hasInverter)
    {
      this.NominalCapacity = chilledWaterFlowRate * WATER_SPECIFIC_HEAT
        * (chilledWaterInletTemperature - chilledWaterOutletTemperature);
      this.MaxChilledWaterFlowRate = chilledWaterFlowRate;
      this.NominalInput = nominalInput;
      this.MinimumPartialLoadRatio = minimumPartialLoadRatio;
      this.HasInverter = hasInverter;
      this.theoreticalCOP = (chilledWaterOutletTemperature + 273.15)
        / (coolingWaterOutletTemperature - chilledWaterOutletTemperature);
      this.NominalCOP = NominalCapacity / NominalInput;
      this.ChilledWaterOutletSetPointTemperature = chilledWaterOutletTemperature;

      //INV機の特性係数
      if (HasInverter) coef = new double[] { 4.425e-4, -2.532e-1, 4.256e-1, 3.098e-1, 6.910e-2, 4.482e-1 };
      //定速機の特性係数
      else coef = new double[] { -4.964e-4, 4.865e-1, 5.053e-1, 1.781e-1, -1.923e-1, 2.292e-2 };
    }

    #endregion

    #region publicメソッド

    /// <summary>機器を停止させる</summary>
    public void ShutOff()
    {
      ChilledWaterOutletTemperature = ChilledWaterInletTemperature;
      CoolingWaterOutletTemperature = CoolingWaterInletTemperature;
      CoolingLoad = ChilledWaterFlowRate = CoolingWaterFlowRate = 0;
      ElectricConsumption = 0;
    }

    /// <summary>状態を更新する</summary>
    /// <param name="coolingWaterInletTemperature">冷却水入口温度[C]</param>
    /// <param name="chilledWaterInletTemperature">冷水入口温度[C]</param>
    /// <param name="coolingWaterFlowRate">冷却水質量流量[kg/s]</param>
    /// <param name="chilledWaterFlowRate">冷水質量流量[kg/s]</param>
    public void Update
      (double coolingWaterInletTemperature, double chilledWaterInletTemperature,
      double coolingWaterFlowRate, double chilledWaterFlowRate)
    {
      //状態値を保存
      this.ChilledWaterInletTemperature = chilledWaterInletTemperature;
      this.CoolingWaterInletTemperature = coolingWaterInletTemperature;
      this.ChilledWaterFlowRate = chilledWaterFlowRate;
      this.CoolingWaterFlowRate = coolingWaterFlowRate;

      //非稼働ならば停止処理
      if (!IsOperating)
      {
        ShutOff();
        return;
      }

      //熱容量流量を計算
      double mcch = WATER_SPECIFIC_HEAT * ChilledWaterFlowRate;
      double mccd = WATER_SPECIFIC_HEAT * CoolingWaterFlowRate;

      //負荷[kW]を計算
      CoolingLoad = mcch * (ChilledWaterInletTemperature - ChilledWaterOutletSetPointTemperature);

      //過負荷の場合には出口温度を補正
      IsOverLoad = NominalCapacity < CoolingLoad;
      if (IsOverLoad)
      {
        CoolingLoad = NominalCapacity;
        ChilledWaterOutletTemperature = ChilledWaterInletTemperature - CoolingLoad / mcch;
      }
      else ChilledWaterOutletTemperature = ChilledWaterOutletSetPointTemperature;

      //容量制御下限値未満の場合には下限値として消費電力を計算
      double load = Math.Max(MinimumPartialLoadRatio * NominalCapacity, CoolingLoad);

      //冷却水出口温度の計算
      double tcho = ChilledWaterOutletTemperature + 273.15;
      double tcdi = coolingWaterInletTemperature + 273.15;
      double pL = load / NominalCapacity;
      double dcd = (coef[2] + coef[3] * pL) * theoreticalCOP;
      double ecd = coef[0] + pL * (coef[1] + coef[5] * pL);
      double acd = coef[4] * theoreticalCOP * theoreticalCOP;
      double bcd = tcho * (-2 * acd + dcd - (tcho * mccd) / NominalInput);
      double ccd = tcho * tcho * (acd - dcd + ecd + (load + tcdi * mccd) / NominalInput);

      //2次方程式の解の公式
      CoolingWaterOutletTemperature = (-bcd - Math.Sqrt(bcd * bcd - 4 * acd * ccd)) / (2d * acd) - 273.15;

      //消費電力の計算
      double tCop = tcho / (CoolingWaterOutletTemperature - ChilledWaterOutletTemperature);
      ElectricConsumption = NominalInput * ((acd / tCop + dcd) / tCop + ecd);
    }

    #endregion

  }
}
