﻿/* OutdoorAirHeatExchanger.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.ThermophysicalProperty;
using Popolo.HVAC.HeatExchanger;

namespace Popolo.HVAC.MultiplePackagedHeatPump
{
  /// <summary>室外機熱交換器</summary>
  [Serializable]
  public class OutdoorAirHeatExchanger: ImmutableOutdoorAirHeatExchanger
  {

    #region インスタンス変数・プロパティ

    /// <summary>冷房用熱交換器（凝縮器）</summary>
    private CrossFinCondensor coolingHEX;

    /// <summary>暖房用熱交換器（蒸発器）</summary>
    private CrossFinEvaporator heatingHEX;

    /// <summary>定格風量[kg/s]を取得する</summary>
    public double NominalAirFlowRate
    {
      get
      {
        if (IsCooling) return heatingHEX.NominalAirFlowRate;
        else return coolingHEX.NominalAirFlowRate;
      }
    }
    
    /// <summary>屋外空気条件を設定・取得する</summary>
    public ImmutableMoistAir OutdoorAir { get; set; }

    /// <summary>出口温度[C]を取得する</summary>
    public double OutletAirDrybulbTemperature
    {
      get
      {
        if (IsCooling) return heatingHEX.OutletAirTemperature;
        else return coolingHEX.OutletAirTemperature;
      }
    }

    /// <summary>出口絶対湿度[kg/kg]を取得する</summary>
    public double OutletAirHumidityRatio
    {
      get
      {
        if (IsCooling) return heatingHEX.OutletAirHumidityRatio;
        else return coolingHEX.OutletAirHumidityRatio;
      }
    }

    /// <summary>定格消費電力を設定・取得する</summary>
    public double NominalElectricConsumption { get; private set; }

    /// <summary>稼働率[-]を設定・取得する</summary>
    public double OperatingRate { get; private set; }

    /// <summary>消費電力[kW]を設定・取得する</summary>
    public double ElectricConsumption
    {
      get
      {
        return NominalElectricConsumption * OperatingRate;
      }
    }

    /// <summary>冷房モードか否か</summary>
    public bool IsCooling { get; set; }

    /// <summary>熱交換量[kW]を設定・取得する（冷媒加熱が正）</summary>
    public double HeatExchange { get; set; }

    /// <summary>水噴霧設備を用いるか否か</summary>
    public bool UseWaterSpray
    {
      get { return coolingHEX.UseWaterSpray; }
      set { coolingHEX.UseWaterSpray = value; }
    }

    /// <summary>デフロスト負荷[kW]を取得する</summary>
    public double DefrostLoad { get { return heatingHEX.DefrostLoad; } }

    #endregion

    #region コンストラクタ

    /// <summary>コンストラクタ</summary>
    /// <param name="evaporatingTemperature">定格蒸発温度[C]</param>
    /// <param name="condensingTemperature">定格凝縮温度[C]</param>
    /// <param name="coolingAirDrybulbTemperature">冷房時定格乾球温度[C]</param>
    /// <param name="heatingAirDrybulbTemperature">暖房時定格乾球温度[C]</param>
    /// <param name="coolingAirHumidityRatio">冷房時定格絶対湿度[kg/kg]</param>
    /// <param name="heatingAirHumidityRatio">暖房時定格絶対湿度[kg/kg]</param>
    /// <param name="coolingCapacity">定格冷却能力[kW]</param>
    /// <param name="heatingCapacity">定格暖房能力[kW]</param>
    /// <param name="coolingAirFlowRate">冷房時定格風量[kg/s]</param>
    /// <param name="heatingAirFlowRate">暖房時定格風量[kg/s]</param>
    /// <param name="electricConsumption">定格消費電力[kW]</param>
    /// <param name="borderRalativeHumidity">乾湿境界相対湿度[%]</param>
    public OutdoorAirHeatExchanger
      (double evaporatingTemperature, double condensingTemperature,
      double coolingAirDrybulbTemperature, double heatingAirDrybulbTemperature,
      double coolingAirHumidityRatio, double heatingAirHumidityRatio,
      double coolingCapacity, double heatingCapacity,
      double coolingAirFlowRate, double heatingAirFlowRate,
      double electricConsumption, double borderRalativeHumidity)
    {
      //熱交換器初期化
      heatingHEX = new CrossFinEvaporator
        (evaporatingTemperature, heatingCapacity, heatingAirFlowRate,
        heatingAirDrybulbTemperature, heatingAirHumidityRatio, borderRalativeHumidity);
      coolingHEX = new CrossFinCondensor
        (condensingTemperature, coolingCapacity, coolingAirFlowRate,
        coolingAirDrybulbTemperature, coolingAirHumidityRatio);

      //プロパティ初期化
      NominalElectricConsumption = electricConsumption;
      OperatingRate = 1;
      OutdoorAir = new MoistAir(26, 0.02);
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>必要な冷媒温度[C]を計算する</summary>
    /// <param name="heatLoad">熱負荷[kW]（加熱負荷が正）</param>
    /// <returns>必要な冷媒温度[C]</returns>
    public double GetRefrigerantTemperature(double heatLoad)
    {
      HeatExchange = heatLoad;
      if (IsCooling)
      {
        return coolingHEX.GetCondensingTemperature
          (-HeatExchange, coolingHEX.NominalAirFlowRate,
          OutdoorAir.DrybulbTemperature, OutdoorAir.HumidityRatio);
      }
      else
      {
        return heatingHEX.GetEvaporatingTemperature
          (HeatExchange, heatingHEX.NominalAirFlowRate,
          OutdoorAir.DrybulbTemperature, OutdoorAir.HumidityRatio, true);
      }
    }

    /// <summary>冷媒温度[C]にもとづき状態を更新する</summary>
    /// <param name="refrigerantTemperature">冷媒温度[C]</param>
    public void UpdateState(double refrigerantTemperature)
    {
      double q;
      if (IsCooling)
      {
        q = -coolingHEX.GetHeatTransfer
          (refrigerantTemperature, heatingHEX.NominalAirFlowRate,
          OutdoorAir.DrybulbTemperature, OutdoorAir.HumidityRatio);
      }
      else
      {
        q = heatingHEX.GetHeatTransfer
          (refrigerantTemperature, heatingHEX.NominalAirFlowRate,
          OutdoorAir.DrybulbTemperature, OutdoorAir.HumidityRatio);
      }
      //稼働率の計算
      OperatingRate = HeatExchange / q;
    }

    #endregion    

  }

  #region 読み取り専用インターフェース

  /// <summary>読み取り専用の室外機熱交換器</summary>
  public interface ImmutableOutdoorAirHeatExchanger
  {
    
    /// <summary>定格風量[kg/s]を取得する</summary>
    double NominalAirFlowRate { get; }

    /// <summary>屋外空気条件を取得する</summary>
    ImmutableMoistAir OutdoorAir { get; }

    /// <summary>出口温度[C]を取得する</summary>
    double OutletAirDrybulbTemperature { get; }

    /// <summary>出口絶対湿度[kg/kg]を取得する</summary>
    double OutletAirHumidityRatio { get; }

    /// <summary>定格消費電力を取得する</summary>
    double NominalElectricConsumption { get; }

    /// <summary>稼働率[-]を取得する</summary>
    double OperatingRate { get; }

    /// <summary>消費電力[kW]を取得する</summary>
    double ElectricConsumption { get; }

    /// <summary>冷房モードか否か</summary>
    bool IsCooling { get; }

    /// <summary>熱交換量[kW]を取得する（冷媒加熱が正）</summary>
    double HeatExchange { get; }

    /// <summary>水噴霧設備を用いるか否か</summary>
    bool UseWaterSpray { get; }

    /// <summary>デフロスト負荷[kW]を取得する</summary>
    double DefrostLoad { get; }
  }

  #endregion

}
