﻿/* SimpleSolarCollector.cs
 * 
 * Copyright (C) 2014 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.Weather;

namespace Popolo.HVAC.SolarEnergy
{

  /// <summary>特性式による簡易集熱器</summary>
  [Serializable]
  public class SimpleSolarCollector
  {

    #region 定数宣言

    /// <summary>水の定圧比熱[kJ/kgK]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    #endregion

    #region 列挙型定義

    /// <summary>受熱面タイプ</summary>
    public enum HeatReceiver
    {
      /// <summary>真空ガラス管型</summary>
      VacuumTube,
      /// <summary>平板型</summary>
      FlatPlate
    }

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>傾斜面情報</summary>
    private Incline incline;

    /// <summary>傾斜面情報を取得する</summary>
    public ImmutableIncline Incline { get { return incline; } }

    /// <summary>特性式の係数A[W/K]を取得する</summary>
    public double CoefficientA { get; private set; }

    /// <summary>特性式の係数B[-]を取得する</summary>
    public double CoefficientB { get; private set; }

    /// <summary>集熱効率[-]を取得する</summary>
    public double Efficiency { get; private set; }

    /// <summary>パネル表面積[m2]を取得する</summary>
    public double SurfaceArea { get; private set; }

    /// <summary>入口水温[C]を取得する</summary>
    public double InletWaterTemperature { get; private set; }

    /// <summary>出口水温[C]を取得する</summary>
    public double OutletWaterTemperature { get; private set; }

    /// <summary>水量[kg/s]を取得する</summary>
    public double WaterFlowRate { get; private set; }

    /// <summary>集熱量[kW]を取得する</summary>
    public double HeatCollection
    { get { return (OutletWaterTemperature - InletWaterTemperature) * WaterFlowRate * WATER_SPECIFIC_HEAT; } }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="incline">傾斜面</param>
    /// <param name="surfaceArea">集熱面積[m2]</param>
    /// <param name="coefficientA">特性係数A</param>
    /// <param name="coefficientB">特性係数B</param>
    public SimpleSolarCollector
      (ImmutableIncline incline, double surfaceArea, double coefficientA, double coefficientB)
    {
      initialize(incline, surfaceArea, coefficientA, coefficientB);
    }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="incline">傾斜面</param>
    /// <param name="surfaceArea">集熱面積[m2]</param>
    /// <param name="heatReceiver">受熱面タイプ</param>
    public SimpleSolarCollector(ImmutableIncline incline, double surfaceArea, HeatReceiver heatReceiver)
    {
      switch (heatReceiver)
      {
        case HeatReceiver.FlatPlate:
          initialize(incline, surfaceArea, -4.14, 0.8);
          return;
        default:
          initialize(incline, surfaceArea, -1.75, 0.6);
          return;
      }
    }

    /// <summary>パラメータを初期化する</summary>
    /// <param name="incline">傾斜面</param>
    /// <param name="surfaceArea">集熱面積[m2]</param>
    /// <param name="coefficientA">特性係数A</param>
    /// <param name="coefficientB">特性係数B</param>
    private void initialize
      (ImmutableIncline incline, double surfaceArea, double coefficientA, double coefficientB)
    {
      this.incline = new Incline(incline);
      this.CoefficientA = coefficientA;
      this.CoefficientB = coefficientB;
      this.SurfaceArea = surfaceArea;
    }

    #endregion

    #region publicメソッド

    /// <summary>水量[kg/s]を計算する</summary>
    /// <param name="totalIrradiance">傾斜面入射量[W/m2]</param>
    /// <param name="waterTemperature">入口水温[C]</param>
    /// <param name="temperatureDifferrence">出入口温度差[K]</param>
    /// <param name="ambientTemperature">周囲温度[C]</param>
    /// <returns>水量[kg/s]</returns>
    public double GetWaterFlowRate
      (double totalIrradiance, double waterTemperature, double temperatureDifferrence, double ambientTemperature)
    {
      //入口水温保存
      InletWaterTemperature = waterTemperature;

      //温度差が負または日射0の場合には停止
      if (temperatureDifferrence <= 0 || totalIrradiance <= 0)
      {
        OutletWaterTemperature = InletWaterTemperature;
        Efficiency = 0;
        WaterFlowRate = 0;
        return 0;
      }

      //出口水温[C]     
      OutletWaterTemperature = waterTemperature + temperatureDifferrence;

      //平均集熱温度[C]
      double tM = 0.5 * (InletWaterTemperature + OutletWaterTemperature);

      //外気温度差[K]
      double dT = tM - ambientTemperature;

      //集熱効率[-]
      Efficiency = Math.Max(0, CoefficientA * (dT / totalIrradiance) + CoefficientB);

      //水量[kg/s]
      WaterFlowRate = 0.001 * totalIrradiance * SurfaceArea
        * Efficiency / (temperatureDifferrence * WATER_SPECIFIC_HEAT);
      return WaterFlowRate;
    }

    /// <summary>水量[kg/s]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <param name="waterTemperature">入口水温[C]</param>
    /// <param name="temperatureDifferrence">出入口温度差[K]</param>
    /// <param name="ambientTemperature">周囲温度[C]</param>
    /// <returns>水量[kg/s]</returns>
    public double GetWaterFlowRate
      (ImmutableSun sun, double waterTemperature, double temperatureDifferrence, double ambientTemperature)
    {
      double totalIrradiance = Incline.GetDirectSolarRadiationRate(sun) * sun.DirectNormalRadiation 
        + Incline.ConfigurationFactorToSky * sun.DiffuseHorizontalRadiation;
      return GetWaterFlowRate(totalIrradiance, waterTemperature, temperatureDifferrence, ambientTemperature);
    }

    /// <summary>出口温度[C]を計算する</summary>
    /// <param name="totalIrradiance">傾斜面入射量[W/m2]</param>
    /// <param name="waterTemperature">入口水温[C]</param>
    /// <param name="waterFlowRate">水量[kg/s]</param>
    /// <param name="ambientTemperature">周囲温度[C]</param>
    /// <returns>出口温度[C]</returns>
    public double GetOutletTemperature
      (double totalIrradiance, double waterTemperature, double waterFlowRate, double ambientTemperature)
    {
      //流量が0の場合には停止
      if (waterFlowRate <= 0)
      {
        OutletWaterTemperature = InletWaterTemperature;
        Efficiency = 0;
        WaterFlowRate = 0;
        return 0;
      }

      //出入口温度差の計算
      double dT = 0.001 * SurfaceArea * 
        (CoefficientA * (waterTemperature - ambientTemperature) + CoefficientB * totalIrradiance)
        / (WATER_SPECIFIC_HEAT * waterFlowRate - 0.0005 * SurfaceArea * CoefficientA);

      GetWaterFlowRate(totalIrradiance, waterTemperature, dT, ambientTemperature);
      return OutletWaterTemperature;
    }

    /// <summary>出口温度[C]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <param name="waterTemperature">入口水温[C]</param>
    /// <param name="waterFlowRate">水量[kg/s]</param>
    /// <param name="ambientTemperature">周囲温度[C]</param>
    /// <returns>出口温度[C]</returns>
    public double GetOutletTemperature
      (ImmutableSun sun, double waterTemperature, double waterFlowRate, double ambientTemperature)
    {
      double totalIrradiance = Incline.GetDirectSolarRadiationRate(sun) * sun.DirectNormalRadiation
        + Incline.ConfigurationFactorToSky * sun.DiffuseHorizontalRadiation;
      return GetOutletTemperature(totalIrradiance, waterTemperature, waterFlowRate, ambientTemperature);
    }

    #endregion

  }
}
