﻿/* AHUSystem.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;
using System.Collections.Generic;

using Popolo.Numerics;
using Popolo.Numerics.MatrixOperation;
using Popolo.ThermalLoad;
using Popolo.HVAC.AirConditioner;
using Popolo.ThermophysicalProperty;

namespace Popolo.HVAC.SubSystem
{

  /// <summary>AHUによる二次側空調システム</summary>
  [Serializable]
  public class AHUSystem : IAirConditioningSystemModel
  {

    #region 定数宣言

    /// <summary>湿り空気の定圧比熱[J/(kgK)]</summary>
    /// <remarks>24C, 50%, 9.5g/kg</remarks>
    private const double AIR_SPECIFICHEAT = 1005 + 1846 * 0.0095;

    #endregion

    #region 列挙型定義

    /// <summary>AHU運転モード</summary>
    public enum OperatingMode
    {
      /// <summary>停止</summary>
      ShutOff,
      /// <summary>換気（温調無）</summary>
      Ventilation,
      /// <summary>冷房専用</summary>
      Cooling,
      /// <summary>暖房専用</summary>
      Heating
    }

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>外気条件</summary>
    private ImmutableMoistAir outdoorAir;

    /// <summary>建物熱負荷計算モデル</summary>
    private BuildingThermalModel bModel;

    /// <summary>AHUリスト</summary>
    private AirHandlingUnit[] ahu;

    /// <summary>AHU毎のゾーン制御リスト</summary>
    private VolumeController[][] vlmCtrl;

    /// <summary>AHU制御点が確定済か否か</summary>
    private bool[] ctrlFixed;

    /// <summary>空調機リストを取得する</summary>
    public ImmutableAirHandlingUnit[] AHUs { get { return ahu; } }

    /// <summary>空調機制御リストを取得する</summary>
    public AHUController[] Controllers { get; private set; }

    /// <summary>収束計算の初回か否か</summary>
    private bool isFirstCall = true;

    /// <summary>AHU入口絶対湿度リスト</summary>
    private double[] hrAHUs;

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="bModel">建物熱負荷計算モデル</param>
    /// <param name="ahu">空気調和機リスト</param>
    public AHUSystem(BuildingThermalModel bModel, AirHandlingUnit[] ahu)
    {
      this.bModel = bModel;
      this.ahu = ahu;

      hrAHUs = new double[ahu.Length];//DEBUG
      ctrlFixed = new bool[ahu.Length];
      Controllers = new AHUController[ahu.Length];
      vlmCtrl = new VolumeController[ahu.Length][];
      for (int i = 0; i < ahu.Length; i++)
      {
        Controllers[i] = new AHUController();
        vlmCtrl[i] = new VolumeController[0];
      }
    }

    #endregion

    #region インスタンスメソッド（モデル構築関連）

    /// <summary>AHU外気量を設定する</summary>
    /// <param name="ahuIndex">AHU番号</param>
    /// <param name="minOA">最小外気量[kg/s]</param>
    /// <param name="maxOA">最大外気量[kg/s]</param>
    public void SetOutdoorAirFlow(int ahuIndex, double minOA, double maxOA)
    { ahu[ahuIndex].SetOutdoorAirFlowRange(minOA, maxOA); }

    /// <summary>空調機にVAVを登録する</summary>
    /// <param name="ahuIndex">空調機番号</param>
    /// <param name="vavs">VAVリスト</param>
    public void SetVAV(int ahuIndex, VolumeController[] vavs)
    {
      vlmCtrl[ahuIndex] = vavs;
      Controllers[ahuIndex].IsCAVControl = false;
    }

    /// <summary>空調機にCAVを登録する</summary>
    /// <param name="ahuIndex">空調機番号</param>
    /// <param name="cavs">CAVリスト</param>
    public void SetCAV(int ahuIndex, VolumeController[] cavs)
    {
      vlmCtrl[ahuIndex] = cavs;
      Controllers[ahuIndex].IsCAVControl = true;
    }

    /// <summary>ゾーン温度を制御する</summary>
    /// <param name="ahuIndex">空調機番号</param>
    /// <param name="controlZoneIndex">制御ゾーン番号</param>
    /// <param name="setPointTemperature">設定温度[C]</param>
    public void ControlZoneTemperature
      (int ahuIndex, int controlZoneIndex, double setPointTemperature)
    {
      if (Controllers[ahuIndex].IsCAVControl)
      {
        Controllers[ahuIndex].SetpointTemperature = setPointTemperature;
        Controllers[ahuIndex].IsRATemperatureControl = false;
      }
      else
      {
        VolumeController vc = vlmCtrl[ahuIndex][controlZoneIndex];
        vc.IsShutOff = false;
        vc.SetPointTemperature = setPointTemperature;
      }
    }

    /// <summary>ゾーンの空調を停止する</summary>
    /// <param name="ahuIndex">空調機番号</param>
    /// <param name="controlZoneIndex">制御ゾーン番号</param>
    public void ShutOff(int ahuIndex, int controlZoneIndex)
    { vlmCtrl[ahuIndex][controlZoneIndex].IsShutOff = true; }

    /// <summary>風量コントローラを取得する</summary>
    /// <param name="ahuIndex">AHU番号</param>
    /// <param name="vavNumber">VAV番号</param>
    /// <returns>風量コントローラ</returns>
    public VolumeController GetVolumeController(int ahuIndex, int vavNumber)
    { return vlmCtrl[ahuIndex][vavNumber]; }

    /// <summary>風量コントローラの数を取得する</summary>
    /// <param name="ahuIndex">AHU番号</param>
    /// <returns>風量コントローラの数</returns>
    public int GetVolumeControllerNumber(int ahuIndex)
    { return vlmCtrl[ahuIndex].Length; }

    #endregion

    #region IAirConditioningSystemModel実装

    /// <summary>現在の日時を設定・取得する</summary>
    public DateTime CurrentDateTime { get; set; }

    /// <summary>タイムステップを設定・取得する</summary>
    public double TimeStep { get; set; }

    /// <summary>外気条件を設定・取得する</summary>
    public ImmutableMoistAir OutdoorAir
    {
      get { return outdoorAir; }
      set
      {
        outdoorAir = value;
        for (int i = 0; i < ahu.Length; i++)
        {
          ahu[i].OATemperature = value.DrybulbTemperature;
          ahu[i].OAHumidityRatio = value.HumidityRatio;
        }
      }
    }

    /// <summary>計算対象の建物熱負荷計算モデルを取得する</summary>
    public ImmutableBuildingThermalModel BuildingThermalModel
    { get { return bModel; } }

    /// <summary>冷水往温度[C]を取得する</summary>
    public double ChilledWaterSupplyTemperature { get; private set; }

    /// <summary>冷水還温度[C]を取得する</summary>
    public double ChilledWaterReturnTemperature { get; private set; }

    /// <summary>冷水流量[kg/s]を取得する</summary>
    public double ChilledWaterFlowRate { get; private set; }

    /// <summary>温水往温度[C]を取得する</summary>
    public double HotWaterSupplyTemperature { get; private set; }

    /// <summary>温水還温度[C]を取得する</summary>
    public double HotWaterReturnTemperature { get; private set; }

    /// <summary>温水流量[kg/s]を取得する</summary>
    public double HotWaterFlowRate { get; private set; }

    /// <summary>冷温水還温度を予想する</summary>
    /// <param name="chilledWaterSupplyTemperature">冷水往温度[C]</param>
    /// <param name="hotWaterSupplyTemperature">温水往温度[C]</param>
    public void ForecastReturnWaterTemperature
      (double chilledWaterSupplyTemperature, double hotWaterSupplyTemperature)
    {
      ChilledWaterSupplyTemperature = chilledWaterSupplyTemperature;
      HotWaterSupplyTemperature = hotWaterSupplyTemperature;

      if (isFirstCall)
      {
        isFirstCall = false;
        for (int i = 0; i < ahu.Length; i++)
        {
          hrAHUs[i] = 0;
          double mRA = 0;
          foreach (VolumeController vc in vlmCtrl[i])
          {
            ImmutableZone zn = bModel.MultiRoom[vc.RoomIndex].Zones[vc.ZoneIndex];
            hrAHUs[i] += zn.HumidityRatio * vc.MaxRAFlow;
            mRA += vc.MaxRAFlow;
          }
          hrAHUs[i] /= mRA;
        }
      }
      for (int i = 0; i < ahu.Length; i++)
        if (Controllers[i].Mode != OperatingMode.ShutOff) ahu[i].RAHumidityRatio = hrAHUs[i];

      //AHUの冷温水と風量初期化
      bool isAllAHUShutoff = true;
      for (int i = 0; i < ahu.Length; i++)
      {
        //冷温水コイルに通水する
        ahu[i].ChilledWaterInletTemperature = chilledWaterSupplyTemperature;
        ahu[i].HotWaterInletTemperature = hotWaterSupplyTemperature;

        //給気風量を初期化
        double maxSA = 0;
        double maxRA = 0;
        if (Controllers[i].Mode != OperatingMode.ShutOff)
        {
          bool allClosed = true;
          foreach (VolumeController ctrl in vlmCtrl[i])
          {
            if (!ctrl.IsShutOff) allClosed = false;
            ctrl.SAFlow = ctrl.MaxSAFlow;
            maxSA += ctrl.MaxSAFlow;
            maxRA += ctrl.MaxRAFlow;
          }
          //全VAV,CAVが停止の場合にはAHUも停止させる
          if (allClosed) Controllers[i].Mode = OperatingMode.ShutOff;
          else ahu[i].SetAirFlowRate(maxRA, maxSA);
        }

        if (Controllers[i].Mode == OperatingMode.ShutOff)
        {
          ahu[i].ShutOff();
          ctrlFixed[i] = true;
          for (int j = 0; j < vlmCtrl[i].Length; j++) vlmCtrl[i][j].SAFlow = 0; //2017.12.02追加
        }
        else
        {
          isAllAHUShutoff = false;
          ctrlFixed[i] = false;
        }
      }

      //すべてのAHUが非稼働の場合には給気量0で成行計算を1回実行して終了
      if (isAllAHUShutoff)
      {
        for (int i = 0; i < ahu.Length; i++)
        {
          foreach (VolumeController vc in vlmCtrl[i])
          {
            bModel.SetSupplyAir(vc.RoomIndex, vc.ZoneIndex, 0, 0, 0);
            bModel.ControlHeatSupply(vc.RoomIndex, vc.ZoneIndex, 0);
            bModel.ControlWaterSupply(vc.RoomIndex, vc.ZoneIndex, 0);
          }
        }
        bModel.ForecastHeatTransfer();
        bModel.ForecastWaterTransfer();
        updateReturnWaterState();
        return;
      }

      //顕熱平衡の収束計算
      solveHeatTransfer();

      //水分平衡の計算
      for (int i = 0; i < ahu.Length; i++)
      {
        foreach (VolumeController vc in vlmCtrl[i])
        {
          if (Controllers[i].Mode == OperatingMode.ShutOff)
            bModel.SetSupplyAir(vc.RoomIndex, vc.ZoneIndex, 0, 0, 0);
          else bModel.SetSupplyAir
              (vc.RoomIndex, vc.ZoneIndex, ahu[i].SATemperature, ahu[i].SAHumidityRatio, vc.SAFlow);
          bModel.ControlHeatSupply(vc.RoomIndex, vc.ZoneIndex, 0);
          bModel.ControlWaterSupply(vc.RoomIndex, vc.ZoneIndex, 0);
        }
      }
      bModel.ForecastWaterTransfer();

      //冷温水量を更新
      updateReturnWaterState();
    }

    /// <summary>状態を確定する</summary>
    public void FixState() { isFirstCall = true; }

    #endregion

    #region インスタンスメソッド（平衡状態計算関連）

    /// <summary>顕熱平衡を解く</summary>
    private void solveHeatTransfer()
    {
      //完全に制御した場合の負荷にもとづいて収束計算初期値を作成
      double[] initVec = new double[ahu.Length];
      for (int i = 0; i < ahu.Length; i++)
      {
        if (Controllers[i].Mode == OperatingMode.Cooling 
          || Controllers[i].Mode == OperatingMode.Heating)
        {
          foreach (VolumeController vc in vlmCtrl[i])
          {
            double sp;
            if (Controllers[i].IsCAVControl) sp = Controllers[i].SetpointTemperature;
            else sp = vc.SetPointTemperature;
            bModel.ControlDrybulbTemperature(vc.RoomIndex, vc.ZoneIndex, sp);
            bModel.SetSupplyAir(vc.RoomIndex, vc.ZoneIndex, 0, 0, 0); //DEBUG

            //加湿系統は潜熱負荷を計算する
            if (Controllers[i].Mode == OperatingMode.Heating 
              && ahu[i].Humidifier != AirHandlingUnit.HumidifierType.None)
              bModel.ControlHumidityRatio(vc.RoomIndex, vc.ZoneIndex, Controllers[i].MinimumHumidity);
          }
        }
      }
      bModel.ForecastHeatTransfer();
      bModel.ForecastWaterTransfer();

      for (int i = 0; i < ahu.Length; i++)
      {
        initVec[i] = 0;
        Controllers[i].splyHumidSP = 0;
        double sHL = 0;
        double mRA = 0;
        double mSA = 0;
        foreach (VolumeController ctrl in vlmCtrl[i])
        {
          ImmutableZone zn = bModel.MultiRoom[ctrl.RoomIndex].Zones[ctrl.ZoneIndex];
          initVec[i] += zn.Temperature * ctrl.RAFlow;
          sHL += zn.HeatSupply;
          mRA += ctrl.MaxRAFlow;
          mSA += ctrl.MaxSAFlow;

          //加湿系統のAHU出口湿度を計算する
          if (Controllers[i].Mode == OperatingMode.Heating
            && ahu[i].Humidifier != AirHandlingUnit.HumidifierType.None)
          {
            double wAHUo = zn.WaterSupply / ctrl.MaxSAFlow + zn.HumidityRatio;
            Controllers[i].splyHumidSP = Math.Max(Controllers[i].splyHumidSP, wAHUo);
          }
        }
        initVec[i] = initVec[i] / mRA + sHL / (mSA * AIR_SPECIFICHEAT);
      }

      //過負荷に変化するAHUがなくなるまで繰り返し計算
      bool lastCalc = false;
      while (true)
      {
        //収束計算対象の給気温度を初期化
        List<double> vars = new List<double>();
        for (int i = 0; i < ahu.Length; i++)
        {
          AHUController ctr = Controllers[i];
          if (ctr.Mode != OperatingMode.ShutOff 
            && (ctrlFixed[i] || ctr.Mode == OperatingMode.Ventilation || ctr.IsCAVControl))
            vars.Add(initVec[i]);
        }

        //収束計算実行
        IVector tandh = new Vector(vars.Count);
        IVector fx = new Vector(vars.Count);
        if (vars.Count != 0)
        {
          int iter;
          for (int i = 0; i < tandh.Length; i++) tandh[i] = vars[i];
          //絶対誤差0.01度未満まで収束計算
          double err;
          MultiRoots.Newton(errorFnc, ref tandh, 1e-3, 1e-3, 50, out iter, out err);
        }
        else errorFnc(tandh, ref fx);

        //過負荷に変化する系統がなくなればVAV風量を最小化して最終の計算処理
        if (!overLoadAHUChanged(tandh))
        {
          if (lastCalc) return;
          else if (minimizeVAV(initVec)) lastCalc = true;
          else return;
        }
      }
    }

    /// <summary>誤差評価関数</summary>
    /// <param name="vecX">AHU吹出温湿度ベクトル</param>
    /// <param name="vecF">誤差ベクトル</param>
    private void errorFnc(IVector vecX, ref IVector vecF)
    {
      //制御点確定済・換気・CAV制御の場合には給気温度が収束計算対象
      int indx = 0;
      for (int i = 0; i < ahu.Length; i++)
      {
        AHUController ctr = Controllers[i];
        if (ctr.Mode != OperatingMode.ShutOff &&
          (ctrlFixed[i] || ctr.Mode == OperatingMode.Ventilation || ctr.IsCAVControl))
        {
          foreach (VolumeController ctrl in vlmCtrl[i])
          {
            bModel.SetSupplyAir(ctrl.RoomIndex, ctrl.ZoneIndex, vecX[indx], 0, ctrl.SAFlow);
            bModel.ControlHeatSupply(ctrl.RoomIndex, ctrl.ZoneIndex, 0);
          }
          indx++;
        }
      }
      //顕熱平衡を更新
      bModel.ForecastHeatTransfer();

      //誤差集計
      indx = 0;
      for (int i = 0; i < ahu.Length; i++)
      {
        AHUController ctr = Controllers[i];
        if (ctr.Mode != OperatingMode.ShutOff)
        {
          //還気温度を計算
          double tra = 0;
          double mra = 0;
          foreach (VolumeController ctrl in vlmCtrl[i])
          {
            ImmutableZone zn = bModel.MultiRoom[ctrl.RoomIndex].Zones[ctrl.ZoneIndex];
            tra += zn.Temperature * ctrl.RAFlow;
            mra += ctrl.RAFlow;
          }
          tra /= mra;
          ahu[i].RATemperature = tra;

          //制御点確定済または換気のみのAHUの給気温度の誤差を計算
          if (ctrlFixed[i] || ctr.Mode == OperatingMode.Ventilation)
          {
            //AHU成行計算処理
            if (ctr.Mode == OperatingMode.Cooling) ahu[i].CoolAir();
            else if (ctr.Mode == OperatingMode.Heating) ahu[i].HeatAir();
            else ahu[i].Ventilate();
            //誤差評価
            vecF[indx] = Math.Abs(ahu[i].SATemperature - vecX[indx]);
            indx++;
          }
          //軽負荷CAV系統の制御誤差評価
          else if (ctr.IsCAVControl)
          {
            double sp = ctr.SetpointTemperature;
            if (ctr.IsRATemperatureControl) vecF[indx] = Math.Abs(tra - sp);
            else
            {
              ImmutableZone zn = bModel.MultiRoom[ctr.TargetRoomIndex].Zones[ctr.TargetZoneIndex];
              vecF[indx] = Math.Abs(zn.Temperature - sp);
            }
            indx++;
          }
        }
      }
    }

    /// <summary>過負荷系統のAHU台数に変化は無いか</summary>
    /// <param name="vec">状態変数リスト</param>
    /// <returns>変化の有無</returns>
    private bool overLoadAHUChanged(IVector vec)
    {
      bool incrsd = false;
      int indx = 0;
      for (int i = 0; i < ahu.Length; i++)
      {
        AHUController ctr = Controllers[i];
        if (ctr.Mode != OperatingMode.ShutOff && !ctrlFixed[i]
          && (ctr.Mode == OperatingMode.Cooling || ctr.Mode == OperatingMode.Heating))
        {
          //還気温度を計算
          ahu[i].RATemperature = 0;
          double mra = 0;
          foreach (VolumeController ctrl in vlmCtrl[i])
          {
            ImmutableZone zn = bModel.MultiRoom[ctrl.RoomIndex].Zones[ctrl.ZoneIndex];
            ahu[i].RATemperature += zn.Temperature * ctrl.RAFlow;
            mra += ctrl.RAFlow;
          }
          ahu[i].RATemperature /= mra;

          //過負荷判定
          double tSP;
          //CAV制御の場合には収束計算対象の状態変数が出口温度設定値
          if (ctr.IsCAVControl) tSP = vec[indx];
          //VAVの場合には顕熱負荷にもとづき出口温度設定値を計算
          else
          {
            if (ctr.Mode == OperatingMode.Cooling) tSP = 60;
            else tSP = -10;
            foreach (VolumeController ctrl in vlmCtrl[i])
            {
              ImmutableZone zn = bModel.MultiRoom[ctrl.RoomIndex].Zones[ctrl.ZoneIndex];
              double bf = zn.Temperature + zn.HeatSupply / (AIR_SPECIFICHEAT * ctrl.SAFlow);
              if (ctr.Mode == OperatingMode.Cooling) tSP = Math.Min(tSP, bf);
              else tSP = Math.Max(tSP, bf);
            }
          }
          //出口温度実現可能か
          if (ctr.Mode == OperatingMode.Cooling) ctrlFixed[i] = !ahu[i].CoolAir(tSP, 0);
          else ctrlFixed[i] = !ahu[i].HeatAir(tSP, ctr.splyHumidSP);
          if (ctrlFixed[i]) incrsd = true;
        }

        if (ctr.Mode != OperatingMode.ShutOff &&
          (ctrlFixed[i] || ctr.Mode == OperatingMode.Ventilation || ctr.IsCAVControl))
          indx++;
      }
      return incrsd;
    }

    /// <summary>VAV風量を最小化</summary>
    /// <remarks>VAVが無ければfalseを出力</remarks>
    private bool minimizeVAV(double[] initVec)
    {
      bool hasVAV = false;
      for (int i = 0; i < ahu.Length; i++)
      {
        //制御点未確定でVAV制御のAHUについて計算
        if (Controllers[i].Mode != OperatingMode.ShutOff && !Controllers[i].IsCAVControl && !ctrlFixed[i])
        {
          bool isCooling = (Controllers[i].Mode == OperatingMode.Cooling);
          int vcNum = vlmCtrl[i].Length;
          bool[] vavSt = new bool[vcNum];
          double[] znT = new double[vcNum];
          double[] znW = new double[vcNum];
          double[] znHL = new double[vcNum];
          double[] minSA = new double[vcNum];
          double[] maxSA = new double[vcNum];
          double[] maxRA = new double[vcNum];
          for (int j = 0; j < vlmCtrl[i].Length; j++)
          {
            VolumeController vc = vlmCtrl[i][j];
            ImmutableZone zn = bModel.MultiRoom[vc.RoomIndex].Zones[vc.ZoneIndex];
            znT[j] = zn.Temperature;
            znW[j] = zn.HumidityRatio;
            znHL[j] = zn.HeatSupply;
            vavSt[j] = vc.IsShutOff;
            minSA[j] = vc.MinSAFlow;
            maxSA[j] = vc.MaxSAFlow;
            maxRA[j] = vc.MaxRAFlow;
          }
          bool sc;
          double[] aFlow = ahu[i].OptimizeVAV
            (isCooling, Controllers[i].splyHumidSP, vavSt, znT, znW, znHL, minSA, maxSA, maxRA, out sc);
          for (int j = 0; j < vlmCtrl[i].Length; j++) vlmCtrl[i][j].SAFlow = aFlow[j];          

          //過剰処理系統があれば制御点を確定して収束計算実施
          if (!sc)
          {
            hasVAV = true;
            ctrlFixed[i] = true;
            initVec[i] = ahu[i].SATemperature; //収束計算用給気温度仮定値を更新
          }
        }
      }
      return hasVAV;
    }

    /// <summary>還冷温水状態を更新する</summary>
    private void updateReturnWaterState()
    {
      double tcw, thw;
      tcw = thw = ChilledWaterFlowRate = HotWaterFlowRate = 0;
      foreach (AirHandlingUnit ah in ahu)
      {
        tcw += ah.CoolingCoil.OutletWaterTemperature * ah.CoolingCoil.WaterFlowRate;
        thw += ah.HeatingCoil.OutletWaterTemperature * ah.HeatingCoil.WaterFlowRate;
        ChilledWaterFlowRate += ah.CoolingCoil.WaterFlowRate;
        HotWaterFlowRate += ah.HeatingCoil.WaterFlowRate;
      }
      if (ChilledWaterFlowRate == 0) ChilledWaterReturnTemperature = ChilledWaterSupplyTemperature;
      else ChilledWaterReturnTemperature = tcw / ChilledWaterFlowRate;
      if (HotWaterFlowRate == 0) HotWaterReturnTemperature = HotWaterSupplyTemperature;
      else HotWaterReturnTemperature = thw / HotWaterFlowRate;
    }

    #endregion

    #region インナークラスの定義

    /// <summary>VAVまたはCAV制御クラス</summary>
    [Serializable]
    public class VolumeController
    {
      /// <summary>給気量[kg/s]を取得する</summary>
      public double SAFlow { get; internal set; }

      /// <summary>還気量[kg/s]を取得する</summary>
      public double RAFlow { get { return SAFlow - (MaxSAFlow - MaxRAFlow); } }

      /// <summary>最大給気量[kg/g]を取得する</summary>
      public double MaxSAFlow { get; internal set; }

      /// <summary>最大還気量[kg/g]を取得する</summary>
      public double MaxRAFlow { get; internal set; }

      /// <summary>最小給気量[kg/g]を取得する</summary>
      public double MinSAFlow { get; internal set; }

      /// <summary>給気する室番号を取得する</summary>
      public int RoomIndex { get; internal set; }

      /// <summary>給気するゾーン番号を取得する</summary>
      public int ZoneIndex { get; internal set; }

      /// <summary>温度設定値を取得する</summary>
      public double SetPointTemperature { get; internal set; }

      /// <summary>停止中か否かを取得する</summary>
      public bool IsShutOff { get; internal set; }

      /// <summary>インスタンスを初期化する</summary>
      /// <param name="rmIndex">多数室番号</param>
      /// <param name="znIndex">ゾーン番号</param>
      /// <param name="maxSAFlow">最大給気量[kg/s]</param>
      /// <param name="maxRAFlow">最大還気量[kg/s]</param>
      /// <param name="minSAFlow">最小給気量[kg/s]</param>
      public VolumeController(int rmIndex, int znIndex, double maxSAFlow, double maxRAFlow, double minSAFlow)
      {
        RoomIndex = rmIndex;
        ZoneIndex = znIndex;
        MaxSAFlow = maxSAFlow;
        MaxRAFlow = maxRAFlow;
        MinSAFlow = minSAFlow;
      }
    }

    /// <summary>AHU制御クラス</summary>
    [Serializable]
    public class AHUController
    {
      /// <summary>インスタンスを初期化する</summary>
      internal AHUController() { }

      /// <summary>運転モードを設定・取得する</summary>
      public OperatingMode Mode { get; set; } = OperatingMode.ShutOff;

      /// <summary>CAV制御か否かを取得する</summary>
      public bool IsCAVControl { get; internal set; } = true;

      /// <summary>還気温度制御か否かを設定・取得する</summary>
      public bool IsRATemperatureControl { get; set; } = true;

      /// <summary>制御対象室番号を設定・取得する</summary>
      public int TargetRoomIndex { get; set; }

      /// <summary>制御対象ゾーン番号を設定・取得する</summary>
      public int TargetZoneIndex { get; set; }

      /// <summary>CAV温度設定値[C]を設定・取得する</summary>
      public double SetpointTemperature { get; set; } = 24.0;

      /// <summary>最低湿度[kg/kg]を設定・取得する</summary>
      public double MinimumHumidity { get; set; }

      /// <summary>給気湿度設定値[kg/kg]を設定・取得する</summary>
      internal double splyHumidSP { get; set; }
    }

    #endregion

  }
}
