﻿/* DirectFiredAbsorptionChillerSystem.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.HVAC.HeatSource;
using Popolo.HVAC.HeatExchanger;
using Popolo.HVAC.Circuit;
using Popolo.ThermophysicalProperty;
using Popolo.Numerics;

namespace Popolo.HVAC.SubSystem
{
  /// <summary>直焚吸収冷温水機による熱源サブシステム</summary>
  [Serializable]
  public class DirectFiredAbsorptionChillerSystem: IHeatSourceSubSystem
  {

    #region 定数宣言
    
    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>直焚吸収冷温水機</summary>
    private DirectFiredAbsorptionChiller chiller;

    /// <summary>ポンプ</summary>
    private CentrifugalPump cdwPump, chwPump, hwPump;

    /// <summary>冷却塔</summary>
    private CoolingTower cTower;

    /// <summary>直焚吸収冷温水機を取得する</summary>
    public ImmutableDirectFiredAbsorptionChiller DirectFiredAbsorptionChiller { get { return chiller; } }

    /// <summary>冷水ポンプを取得する</summary>
    public ImmutableCentrifugalPump ChilledWaterPump { get { return chwPump; } }

    /// <summary>温水ポンプを取得する</summary>
    public ImmutableCentrifugalPump HotWaterPump { get { return hwPump; } }

    /// <summary>冷却水ポンプを取得する</summary>
    public ImmutableCentrifugalPump CoolingWaterPump { get { return cdwPump; } }

    /// <summary>冷却塔を取得する</summary>
    public ImmutableCoolingTower CoolingTower { get { return cTower; } }

    /// <summary>冷凍機のセル数を取得する</summary>
    public int ChillerNumber { get; private set; }

    /// <summary>冷却塔の台数（1台の冷凍機あたり）を取得する</summary>
    public int CoolingTowerNumber { get; private set; }

    /// <summary>運転台数を取得する</summary>
    public int OperatingChillerNumber { get; private set; }

    /// <summary>冷却塔を1対1で稼働させるか否か</summary>
    public bool OperateCoolingTowerOneOnOne { get; set; } = true;

    /// <summary>冷却水温度を制御するか否か</summary>
    public bool ControlCoolingWaterTemperature { get; set; }

    /// <summary>冷却水流量を制御するか否か（負荷率比例）</summary>
    public bool ControlCoolingWaterFlowRate { get; set; }

    /// <summary>最小冷却水流量比[-]を設定・取得する</summary>
    public double MinimumCoolingWaterFlowRatio { get; set; } = 0.5;

    /// <summary>冷却水温度設定値[C]を設定・取得する</summary>
    public double CoolingWaterTemperatureSetpoint { get; set; } = 32;

    #endregion

    #region IHeatSourceSubSystem実装

    /// <summary>冷水供給が過負荷か否か</summary>
    public bool IsOverLoad_C { get; private set; }

    /// <summary>温水供給が過負荷か否か</summary>
    public bool IsOverLoad_H { get; private set; }

    /// <summary>運転可能モードを取得する</summary>
    public HeatSourceSystemModel.OperatingMode SelectableMode { get; }

    /// <summary>運転モードを設定・取得する</summary>
    public HeatSourceSystemModel.OperatingMode Mode { get; set; }

    /// <summary>現在の日時を設定・取得する</summary>
    public DateTime CurrentDateTime { get; set; }

    /// <summary>タイムステップを設定・取得する</summary>
    public double TimeStep { get; set; }

    /// <summary>温水還温度[C]を設定・取得する</summary>
    public double HotWaterReturnTemperature { get; set; } = 40;

    /// <summary>温水往温度設定値[C]を設定・取得する</summary>
    public double HotWaterSupplyTemperatureSetpoint { get; set; } = 45;

    /// <summary>温水往温度[C]を取得する</summary>
    public double HotWaterSupplyTemperature { get; private set; } = 45;

    /// <summary>温水流量[kg/s]を取得する</summary>
    public double HotWaterFlowRate { get; private set; }

    /// <summary>温水上限流量[kg/s]を取得する</summary>
    public double MaxHotWaterFlowRate
    { get { return chiller.MaxHotWaterFlowRate * ChillerNumber; } }

    /// <summary>温水下限流量比[-]を取得する</summary>
    public double MinHotWaterFlowRatio
    { get { return chiller.MinHotWaterFlowRate / MaxHotWaterFlowRate; } }

    /// <summary>冷水還温度[C]を設定・取得する</summary>
    public double ChilledWaterReturnTemperature { get; set; } = 12;

    /// <summary>冷水往温度設定値[C]を設定・取得する</summary>
    public double ChilledWaterSupplyTemperatureSetpoint { get; set; } = 7;

    /// <summary>冷水往温度[C]を取得する</summary>
    public double ChilledWaterSupplyTemperature { get; private set; } = 7;

    /// <summary>冷水流量[kg/s]を取得する</summary>
    public double ChilledWaterFlowRate { get; private set; }

    /// <summary>冷水上限流量[kg/s]を取得する</summary>
    public double MaxChilledWaterFlowRate
    { get { return chiller.MaxChilledWaterFlowRate * ChillerNumber; } }

    /// <summary>冷水下限流量比[-]を取得する</summary>
    public double MinChilledWaterFlowRatio
    { get { return chiller.MinChilledWaterFlowRate / MaxChilledWaterFlowRate; } }

    /// <summary>外気条件を設定・取得する</summary>
    public ImmutableMoistAir OutdoorAir { get; set; }

    /// <summary>熱源サブシステムを停止させる</summary>
    public void ShutOff()
    {
      IsOverLoad_C = IsOverLoad_H = false;
      OperatingChillerNumber = 0;
      chiller.ShutOff();
      chwPump.ShutOff();
      hwPump.ShutOff();
      cdwPump.ShutOff();
      cTower.ShutOff();
    }

    /// <summary>冷温水往温度を予測する</summary>
    /// <param name="chilledWaterFlowRate">冷水流量[kg/s]</param>
    /// <param name="hotWaterFlowRate">温水流量[kg/s]</param>
    public void ForecastSupplyWaterTemperature(double chilledWaterFlowRate, double hotWaterFlowRate)
    {
      ChilledWaterFlowRate = chilledWaterFlowRate;
      HotWaterFlowRate = hotWaterFlowRate;

      if (Mode == HeatSourceSystemModel.OperatingMode.Cooling) forecastCooling(chilledWaterFlowRate);
      else if (Mode == HeatSourceSystemModel.OperatingMode.Heating) forecastHeating(hotWaterFlowRate);
      else ShutOff();
    }

    /// <summary>将来状態を予測する（暖房）</summary>
    /// <param name="hotWaterFlowRate">温水流量[kg/s]</param>
    private void forecastHeating(double hotWaterFlowRate)
    {
      if (hotWaterFlowRate == 0)
      {
        ShutOff();
        return;
      }
      chiller.IsCoolingMode = false;

      ChilledWaterSupplyTemperature = ChilledWaterReturnTemperature;
      cTower.ShutOff();
      cdwPump.ShutOff();
      chwPump.ShutOff();

      OperatingChillerNumber = (int)Math.Ceiling(hotWaterFlowRate / chiller.MaxHotWaterFlowRate);
      OperatingChillerNumber = Math.Min(ChillerNumber, OperatingChillerNumber);
      chiller.OutletWaterSetPointTemperature = HotWaterSupplyTemperatureSetpoint;
      while (true)
      {
        //ポンプによる昇温を評価
        double hwFlow = hotWaterFlowRate / OperatingChillerNumber;
        hwPump.UpdateState(0.001 * hwFlow);
        double twi = HotWaterReturnTemperature + hwPump.GetElectricConsumption() / (WATER_SPECIFIC_HEAT * hwFlow);

        chiller.Update(32, twi, 0, hwFlow);
        if (OperatingChillerNumber == ChillerNumber || !chiller.IsOverLoad) break;
        else OperatingChillerNumber++;
      }

      IsOverLoad_H = chiller.IsOverLoad;
      if (IsOverLoad_H) HotWaterSupplyTemperature = chiller.OutletWaterTemperature;
      else HotWaterSupplyTemperature = HotWaterSupplyTemperatureSetpoint;
    }

    /// <summary>将来状態を予測する（冷房）</summary>
    /// <param name="chilledWaterFlowRate">冷水流量[kg/s]</param>
    private void forecastCooling(double chilledWaterFlowRate)
    {
      if (chilledWaterFlowRate == 0)
      {
        ShutOff();
        return;
      }
      chiller.IsCoolingMode = true;

      HotWaterSupplyTemperature = HotWaterReturnTemperature;
      cTower.SetOutdoorAirState(OutdoorAir.WetbulbTemperature, OutdoorAir.HumidityRatio);
      hwPump.ShutOff();

      OperatingChillerNumber = (int)Math.Ceiling(chilledWaterFlowRate / chiller.MaxChilledWaterFlowRate);
      OperatingChillerNumber = Math.Min(ChillerNumber, OperatingChillerNumber);
      chiller.OutletWaterSetPointTemperature = ChilledWaterSupplyTemperatureSetpoint;
      while (true)
      {
        //冷水・冷却水流量を計算
        double chwFlow = chilledWaterFlowRate / OperatingChillerNumber;
        double pLoad = chwFlow / chiller.MaxChilledWaterFlowRate;
        double cdwFlow;
        if (ControlCoolingWaterFlowRate) cdwFlow = cTower.WaterFlowRate 
            = cTower.MaxWaterFlowRate * Math.Max(pLoad, MinimumCoolingWaterFlowRatio);  //2017.12.15 BugFix
        else cdwFlow = cTower.WaterFlowRate = cTower.MaxWaterFlowRate; //2017.12.15 BugFix
        if (!OperateCoolingTowerOneOnOne) cdwFlow /= ChillerNumber;

        //ポンプによる昇温を評価
        cdwPump.UpdateState(0.001 * cdwFlow);
        double dCDT = cdwPump.GetElectricConsumption() / (WATER_SPECIFIC_HEAT * cdwFlow);
        chwPump.UpdateState(0.001 * chwFlow);
        double twi = ChilledWaterReturnTemperature 
          + chwPump.GetElectricConsumption() / (WATER_SPECIFIC_HEAT * chwFlow);

        //冷却塔と冷凍機の連成計算（冷却水温度の計算）
        bool needIteration = !ControlCoolingWaterTemperature;
        if (ControlCoolingWaterTemperature)
        {
          chiller.Update(CoolingWaterTemperatureSetpoint + dCDT, twi, cdwFlow, chwFlow);
          cTower.OutletWaterSetPointTemperature = CoolingWaterTemperatureSetpoint;
          cTower.Update(chiller.CoolingWaterOutletTemperature, true);
          if (cTower.IsOverLoad) needIteration = true;
        }
        //過負荷または冷却水温度成行の場合には収束計算
        if (needIteration)
        {
          Roots.ErrorFunction eFnc = delegate (double cdt)
          {
            chiller.Update(cdt + dCDT, twi, cdwFlow, chwFlow);
            cTower.Update(chiller.CoolingWaterOutletTemperature, cTower.MaxAirFlowRate);
            return cTower.OutletWaterTemperature - cdt;
          };
          if (eFnc(10) < 0) break;
          else if (0 < eFnc(32)) break;
          else Roots.Bisection(eFnc, 10, 32, 0.01, 0.001, 10);
        }

        if (OperatingChillerNumber == ChillerNumber || !chiller.IsOverLoad) break;
        else OperatingChillerNumber++;
      }

      IsOverLoad_C = chiller.IsOverLoad;
      if (IsOverLoad_C) ChilledWaterSupplyTemperature = chiller.OutletWaterTemperature;
      else ChilledWaterSupplyTemperature = ChilledWaterSupplyTemperatureSetpoint;
    }

    /// <summary>状態を確定する</summary>
    public void FixState() { }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="chiller">直焚吸収冷温水発生機</param>
    /// <param name="chwPump">冷水ポンプ</param>
    /// <param name="hwPump">温水ポンプ</param>
    /// <param name="cdwPump">冷却水ポンプ</param>
    /// <param name="cTower">冷却塔</param>
    /// <param name="chillerNumber">冷凍機台数</param>
    /// <param name="coolingTowerNumber">冷却塔台数（1台の冷凍機あたりの台数）</param>
    public DirectFiredAbsorptionChillerSystem
      (DirectFiredAbsorptionChiller chiller, CentrifugalPump chwPump, CentrifugalPump hwPump, 
      CentrifugalPump cdwPump, CoolingTower cTower, int chillerNumber, int coolingTowerNumber)
    {
      this.chiller = chiller;
      this.chwPump = chwPump;
      this.hwPump = hwPump;
      this.cdwPump = cdwPump;
      this.cTower = cTower;
      this.ChillerNumber = chillerNumber;
      this.CoolingTowerNumber = coolingTowerNumber;

      //加熱運転・冷却運転対応
      SelectableMode = HeatSourceSystemModel.OperatingMode.Cooling | HeatSourceSystemModel.OperatingMode.Heating;
      Mode = HeatSourceSystemModel.OperatingMode.ShutOff;      
    }

    #endregion

  }
}
