﻿/* GroundHeatSourceHeatPumpSystem.cs
 * 
 * Copyright (C) 2018 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;

using Popolo.HVAC.HeatSource;
using Popolo.HVAC.Circuit;
using Popolo.ThermophysicalProperty;
using Popolo.HVAC.HeatExchanger;
using Popolo.Numerics;

namespace Popolo.HVAC.SubSystem
{
  /// <summary>地中熱ヒートポンプシステム</summary>
  [Serializable]
  public class GroundHeatSourceHeatPumpSystem : IHeatSourceSubSystem
  {

    #region 定数宣言

    /// <summary>加熱運転の熱源水下限温度[C]</summary>
    private const double HEAT_MIN_TEMP = 0;

    /// <summary>冷却運転時の熱源水上限温度[C]</summary>
    private const double COOL_MAX_TEMP = 40;

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    #endregion

    #region IHeatSourceSubSystem実装（プロパティ）

    /// <summary>運転可能モードを取得する</summary>
    public HeatSourceSystemModel.OperatingMode SelectableMode
    { get { return HeatSourceSystemModel.OperatingMode.Cooling | HeatSourceSystemModel.OperatingMode.Heating; } }

    /// <summary>運転モードを設定・取得する</summary>
    public HeatSourceSystemModel.OperatingMode Mode { get; set; }

    /// <summary>外気条件を設定・取得する</summary>
    public ImmutableMoistAir OutdoorAir { get; set; }

    /// <summary>現在の日時を設定・取得する</summary>
    public DateTime CurrentDateTime { get; set; }

    /// <summary>タイムステップ[sec]を設定・取得する</summary>
    public double TimeStep
    {
      get { return gHex.TimeStep; }
      set { gHex.TimeStep = value; }
    }

    /// <summary>冷水供給が過負荷か否か</summary>
    public bool IsOverLoad_C { get; private set; }

    /// <summary>温水供給が過負荷か否か</summary>
    public bool IsOverLoad_H { get; private set; }

    /// <summary>温水還温度[C]を設定・取得する</summary>
    public double HotWaterReturnTemperature { get; set; }

    /// <summary>温水往温度設定値[C]を設定・取得する</summary>
    public double HotWaterSupplyTemperatureSetpoint
    {
      get { return whp.HotWaterSetPoint; }
      set { whp.HotWaterSetPoint = value; }
    }

    /// <summary>温水往温度[C]を取得する</summary>
    public double HotWaterSupplyTemperature { get; private set; }

    /// <summary>温水流量[kg/s]を取得する</summary>
    public double HotWaterFlowRate { get; private set; }

    /// <summary>温水上限流量[kg/s]を取得する</summary>
    public double MaxHotWaterFlowRate { get { return whp.NominalHotWaterFlowRate; } }

    /// <summary>温水下限流量比[-]を取得する</summary>
    public double MinHotWaterFlowRatio { get { return 0.5; } }

    /// <summary>冷水還温度[C]を設定・取得する</summary>
    public double ChilledWaterReturnTemperature { get; set; }

    /// <summary>冷水往温度設定値[C]を設定・取得する</summary>
    public double ChilledWaterSupplyTemperatureSetpoint
    {
      get { return whp.ChilledWaterSetPoint; }
      set { whp.ChilledWaterSetPoint = value; }
    }

    /// <summary>冷水往温度[C]を取得する</summary>
    public double ChilledWaterSupplyTemperature { get; private set; }

    /// <summary>冷水流量[kg/s]を設定する</summary>
    public double ChilledWaterFlowRate { get; private set; }

    /// <summary>冷水上限流量[kg/s]を取得する</summary>
    public double MaxChilledWaterFlowRate { get { return whp.NominalChilledWaterFlowRate; } }

    /// <summary>冷水下限流量比[-]を取得する</summary>
    public double MinChilledWaterFlowRatio { get { return 0.5; } }

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>過負荷か否かを取得する</summary>
    public bool IsGroundHEX_OverLoad { get; private set; }

    /// <summary>水熱源ヒートポンプ</summary>
    private WaterHeatPump whp;

    /// <summary>地中熱交換器</summary>
    private SimpleGroundHeatExchanger gHex;

    /// <summary>ポンプ</summary>
    private CentrifugalPump groundWaterPump, supplyPump;

    /// <summary>水熱源ヒートポンプを取得する</summary>
    public ImmutableWaterHeatPump WaterHeatPump { get { return whp; } }

    /// <summary>地中熱交換器を取得する</summary>
    public ImmutableSimpleGroundHeatExchanger GroundHeatExchanger { get { return gHex; } }

    /// <summary>熱源水ポンプを取得する</summary>
    public ImmutableCentrifugalPump GroundWaterPump { get { return groundWaterPump; } }

    /// <summary>冷温水ポンプを取得する</summary>
    public ImmutableCentrifugalPump SupplyPump { get { return supplyPump; } }

    /// <summary>熱源水流量[kg/s]を取得する</summary>
    public double GroundWaterFlowRate { get; private set; }

    /// <summary>熱源水流量設定値[kg/s]を取得する</summary>
    public double GroundWaterFlowRateSetpoint
    {
      get { return grndFlowSP; }
      set { if (0 < value) grndFlowSP = value; }
    }

    /// <summary>熱源水流量設定値[kg/s]</summary>
    private double grndFlowSP;

    #endregion

    #region IHeatSourceSubSystem実装（メソッド）

    /// <summary>冷温水往温度を予測する</summary>
    /// <param name="chilledWaterFlowRate">冷水流量[kg/s]</param>
    /// <param name="hotWaterFlowRate">温水流量[kg/s]</param>
    public void ForecastSupplyWaterTemperature(double chilledWaterFlowRate, double hotWaterFlowRate)
    {
      ChilledWaterFlowRate = chilledWaterFlowRate;
      HotWaterFlowRate = hotWaterFlowRate;

      if (Mode == HeatSourceSystemModel.OperatingMode.Cooling)
        forecastCooling(chilledWaterFlowRate);
      else if (Mode == HeatSourceSystemModel.OperatingMode.Heating)
        forecastHeating(hotWaterFlowRate);
      else
      {
        gHex.Update(0, 0);  //土壌熱流のみを計算
        ShutOff();
      }
    }

    /// <summary>将来状態を予測する（冷房）</summary>
    /// <param name="chilledWaterFlowRate">冷水流量[kg/s]</param>
    private void forecastCooling(double chilledWaterFlowRate)
    {
      if (chilledWaterFlowRate <= 0 || ChilledWaterReturnTemperature < ChilledWaterSupplyTemperatureSetpoint)
      {
        gHex.Update(0, 0);  //土壌熱流のみを計算
        ShutOff();
        return;
      }

      //冷温水ポンプの計算
      supplyPump.UpdateState(0.001 * chilledWaterFlowRate);
      double deltaTW = supplyPump.GetElectricConsumption() / (WATER_SPECIFIC_HEAT * chilledWaterFlowRate);

      //熱源水ポンプの計算
      GroundWaterFlowRate = GroundWaterFlowRateSetpoint;
      groundWaterPump.UpdateState(0.001 * GroundWaterFlowRate);  //ポンプ状態を更新:比重量はブラインでも同じで良いのか？？？
      double deltaTG = groundWaterPump.GetElectricConsumption() / (gHex.FluidSpecificHeat * GroundWaterFlowRate);

      //熱源水上限温度で計算
      whp.CoolWater(chilledWaterFlowRate, GroundWaterFlowRate, ChilledWaterReturnTemperature + deltaTW, COOL_MAX_TEMP);
      gHex.ForecastSate(whp.CoolingWaterOutletTemperature + deltaTG, GroundWaterFlowRate);

      //過負荷の場合は下限温度で計算を打ち切り
      if (COOL_MAX_TEMP < gHex.FluidOutletTemperature) IsGroundHEX_OverLoad = true;
      //熱源水温度を収束計算
      else
      {
        //黄金分割法
        Roots.ErrorFunction eFnc = delegate (double x)
        {
          whp.CoolWater(chilledWaterFlowRate, GroundWaterFlowRate, ChilledWaterReturnTemperature + deltaTW, x);
          gHex.ForecastSate(whp.CoolingWaterOutletTemperature + deltaTG, GroundWaterFlowRate);
          return gHex.FluidOutletTemperature - x;
        };
        Roots.Brent(gHex.NearGroundTemperature, COOL_MAX_TEMP, 0.01, eFnc);

        IsGroundHEX_OverLoad = false;
      }

      HotWaterSupplyTemperature = HotWaterReturnTemperature;
      IsOverLoad_H = false;

      IsOverLoad_C = whp.IsOverLoad;
      if (IsOverLoad_C) ChilledWaterSupplyTemperature = whp.ChilledWaterOutletTemperature;
      else ChilledWaterSupplyTemperature = ChilledWaterSupplyTemperatureSetpoint;
    }

    /// <summary>将来状態を予測する（暖房）</summary>
    /// <param name="hotWaterFlowRate">温水流量[kg/s]</param>
    private void forecastHeating(double hotWaterFlowRate)
    {
      if (hotWaterFlowRate <= 0 || HotWaterSupplyTemperatureSetpoint < HotWaterReturnTemperature)
      {
        gHex.Update(0, 0);  //土壌熱流のみを計算
        ShutOff();
        return;
      }

      //冷温水ポンプの計算
      supplyPump.UpdateState(0.001 * hotWaterFlowRate);
      double deltaTW = supplyPump.GetElectricConsumption() / (WATER_SPECIFIC_HEAT * hotWaterFlowRate);

      //熱源水ポンプの計算
      GroundWaterFlowRate = GroundWaterFlowRateSetpoint;
      groundWaterPump.UpdateState(0.001 * GroundWaterFlowRate);  //ポンプ状態を更新:比重量はブラインでも同じで良いのか？？？
      double deltaTG = groundWaterPump.GetElectricConsumption() / (gHex.FluidSpecificHeat * GroundWaterFlowRate);

      //熱源水下限温度で計算
      whp.HeatWater(hotWaterFlowRate, GroundWaterFlowRate, HotWaterReturnTemperature + deltaTW, HEAT_MIN_TEMP);
      gHex.ForecastSate(whp.HeatSourceWaterOutletTemperature + deltaTG, GroundWaterFlowRate);

      //過負荷の場合は下限温度で計算を打ち切り
      if (gHex.FluidOutletTemperature < HEAT_MIN_TEMP) IsGroundHEX_OverLoad = true;
      //熱源水温度を収束計算
      else
      {
        //黄金分割法
        Roots.ErrorFunction eFnc = delegate (double x) 
        {
          whp.HeatWater(hotWaterFlowRate, GroundWaterFlowRate, HotWaterReturnTemperature + deltaTW, x);
          gHex.ForecastSate(whp.HeatSourceWaterOutletTemperature + deltaTG, GroundWaterFlowRate);
          return gHex.FluidOutletTemperature - x;
        };
        Roots.Brent(HEAT_MIN_TEMP, gHex.NearGroundTemperature, 0.01, eFnc);

        IsGroundHEX_OverLoad = false;
      }

      ChilledWaterSupplyTemperature = ChilledWaterReturnTemperature;
      IsOverLoad_C = false;

      IsOverLoad_H = whp.IsOverLoad;
      if (IsOverLoad_H) HotWaterSupplyTemperature = whp.HotWaterOutletTemperature;
      else HotWaterSupplyTemperature = HotWaterSupplyTemperatureSetpoint;
    }

    /// <summary>状態を確定する</summary>
    public void FixState()
    { gHex.FixState(); }

    /// <summary>熱源サブシステムを停止させる</summary>
    public void ShutOff()
    {
      IsOverLoad_C = IsOverLoad_H = false;
      GroundWaterFlowRate = 0;
      groundWaterPump.ShutOff();
      supplyPump.ShutOff();
      whp.ShutOff();
    }

    #endregion

    #region コンストラクタ

    /// <summary></summary>
    /// <param name="whp"></param>
    /// <param name="gHex"></param>
    /// <param name="groundWaterPump"></param>
    /// <param name="supplyPump"></param>
    public GroundHeatSourceHeatPumpSystem
      (WaterHeatPump whp, SimpleGroundHeatExchanger gHex, CentrifugalPump groundWaterPump, CentrifugalPump supplyPump)
    {
      this.whp = whp;
      this.gHex = gHex;
      this.groundWaterPump = groundWaterPump;
      this.supplyPump = supplyPump;

      GroundWaterFlowRateSetpoint = 1000 * groundWaterPump.DesignFlowRate;
    }

    #endregion

  }
}
