﻿/* HeatSourceSystemModel.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.HVAC.Circuit;
using Popolo.ThermophysicalProperty;

namespace Popolo.HVAC.SubSystem
{
  /// <summary>熱源一次システム</summary>
  [Serializable]
  public class HeatSourceSystemModel : ImmutableHeatSourceSystemModel
  {

    #region 定数宣言

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    #endregion

    #region 列挙型定義

    /// <summary>運転モード</summary>
    [Flags]
    public enum OperatingMode
    {
      /// <summary>停止</summary>
      ShutOff = 0,
      /// <summary>加熱運転</summary>
      Heating = 1,
      /// <summary>冷却運転</summary>
      Cooling = 2,
      /// <summary>加熱・冷却運転</summary>
      HeatingAndCooling = 4
    }

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>サブシステム</summary>
    private IHeatSourceSubSystem[] subSystems;

    /// <summary>2次ポンプ</summary>
    private PumpSystem chilledWaterPumps, hotWaterPumps;

    /// <summary>運転順位</summary>
    private int[] opRankC, opRankH;

    /// <summary>冷水往温度設定値を設定・取得する</summary>
    public double ChilledWaterSupplyTemperatureSetpoint { get; set; } = 7;

    /// <summary>冷水往温度[C]を取得する</summary>
    public double ChilledWaterSupplyTemperature { get; private set; } = 7;

    /// <summary>冷水還温度[C]を取得する</summary>
    public double ChilledWaterReturnTemperature { get; private set; } = 12;

    /// <summary>冷水流量[kg/s]を取得する</summary>
    public double ChilledWaterFlowRate { get; private set; }

    /// <summary>冷水バイパス流量[kg/s]を取得する</summary>
    public double ChilledWaterBypassFlowRate { get; private set; }

    /// <summary>冷水供給が過負荷か否か</summary>
    public bool IsOverLoad_C { get; private set; }

    /// <summary>温水往温度設定値を設定・取得する</summary>
    public double HotWaterSupplyTemperatureSetpoint { get; set; } = 45;

    /// <summary>温水往温度[C]を取得する</summary>
    public double HotWaterSupplyTemperature { get; private set; } = 45;

    /// <summary>温水還温度[C]を取得する</summary>
    public double HotWaterReturnTemperature { get; private set; } = 40;

    /// <summary>温水流量[kg/s]を取得する</summary>
    public double HotWaterFlowRate { get; private set; }

    /// <summary>温水バイパス流量[kg/s]を取得する</summary>
    public double HotWaterBypassFlowRate { get; private set; }

    /// <summary>温水供給が過負荷か否か</summary>
    public bool IsOverLoad_H { get; private set; }

    /// <summary>2次ポンプ方式か否か</summary>
    public bool IsSecondaryPumpSystem { get { return chilledWaterPumps != null; } }

    /// <summary>冷水二次ポンプシステムを取得する</summary>
    public ImmutablePumpSystem ChilledWaterPumpSystem { get { return chilledWaterPumps; } }

    /// <summary>温水二次ポンプシステムを取得する</summary>
    public ImmutablePumpSystem HotWaterPumpSystem { get { return hotWaterPumps; } }

    /// <summary>外気条件を設定・取得する</summary>
    public ImmutableMoistAir OutdoorAir { get; set; } = new MoistAir(35, 0.0195);

    /// <summary>現在の日時を設定・取得する</summary>
    public DateTime CurrentDateTime { get; set; }

    /// <summary>計算時間間隔[sec]を設定・取得する</summary>
    public double TimeStep { get; set; } = 3600;

    /// <summary>配管熱損失率[-]を設定・取得する</summary>
    public double PipeHeatLossRate { get; set; } = 0.08;

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する（2ポンプシステム）</summary>
    /// <param name="subSystems">熱源サブシステムリスト</param>
    /// <param name="chilledWaterPumps">冷水二次ポンプ</param>
    /// <param name="hotWaterPumps">温水二次ポンプ</param>
    public HeatSourceSystemModel
      (IHeatSourceSubSystem[] subSystems, PumpSystem chilledWaterPumps, PumpSystem hotWaterPumps)
    {
      this.subSystems = subSystems;
      this.chilledWaterPumps = chilledWaterPumps;
      this.hotWaterPumps = hotWaterPumps;
      opRankC = new int[subSystems.Length];
      opRankH = new int[subSystems.Length];
    }

    /// <summary>インスタンスを初期化する（1ポンプシステム）</summary>
    /// <param name="subSystems">熱源サブシステムリスト</param>
    public HeatSourceSystemModel(IHeatSourceSubSystem[] subSystems)
    {
      this.subSystems = subSystems;
      opRankC = new int[subSystems.Length];
      opRankH = new int[subSystems.Length];
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>将来の状態を予測する</summary>
    /// <param name="chilledWaterFlowRate">冷水流量[kg/s]</param>
    /// <param name="chilledWaterReturnTemperature">冷水還温度[C]</param>
    /// <param name="hotWaterFlowRate">温水流量[kg/s]</param>
    /// <param name="hotWaterReturnTemperature">温水還温度[C]</param>
    public void ForecastSupplyWaterTemperature
      (double chilledWaterFlowRate, double chilledWaterReturnTemperature,
      double hotWaterFlowRate, double hotWaterReturnTemperature)
    {
      ChilledWaterFlowRate = chilledWaterFlowRate;
      HotWaterFlowRate = hotWaterFlowRate;
      ChilledWaterReturnTemperature = chilledWaterReturnTemperature
        - (ChilledWaterSupplyTemperatureSetpoint - chilledWaterReturnTemperature) * PipeHeatLossRate;
      HotWaterReturnTemperature = hotWaterReturnTemperature
        - (HotWaterSupplyTemperatureSetpoint - hotWaterReturnTemperature) * PipeHeatLossRate;

      //日時・冷温水往温度・外気条件を設定
      for (int i = 0; i < subSystems.Length; i++)
      {
        subSystems[i].TimeStep = TimeStep;
        subSystems[i].CurrentDateTime = CurrentDateTime;
        subSystems[i].OutdoorAir = OutdoorAir;
        subSystems[i].ChilledWaterSupplyTemperatureSetpoint = ChilledWaterSupplyTemperatureSetpoint;
        subSystems[i].HotWaterSupplyTemperatureSetpoint = HotWaterSupplyTemperatureSetpoint;
      }

      //2次ポンプによる昇温を反映
      double tcwi = chilledWaterReturnTemperature;
      double thwi = hotWaterReturnTemperature;
      if (IsSecondaryPumpSystem)
      {
        if (chilledWaterFlowRate != 0)
        {
          chilledWaterPumps.TotalFlowRate = 0.001 * chilledWaterFlowRate;
          chilledWaterPumps.UpdateState();
          tcwi += chilledWaterPumps.GetElectricConsumption() / (chilledWaterFlowRate * WATER_SPECIFIC_HEAT);
        }
        else chilledWaterPumps.ShutOff();
        if (hotWaterFlowRate != 0)
        {
          hotWaterPumps.TotalFlowRate = 0.001 * hotWaterFlowRate;
          hotWaterPumps.UpdateState();
          thwi += hotWaterPumps.GetElectricConsumption() / (hotWaterFlowRate * WATER_SPECIFIC_HEAT);
        }
        else hotWaterPumps.ShutOff();
      }

      //負荷流量に応じて最低運転台数を計算
      int cStage, hStage;
      double fcsum = 0; //冷却運転
      if (chilledWaterFlowRate <= 0) cStage = 0;
      else
      {
        for (cStage = 1; cStage <= opRankC.Length; cStage++)
        {
          for (int j = 0; j < subSystems.Length; j++)
            if (opRankC[j] == cStage && isModeAdapts(j, true)) fcsum += subSystems[j].MaxChilledWaterFlowRate;
          if (chilledWaterFlowRate < fcsum) break;
        }
      }
      double fhsum = 0; //加熱運転
      if (hotWaterFlowRate <= 0) hStage = 0;
      else
      {
        for (hStage = 1; hStage <= opRankH.Length; hStage++)
        {
          for (int j = 0; j < subSystems.Length; j++)
            if (opRankH[j] == hStage && isModeAdapts(j, false)) fhsum += subSystems[j].MaxHotWaterFlowRate;
          if (hotWaterFlowRate < fhsum) break;
        }
      }

      //過負荷系統が無くなるまで増段
      double bypsC = 0;
      double bypsH = 0;
      while (true)
      {
        IsOverLoad_C = false;
        IsOverLoad_H = false;

        //最大絞り流量比を計算
        double minMaxC = 0.0;
        double minMaxH = 0.0;
        for (int i = 0; i < subSystems.Length; i++)
        {
          if (opRankC[i] <= cStage && isModeAdapts(i, true))
            minMaxC = Math.Max(minMaxC, subSystems[i].MinChilledWaterFlowRatio);
          if (opRankH[i] <= hStage && isModeAdapts(i, false))
            minMaxH = Math.Max(minMaxH, subSystems[i].MinHotWaterFlowRatio);
        }

        //バイパス流量を考慮して熱源入口水温を計算
        double cpl, hpl;
        if (chilledWaterFlowRate == 0 || fcsum == 0) cpl = 0;
        else
        {
          cpl = chilledWaterFlowRate / fcsum;
          bypsC = Math.Max(0, minMaxC - cpl);
          double tcwi2 = (tcwi * cpl + ChilledWaterSupplyTemperatureSetpoint * bypsC) / (cpl + bypsC);
          foreach (IHeatSourceSubSystem hss in subSystems) hss.ChilledWaterReturnTemperature = tcwi2;
          cpl = Math.Max(minMaxC, cpl);
        }
        if (hotWaterFlowRate == 0 || fhsum == 0) hpl = 0;
        else
        {
          hpl = hotWaterFlowRate / fhsum;
          bypsH = Math.Max(0, minMaxH - hpl);
          double thwi2 = (thwi * hpl + HotWaterSupplyTemperatureSetpoint * bypsH) / (hpl + bypsH);
          foreach (IHeatSourceSubSystem hss in subSystems) hss.HotWaterReturnTemperature = thwi2;
          hpl = Math.Max(minMaxH, hpl);
        }

        //過負荷系統が無いか確認
        for (int i = 0; i < subSystems.Length; i++)
        {
          double cf, hf;
          cf = hf = 0;
          if (opRankC[i] <= cStage && isModeAdapts(i, true)) cf = subSystems[i].MaxChilledWaterFlowRate * cpl;
          if (opRankH[i] <= hStage && isModeAdapts(i, false)) hf = subSystems[i].MaxHotWaterFlowRate * hpl;
          subSystems[i].ForecastSupplyWaterTemperature(cf, hf);

          if (subSystems[i].IsOverLoad_C) IsOverLoad_C = true;
          if (subSystems[i].IsOverLoad_H) IsOverLoad_H = true;
        }
        int maxStage = opRankC.Length;
        if ((!IsOverLoad_C || maxStage <= cStage) && (!IsOverLoad_H || maxStage <= hStage)) break;

        //増段処理
        if (IsOverLoad_C && cStage != maxStage)
        {
          cStage++;
          for (int j = 0; j < subSystems.Length; j++)
            if (opRankC[j] == cStage && isModeAdapts(j, true)) fcsum += subSystems[j].MaxChilledWaterFlowRate;
        }
        if (IsOverLoad_H && hStage != maxStage)
        {
          hStage++;
          for (int j = 0; j < subSystems.Length; j++)
            if (opRankH[j] == hStage && isModeAdapts(j, false)) fhsum += subSystems[j].MaxHotWaterFlowRate;
        }
      }
      ChilledWaterBypassFlowRate = bypsC * fcsum;
      HotWaterBypassFlowRate = bypsH * fhsum;

      //冷水出口温度を計算
      if (IsOverLoad_C)
      {
        ChilledWaterSupplyTemperature = 0;
        double cwSum = 0;
        foreach (IHeatSourceSubSystem ss in subSystems)
        {
          ChilledWaterSupplyTemperature += ss.ChilledWaterSupplyTemperature * ss.ChilledWaterFlowRate;
          cwSum += ss.ChilledWaterFlowRate;
        }
        ChilledWaterSupplyTemperature /= cwSum;
      }
      else ChilledWaterSupplyTemperature = ChilledWaterSupplyTemperatureSetpoint;

      //温水出口温度を計算
      if (IsOverLoad_H)
      {
        HotWaterSupplyTemperature = 0;
        double hwSum = 0;
        foreach (IHeatSourceSubSystem ss in subSystems)
        {
          HotWaterSupplyTemperature += ss.HotWaterSupplyTemperature * ss.HotWaterFlowRate;
          hwSum += ss.HotWaterFlowRate;
        }
        HotWaterSupplyTemperature /= hwSum;
      }
      else HotWaterSupplyTemperature = HotWaterSupplyTemperatureSetpoint;
    }

    /// <summary>状態を確定する</summary>
    public void FixState() { foreach (IHeatSourceSubSystem hs in subSystems) hs.FixState(); }

    /// <summary>冷却運転の段数を設定する</summary>
    /// <param name="subSystemIndex">サブシステム番号</param>
    /// <param name="stage">段数（1, 2, 3...）</param>
    public void SetChillingOperationSequence(int subSystemIndex, int stage) { opRankC[subSystemIndex] = stage; }

    /// <summary>暖房運転の段数を設定する</summary>
    /// <param name="subSystemIndex">サブシステム番号</param>
    /// <param name="stage">段数（1, 2, 3...）</param>
    public void SetHeatingOperationSequence(int subSystemIndex, int stage) { opRankH[subSystemIndex] = stage; }

    /// <summary>運転モードを設定する</summary>
    /// <param name="subSystemIndex">サブシステム番号</param>
    /// <param name="mode">運転モード</param>
    public void SetOperatingMode(int subSystemIndex, OperatingMode mode) { subSystems[subSystemIndex].Mode = mode; }

    /// <summary>運転モードに合致するか否か</summary>
    /// <param name="subsystemNumber">サブシステム番号</param>
    /// <param name="cooling">冷却運転か否か</param>
    /// <returns>運転モードに合致するか否か</returns>
    private bool isModeAdapts(int subsystemNumber, bool cooling)
    {
      IHeatSourceSubSystem ss = subSystems[subsystemNumber];
      if (ss.Mode == OperatingMode.ShutOff) return false;
      if (cooling && ss.Mode == OperatingMode.Heating) return false;
      if (!cooling && ss.Mode == OperatingMode.Cooling) return false;
      return true;
    }

    #endregion

  }

  /// <summary>読み取り専用の熱源一次システム</summary>
  public interface ImmutableHeatSourceSystemModel
  {
    /// <summary>冷水往温度設定値を取得する</summary>
    double ChilledWaterSupplyTemperatureSetpoint { get; }

    /// <summary>冷水往温度[C]を取得する</summary>
    double ChilledWaterSupplyTemperature { get; }

    /// <summary>冷水還温度[C]を取得する</summary>
    double ChilledWaterReturnTemperature { get; }

    /// <summary>冷水流量[kg/s]を取得する</summary>
    double ChilledWaterFlowRate { get; }

    /// <summary>冷水バイパス流量[kg/s]を取得する</summary>
    double ChilledWaterBypassFlowRate { get; }

    /// <summary>冷水供給が過負荷か否か</summary>
    bool IsOverLoad_C { get; }

    /// <summary>温水往温度設定値を設定・取得する</summary>
    double HotWaterSupplyTemperatureSetpoint { get; }

    /// <summary>温水往温度[C]を取得する</summary>
    double HotWaterSupplyTemperature { get; }

    /// <summary>温水還温度[C]を取得する</summary>
    double HotWaterReturnTemperature { get; }

    /// <summary>温水流量[kg/s]を取得する</summary>
    double HotWaterFlowRate { get; }

    /// <summary>温水バイパス流量[kg/s]を取得する</summary>
    double HotWaterBypassFlowRate { get; }

    /// <summary>温水供給が過負荷か否か</summary>
    bool IsOverLoad_H { get; }

    /// <summary>2次ポンプ方式か否か</summary>
    bool IsSecondaryPumpSystem { get; }

    /// <summary>冷水二次ポンプシステムを取得する</summary>
    ImmutablePumpSystem ChilledWaterPumpSystem { get; }

    /// <summary>温水二次ポンプシステムを取得する</summary>
    ImmutablePumpSystem HotWaterPumpSystem { get; }

    /// <summary>外気条件を設定・取得する</summary>
    ImmutableMoistAir OutdoorAir { get; }

    /// <summary>現在の日時を設定・取得する</summary>
    DateTime CurrentDateTime { get; }

    /// <summary>計算時間間隔[sec]を設定・取得する</summary>
    double TimeStep { get; }

    /// <summary>配管熱損失率[-]を設定・取得する</summary>
    double PipeHeatLossRate { get; }
  }

}
