﻿/* HotWaterBoilerSystem.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.HVAC.HeatSource;
using Popolo.HVAC.Circuit;
using Popolo.ThermophysicalProperty;

namespace Popolo.HVAC.SubSystem
{
  /// <summary>温水ボイラによる熱源サブシステム</summary>
  [Serializable]
  public class HotWaterBoilerSystem : IHeatSourceSubSystem
  {

    #region 定数宣言

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>温水ボイラ</summary>
    private HotWaterBoiler boiler;

    /// <summary>温水ポンプ</summary>
    private CentrifugalPump hwPump;

    /// <summary>温水ボイラを取得する</summary>
    public ImmutableHotWaterBoiler Boiler { get { return boiler; } }

    /// <summary>温水ポンプを取得する</summary>
    public ImmutableCentrifugalPump HotWaterPump { get { return hwPump; } }

    /// <summary>ボイラの台数を取得する</summary>
    public int BoilerNumber { get; private set; }

    /// <summary>運転台数を取得する</summary>
    public int OperatingNumber { get; private set; }

    #endregion

    #region IHeatSourceSubSystem実装

    /// <summary>冷水供給が過負荷か否か</summary>
    public bool IsOverLoad_C { get { return false; } }

    /// <summary>温水供給が過負荷か否か</summary>
    public bool IsOverLoad_H { get; private set; }

    /// <summary>運転可能モードを取得する</summary>
    public HeatSourceSystemModel.OperatingMode SelectableMode
    { get { return HeatSourceSystemModel.OperatingMode.Heating; } }

    /// <summary>運転モードを設定・取得する</summary>
    public HeatSourceSystemModel.OperatingMode Mode { get; set; }

    /// <summary>現在の日時を設定・取得する</summary>
    public DateTime CurrentDateTime { get; set; }

    /// <summary>タイムステップを設定・取得する</summary>
    public double TimeStep { get; set; }

    /// <summary>温水還温度[C]を設定・取得する</summary>
    public double HotWaterReturnTemperature { get; set; } = 40;

    /// <summary>温水往温度設定値[C]を設定・取得する</summary>
    public double HotWaterSupplyTemperatureSetpoint { get; set; } = 45;

    /// <summary>温水往温度[C]を取得する</summary>
    public double HotWaterSupplyTemperature { get; private set; } = 45;

    /// <summary>温水流量[kg/s]を設定・取得する</summary>
    public double HotWaterFlowRate { get; set; }

    /// <summary>温水上限流量[kg/s]を取得する</summary>
    public double MaxHotWaterFlowRate
    { get { return boiler.MaxWaterFlowRate * BoilerNumber; } }

    /// <summary>温水下限流量比[-]を取得する</summary>
    public double MinHotWaterFlowRatio
    { get { return boiler.MinWaterFlowRatio / BoilerNumber; } }

    /// <summary>冷水還温度[C]を設定・取得する</summary>
    public double ChilledWaterReturnTemperature { get; set; }

    /// <summary>冷水往温度設定値[C]を設定・取得する</summary>
    public double ChilledWaterSupplyTemperatureSetpoint { get; set; }

    /// <summary>冷水往温度[C]を取得する</summary>
    public double ChilledWaterSupplyTemperature { get; private set; }

    /// <summary>冷水流量[kg/s]を取得する</summary>
    public double ChilledWaterFlowRate { get; private set; }

    /// <summary>冷水上限流量[kg/s]を取得する</summary>
    public double MaxChilledWaterFlowRate { get { return 0; } }

    /// <summary>冷水下限流量比[-]を取得する</summary>
    public double MinChilledWaterFlowRatio { get { return 0; } }

    /// <summary>外気条件を設定・取得する</summary>
    public ImmutableMoistAir OutdoorAir { get; set; }

    /// <summary>熱源サブシステムを停止させる</summary>
    public void ShutOff()
    {
      IsOverLoad_H = false;
      BoilerNumber = 0;
      boiler.ShutOff();
      hwPump.ShutOff();
    }

    /// <summary>冷温水往温度を予測する</summary>
    /// <param name="chilledWaterFlowRate">冷水流量[kg/s]</param>
    /// <param name="hotWaterFlowRate">温水流量[kg/s]</param>
    public void ForecastSupplyWaterTemperature(double chilledWaterFlowRate, double hotWaterFlowRate)
    {
      ChilledWaterFlowRate = chilledWaterFlowRate;

      if (Mode != HeatSourceSystemModel.OperatingMode.Heating || hotWaterFlowRate == 0)
      {
        ShutOff();
        return;
      }

      OperatingNumber = (int)Math.Ceiling(hotWaterFlowRate / boiler.MaxWaterFlowRate);
      OperatingNumber = Math.Min(BoilerNumber, OperatingNumber);
      boiler.AmbientTemperature = OutdoorAir.DrybulbTemperature;
      boiler.OutletWaterSetPointTemperature = HotWaterSupplyTemperatureSetpoint;

      while (true)
      {
        //ポンプによる昇温を評価
        double hwFlow = hotWaterFlowRate / OperatingNumber;
        hwPump.UpdateState(0.001 * hwFlow);
        double twi = HotWaterReturnTemperature + hwPump.GetElectricConsumption() / (WATER_SPECIFIC_HEAT * hwFlow);
        boiler.Update(twi, hwFlow);
        if (OperatingNumber == BoilerNumber || !boiler.IsOverLoad) break;
        else OperatingNumber++;
      }

      IsOverLoad_H = boiler.IsOverLoad;
      if (IsOverLoad_H) HotWaterSupplyTemperature = boiler.OutletWaterTemperature;
      else HotWaterSupplyTemperature = HotWaterSupplyTemperatureSetpoint;
    }

    /// <summary>状態を確定する</summary>
    public void FixState() { }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="boiler">温水ボイラ</param>
    /// <param name="hwPump">温水ポンプ</param>
    /// <param name="number">ボイラ台数</param>
    public HotWaterBoilerSystem(HotWaterBoiler boiler, CentrifugalPump hwPump, int number)
    {
      this.boiler = boiler;
      this.hwPump = hwPump;
      BoilerNumber = number;
      Mode = HeatSourceSystemModel.OperatingMode.ShutOff;
    }

    #endregion

  }
}
