﻿/* MultipleStratifiedWaterTankSystem.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.HVAC.HeatExchanger;
using Popolo.HVAC.HeatSource;
using Popolo.HVAC.Circuit;
using Popolo.ThermophysicalProperty;
using Popolo.Numerics;
using Popolo.HVAC.ThermalStorage;

namespace Popolo.HVAC.SubSystem
{
  /// <summary>温度成層型蓄熱槽による熱源サブシステム</summary>
  [Serializable]
  public class MultipleStratifiedWaterTankSystem : IHeatSourceSubSystem
  {

    #region 定数宣言

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>予測計算中か否か</summary>
    private bool isForecasting = false;

    /// <summary>水槽内温度分布</summary>
    private double[] oldTemps;

    /// <summary>冷凍機ポンプによる昇温[K]</summary>
    private double dtChillingPump, dtCoolingPump;

    /// <summary>ターボ冷凍機</summary>
    private ICentrifugalChiller chiller;

    /// <summary>ポンプ</summary>
    private CentrifugalPump cdwPump, chwPump, chgPump, disPump;

    /// <summary>冷却塔</summary>
    private CoolingTower cTower;

    /// <summary>プレート熱交換器</summary>
    private PlateHeatExchanger pHex;

    /// <summary>温度成層型蓄熱槽</summary>
    private MultipleStratifiedWaterTank wTank;

    /// <summary>ターボ冷凍機を取得する</summary>
    public ImmutableCentrifugalChiller Chiller { get { return chiller; } }

    /// <summary>冷水ポンプを取得する</summary>
    public ImmutableCentrifugalPump ChilledWaterPump { get { return chwPump; } }

    /// <summary>冷却水ポンプを取得する</summary>
    public ImmutableCentrifugalPump CoolingWaterPump { get { return cdwPump; } }

    /// <summary>蓄熱ポンプを取得する</summary>
    public ImmutableCentrifugalPump ChargePump { get { return chgPump; } }

    /// <summary>放熱ポンプを取得する</summary>
    public ImmutableCentrifugalPump DischargePump { get { return disPump; } }

    /// <summary>冷却塔を取得する</summary>
    public ImmutableCoolingTower CoolingTower { get { return cTower; } }

    /// <summary>プレート熱交換器を取得する</summary>
    public ImmutablePlateHeatExchanger PlateHeatExchanger { get { return pHex; } }

    /// <summary>温度成層型蓄熱槽</summary>
    public ImmutableMultipleStratifiedWaterTank WaterTank { get { return wTank; } }

    /// <summary>冷凍機の台数を取得する</summary>
    public int ChillerNumber { get; private set; }

    /// <summary>冷却塔のセル数（1台の冷凍機あたり）を取得する</summary>
    public int CoolingTowerNumber { get; private set; }

    /// <summary>冷却水温度を制御するか否か</summary>
    public bool ControlCoolingWaterTemperature { get; set; }

    /// <summary>冷却水温度設定値[C]を設定・取得する</summary>
    public double CoolingWaterTemperatureSetpoint { get; set; } = 32;

    /// <summary>冷凍機を運転させるか否か</summary>
    public bool OperateChiler
    {
      get { return chiller.IsOperating; }
      set { chiller.IsOperating = value; }
    }

    /// <summary>水蓄熱槽の状態更新タイムステップ[sec]を設定・取得する</summary>
    public double TankUpdateTimeStep { get; set; } = 600;

    /// <summary>蓄熱温度[C]を設定・取得する</summary>
    public double StorageTemperature
    {
      get { return chiller.ChilledWaterOutletSetPointTemperature; }
      set { chiller.ChilledWaterOutletSetPointTemperature = value; }
    }

    /// <summary>蓄熱運転中か否か</summary>
    public bool Charging
    {
      get { return chiller.IsOperating; }
      set { chiller.IsOperating = value; }
    }

    #endregion

    #region IHeatSourceSubSystem実装

    /// <summary>冷水供給が過負荷か否か</summary>
    public bool IsOverLoad_C { get; private set; }

    /// <summary>温水供給が過負荷か否か</summary>
    public bool IsOverLoad_H { get; private set; }

    /// <summary>運転可能モードを取得する</summary>
    public HeatSourceSystemModel.OperatingMode SelectableMode
    { get { return HeatSourceSystemModel.OperatingMode.Cooling; } }

    /// <summary>運転モードを設定・取得する</summary>
    public HeatSourceSystemModel.OperatingMode Mode { get; set; }

    /// <summary>現在の日時を設定・取得する</summary>
    public DateTime CurrentDateTime { get; set; }

    /// <summary>タイムステップを設定・取得する</summary>
    public double TimeStep { get; set; }

    /// <summary>温水還温度[C]を設定・取得する</summary>
    public double HotWaterReturnTemperature { get; set; } = 40;

    /// <summary>温水往温度設定値[C]を設定・取得する</summary>
    public double HotWaterSupplyTemperatureSetpoint { get; set; } = 45;

    /// <summary>温水往温度[C]を取得する</summary>
    public double HotWaterSupplyTemperature { get; private set; } = 45;

    /// <summary>温水流量[kg/s]を取得する</summary>
    public double HotWaterFlowRate { get; private set; }

    /// <summary>温水上限流量[kg/s]を取得する</summary>
    public double MaxHotWaterFlowRate { get { return 0; } }

    /// <summary>温水下限流量比[-]を取得する</summary>
    public double MinHotWaterFlowRatio { get { return 0; } }

    /// <summary>冷水還温度[C]を設定・取得する</summary>
    public double ChilledWaterReturnTemperature { get; set; } = 12;

    /// <summary>冷水往温度設定値[C]を設定・取得する</summary>
    public double ChilledWaterSupplyTemperatureSetpoint
    {
      get { return pHex.SupplyTemperatureSetpoint; }
      set { pHex.SupplyTemperatureSetpoint = value; }
    }

    /// <summary>冷水往温度[C]を取得する</summary>
    public double ChilledWaterSupplyTemperature { get; private set; } = 7;

    /// <summary>冷水流量[kg/s]を取得する</summary>
    public double ChilledWaterFlowRate { get; private set; }

    /// <summary>冷水上限流量[kg/s]を取得する</summary>
    public double MaxChilledWaterFlowRate { get { return pHex.MaxSupplyFlowRate; } }

    /// <summary>冷水下限流量比[-]を取得する</summary>
    public double MinChilledWaterFlowRatio { get { return 0; } }

    /// <summary>外気条件を設定・取得する</summary>
    public ImmutableMoistAir OutdoorAir { get; set; }

    /// <summary>熱源サブシステムを停止させる</summary>
    public void ShutOff()
    {
      IsOverLoad_C = IsOverLoad_H = false;
      chiller.ShutOff();
      chwPump.ShutOff();
      cdwPump.ShutOff();
      chgPump.ShutOff();
      disPump.ShutOff();
      cTower.ShutOff();
    }

    /// <summary>冷温水往温度を予測する</summary>
    /// <param name="chilledWaterFlowRate">冷水流量[kg/s]</param>
    /// <param name="hotWaterFlowRate">温水流量[kg/s]</param>
    public void ForecastSupplyWaterTemperature(double chilledWaterFlowRate, double hotWaterFlowRate)
    {
      ChilledWaterFlowRate = chilledWaterFlowRate;
      HotWaterFlowRate = hotWaterFlowRate;
      cTower.SetOutdoorAirState(OutdoorAir.WetbulbTemperature, OutdoorAir.HumidityRatio);

      //水槽内温度分布の一時保存と復元
      if (isForecasting) wTank.InitializeTemperature(oldTemps);
      else
      {
        isForecasting = true;
        wTank.GetTemperatures(ref oldTemps);
      }

      //タイムステップが経過するまで水槽温度更新を続ける
      double remTime = TimeStep;
      double aveTemp = 0;
      IsOverLoad_C = false;
      while (true)
      {
        bool isLastCalc = remTime < TankUpdateTimeStep;
        wTank.TimeStep = TankUpdateTimeStep;
        if (isLastCalc) wTank.TimeStep = remTime;
        remTime -= wTank.TimeStep;

        //冷凍機稼働かつ二次側負荷有りの場合はHEX所要流量を収束計算（追掛運転）
        if (Charging && 0 < ChilledWaterFlowRate)
        {
          //冷水ポンプの昇温幅を計算
          chgPump.UpdateState(chgPump.DesignFlowRate);
          chwPump.UpdateState(0.001 * ChilledWaterFlowRate);
          double dtCHP = chwPump.GetElectricConsumption() / (WATER_SPECIFIC_HEAT * ChilledWaterFlowRate);

          Roots.ErrorFunction eFnc = delegate (double hexFlow)
          {
            calcHexHeatTransfer(hexFlow, ChilledWaterReturnTemperature + dtCHP);
            return pHex.SupplyTemperature - ChilledWaterSupplyTemperatureSetpoint;
          };

          //最大流量で処理可能か
          if (eFnc(pHex.MaxHeatSourceFlowRate) < 0)
          {
            IsOverLoad_C = false;
            //最小流量で制御可能か
            if (0 < eFnc(pHex.MaxHeatSourceFlowRate * 0.001))
              Roots.Bisection(eFnc, pHex.MaxHeatSourceFlowRate * 0.001,
                pHex.MaxHeatSourceFlowRate, 0.01, pHex.MaxHeatSourceFlowRate * 0.01, 20);
          }
          else IsOverLoad_C = true;

          //水槽内温度更新処理
          double tTNKin;
          bool isDownFlow = chiller.ChilledWaterFlowRate < pHex.HeatSourceFlowRate;
          if (isDownFlow) tTNKin = pHex.HeatSourceOutletTemperature 
              + disPump.GetElectricConsumption() / (WATER_SPECIFIC_HEAT * pHex.HeatSourceFlowRate);
          else tTNKin = chiller.ChilledWaterOutletTemperature;
          wTank.UpdateState
            (tTNKin, 0.001 * Math.Abs(chiller.ChilledWaterFlowRate - pHex.HeatSourceFlowRate), isDownFlow);
        }
        //冷凍機非稼働+二次側負荷有り（放熱運転）
        else if (0 < ChilledWaterFlowRate)
        {
          chiller.ShutOff();
          chgPump.ShutOff();
          cdwPump.ShutOff();
          cTower.ShutOff();

          //冷水ポンプの昇温幅を計算
          chwPump.UpdateState(0.001 * ChilledWaterFlowRate);
          double dtCHP = chwPump.GetElectricConsumption() / (WATER_SPECIFIC_HEAT * ChilledWaterFlowRate);

          //放熱用プレート熱交換器の必要通水量を計算
          pHex.ControlSupplyTemperature
            (WaterTank.LowerOutletTemperarture, ChilledWaterReturnTemperature + dtCHP, ChilledWaterFlowRate);

          //水槽内温度を更新
          double tankInletTemp = 0;
          if (0 < pHex.HeatSourceFlowRate)
          {
            disPump.UpdateState(0.001 * pHex.HeatSourceFlowRate);
            tankInletTemp = pHex.HeatSourceOutletTemperature
              + disPump.GetElectricConsumption() / (WATER_SPECIFIC_HEAT * pHex.HeatSourceFlowRate);
          }
          wTank.UpdateState(tankInletTemp, 0.001 * pHex.HeatSourceFlowRate, true);

          IsOverLoad_C = IsOverLoad_C || pHex.IsOverLoad;
        }
        //冷凍機稼働+二次側負荷無し（蓄熱運転）
        else if (Charging)
        {
          pHex.ShutOff();
          disPump.ShutOff();
          chwPump.ShutOff();
          chgPump.UpdateState(chgPump.DesignFlowRate);
          calcChillerAndCoolingTower(wTank.UpperOutletTemperarture + dtChillingPump);
          wTank.UpdateState
            (chiller.ChilledWaterOutletTemperature, 0.001 * chiller.ChilledWaterFlowRate, false);
          IsOverLoad_C = false;
        }
        //冷凍機非稼働+二次側負荷無し（蓄熱槽熱損失のみ）
        else
        {
          ShutOff();
          wTank.UpdateState(0, 0, true);
          IsOverLoad_C = false;
        }
        aveTemp += pHex.SupplyTemperature * wTank.TimeStep;

        if (isLastCalc) break;
      }

      //過負荷の場合には平均的な供給水温を出力
      if (IsOverLoad_C) ChilledWaterSupplyTemperature = aveTemp / TimeStep;
      else ChilledWaterSupplyTemperature = ChilledWaterSupplyTemperatureSetpoint;
    }

    /// <summary>熱交換器の熱流を更新する</summary>
    /// <param name="hexFlow">熱交換器流量[kg/s]</param>
    /// <param name="rtnTmp">還冷水温度[C]（ポンプ昇温込み）</param>
    private void calcHexHeatTransfer(double hexFlow, double rtnTmp)
    {
      //水槽内の流れ方向を確定
      double ttlChilFlow = chiller.MaxChilledWaterFlowRate * ChillerNumber;
      bool isDownFlow = ttlChilFlow < hexFlow;

      //ターボ冷凍機供給温度を仮定
      double chilOut = StorageTemperature;
      double chilIn = dtChillingPump;
      if (isDownFlow)
      {
        double tHexIn = (WaterTank.LowerOutletTemperarture * (hexFlow - ttlChilFlow)
          + chilOut * ttlChilFlow) / hexFlow;
        pHex.Update(tHexIn, rtnTmp, hexFlow, ChilledWaterFlowRate);
        disPump.UpdateState(pHex.HeatSourceFlowRate);
        double dtDisP = disPump.GetElectricConsumption() / (WATER_SPECIFIC_HEAT * pHex.HeatSourceFlowRate);
        chilIn += pHex.HeatSourceOutletTemperature + dtDisP;
      }
      else
      {
        pHex.Update(chilOut, rtnTmp, hexFlow, ChilledWaterFlowRate);
        disPump.UpdateState(pHex.HeatSourceFlowRate);
        double dtDisP = disPump.GetElectricConsumption() / (WATER_SPECIFIC_HEAT * pHex.HeatSourceFlowRate);
        chilIn += ((pHex.HeatSourceOutletTemperature + dtDisP) * hexFlow 
          + wTank.UpperOutletTemperarture * (ttlChilFlow - hexFlow)) / ttlChilFlow;
      }

      //冷却塔と冷凍機の連成計算
      calcChillerAndCoolingTower(chilIn);

      //ターボ冷凍機過負荷（冷水出口温度が上昇）の場合には1回だけHEX交換熱量を補正
      if (chiller.IsOverLoad)
      {
        if (isDownFlow)
        {
          double tHexIn = (WaterTank.LowerOutletTemperarture * (hexFlow - ttlChilFlow)
            + chiller.ChilledWaterOutletTemperature * ttlChilFlow) / hexFlow;
          pHex.Update(tHexIn, rtnTmp, hexFlow, ChilledWaterFlowRate);
        }
        else
          pHex.Update(chiller.ChilledWaterOutletTemperature, rtnTmp, hexFlow, ChilledWaterFlowRate);
      }
    }

    /// <summary>ターボ冷凍機と冷却塔を連成計算する</summary>
    /// <param name="inletChilledWaterTemp">冷水還温度[C]</param>
    private void calcChillerAndCoolingTower(double inletChilledWaterTemp)
    {
      //冷却塔と冷凍機の連成計算（冷却水温度の計算）
      double mch = 1000 * chgPump.DesignFlowRate;
      double mcd = 1000 * cdwPump.DesignFlowRate;
      cdwPump.UpdateState(cdwPump.DesignFlowRate);

      bool needIteration = !ControlCoolingWaterTemperature;
      if (ControlCoolingWaterTemperature)
      {
        chiller.Update(CoolingWaterTemperatureSetpoint + dtCoolingPump, inletChilledWaterTemp, mcd, mch);
        cTower.OutletWaterSetPointTemperature = CoolingWaterTemperatureSetpoint;
        cTower.Update(chiller.CoolingWaterOutletTemperature, true);
        if (cTower.IsOverLoad) needIteration = true;
      }
      //過負荷または冷却水温度成行の場合には収束計算
      if (needIteration)
      {
        Roots.ErrorFunction eFnc = delegate (double cdt)
        {
          chiller.Update(cdt + dtCoolingPump, inletChilledWaterTemp, mcd, mch);
          cTower.Update(chiller.CoolingWaterOutletTemperature, cTower.MaxAirFlowRate);
          return cTower.OutletWaterTemperature - cdt;
        };
        double fmax = eFnc(OutdoorAir.WetbulbTemperature);
        double fmin = eFnc(37);
        if (0 <= fmin && fmax <= 0)
          Roots.Bisection(eFnc, 37, OutdoorAir.WetbulbTemperature, fmin, fmax, 0.01, 0.01, 10);
      }
    }

    /// <summary>状態を確定する</summary>
    public void FixState() { isForecasting = false; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="waterTank">温度成層型蓄熱槽</param>
    /// <param name="plateHex">プレート熱交換器</param>
    /// <param name="chiller">ターボ冷凍機</param>
    /// <param name="chwPump">冷水ポンプ</param>
    /// <param name="cdwPump">冷却水ポンプ</param>
    /// <param name="chargePump">蓄熱ポンプ</param>
    /// <param name="dischargePump">放熱ポンプ</param>
    /// <param name="cTower">冷却塔</param>
    /// <param name="chillerNumber">冷凍機台数</param>
    /// <param name="coolingTowerNumber">冷却塔台数（1台の冷凍機あたりの台数）</param>
    public MultipleStratifiedWaterTankSystem
      (MultipleStratifiedWaterTank waterTank, PlateHeatExchanger plateHex, ICentrifugalChiller chiller,
      CentrifugalPump chwPump, CentrifugalPump cdwPump, CentrifugalPump chargePump,
      CentrifugalPump dischargePump, CoolingTower cTower, int chillerNumber, int coolingTowerNumber)
    {
      this.chiller = chiller;
      this.chwPump = chwPump;
      this.cdwPump = cdwPump;
      this.chgPump = chargePump;
      this.disPump = dischargePump;
      this.cTower = cTower;
      this.pHex = plateHex;
      this.wTank = waterTank;
      this.ChillerNumber = chillerNumber;
      this.CoolingTowerNumber = coolingTowerNumber;
      oldTemps = new double[WaterTank.LayerNumber];

      //冷凍機ポンプの昇温幅を計算・保存
      chgPump.UpdateState(chgPump.DesignFlowRate);
      dtChillingPump = chgPump.GetElectricConsumption() / (WATER_SPECIFIC_HEAT * 1000 * chgPump.DesignFlowRate);
      cdwPump.UpdateState(cdwPump.DesignFlowRate);
      dtCoolingPump = cdwPump.GetElectricConsumption() / (WATER_SPECIFIC_HEAT * 1000 * cdwPump.DesignFlowRate);

      Mode = HeatSourceSystemModel.OperatingMode.ShutOff;
    }

    #endregion

  }
}
