﻿/* MultiConnectedWaterTank.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.ThermophysicalProperty;
using Popolo.Numerics.MatrixOperation;

namespace Popolo.HVAC.ThermalStorage
{
  /// <summary>連結完全混合型蓄熱槽</summary>
  [Serializable]
  public class MultiConnectedWaterTank
  {

    #region 定数宣言
    
    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>計算領域</summary>
    private IMatrix wMat;

    /// <summary>水槽温度[C]</summary>
    private IVector temperatures, temperatures_Back;

    /// <summary>予測計算中か否か</summary>
    private bool isForecasting = false;

    /// <summary>水槽容量[m3]</summary>
    private double[] volumes;

    /// <summary>熱損失係数[kW/K]</summary>
    private double[] heatLossCoefficients;

    /// <summary>タイムステップ[sec]</summary>
    private double timeStep = 60;

    /// <summary>タイムステップを設定[sec]・取得する</summary>
    public double TimeStep
    {
      get { return timeStep; }
      set { if (0 < value) timeStep = value; }
    }
    
    /// <summary>流入水温[C]を取得する</summary>
    public double WaterInletTemperature { get; private set; }

    /// <summary>流出水温[C]を取得する</summary>
    public double WaterOutletTemperarture
    {
      get
      {
        if (IsForwardFlow) return temperatures[temperatures.Length - 1];
        else return temperatures[0];
      }
    }

    /// <summary>流入量[m3/s]を取得する</summary>
    public double WaterFlowRate { get; private set; }

    /// <summary>周囲の温度[C]を設定・取得する</summary>
    public double AmbientTemperature { get; set; } = 20;

    /// <summary>順流か否か</summary>
    public bool IsForwardFlow { get; private set; }

    /// <summary>水槽の数を取得する</summary>
    public int TankNumber { get { return temperatures.Length; } }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="volumes">水槽容量[m3]</param>
    public MultiConnectedWaterTank(double[] volumes)
    {
      int num = volumes.Length;

      wMat = new Matrix(3, num);
      temperatures = new Vector(num);
      temperatures_Back = new Vector(num);
      temperatures.Initialize(20);
      temperatures_Back.Initialize(20);

      this.volumes = new double[num];
      volumes.CopyTo(this.volumes, 0);
      this.heatLossCoefficients = new double[num];
    }

    #endregion

    #region 水温設定関連の処理

    /// <summary>水槽温度[C]を初期化する</summary>
    /// <param name="temperature">初期化する温度[C]</param>
    public void InitializeTemperature(double temperature)
    {
      RestoreState();
      this.temperatures.Initialize(temperature);
    }

    /// <summary>水槽温度[C]を初期化する</summary>
    /// <param name="temperature">初期化する温度[C]</param>
    /// <param name="tankIndex">水槽番号</param>
    public void InitializeTemperature(int tankIndex, double temperature)
    {
      RestoreState();
      this.temperatures[tankIndex] = temperature;
    }

    /// <summary>水槽温度[C]を取得する</summary>
    /// <param name="tankIndex">水槽番号</param>
    /// <returns>水槽温度[C]</returns>
    public double GetTemperature(int tankIndex) { return temperatures[tankIndex]; }

    #endregion

    #region 水温更新関連の処理

    /// <summary>将来の水槽温度を予測する</summary>
    /// <param name="waterInletTemperature">流入水温[C]</param>
    /// <param name="waterFlowRate">流入流量[m3/s]</param>
    /// <param name="isForwardFlow">水流の向きは順流か否か</param>
    public void ForecastState(double waterInletTemperature, double waterFlowRate, bool isForwardFlow)
    {
      this.WaterInletTemperature = waterInletTemperature;
      this.WaterFlowRate = waterFlowRate;
      this.IsForwardFlow = isForwardFlow;

      //既に予測計算中の場合には状態を復元
      if (isForecasting) RestoreState();
      else
      {
        //現在の温度を保存
        isForecasting = true;
        for (int i = 0; i < temperatures.Length; i++) temperatures_Back[i] = temperatures[i];
      }

      //水温更新
      UpdateTemperature(ref temperatures, ref wMat, timeStep, WaterInletTemperature, WaterFlowRate, 
        heatLossCoefficients, AmbientTemperature, volumes, IsForwardFlow);
    }

    /// <summary>水槽温度を復元する</summary>
    public void RestoreState()
    {
      if (isForecasting)
        for (int i = 0; i < temperatures.Length; i++)
          temperatures[i] = temperatures_Back[i];
    }

    /// <summary>将来の状態を確定する</summary>
    public void FixState() { isForecasting = false; }

    #endregion

    #region 蓄熱量、蓄放熱流の計算処理

    /// <summary>蓄熱量[MJ]を計算する（温水蓄熱が正、冷水蓄熱が負）</summary>
    /// <param name="referenceTemperature">基準温度[C]</param>
    /// <returns>蓄熱量[MJ]</returns>
    public double GetHeatStorage(double referenceTemperature)
    {
      double sum = 0;
      for (int i = 0; i < volumes.Length; i++) sum += volumes[i] * (temperatures[i] - referenceTemperature);
      double rho = Water.GetLiquidDensity(referenceTemperature);
      return 0.001 * sum * rho * WATER_SPECIFIC_HEAT;
    }

    /// <summary>蓄熱流[kW]を計算する</summary>
    /// <returns>蓄熱流[kW]</returns>
    public double GetHeatStorageFlow()
    {
      double aveTemp = 0.5 * (WaterInletTemperature + WaterOutletTemperarture);
      return (WaterInletTemperature - WaterOutletTemperarture)
        * WaterFlowRate * Water.GetLiquidDensity(aveTemp) * WATER_SPECIFIC_HEAT;
    }

    #endregion

    #region 熱損失係数関連の処理

    /// <summary>熱損失係数[kW/K]を設定する</summary>
    /// <param name="tankIndex">水槽番号</param>
    /// <param name="heatLossCoefficient">熱損失係数[kW/K]</param>
    public void SetHeatLossCoefficient(int tankIndex, double heatLossCoefficient)
    { heatLossCoefficients[tankIndex] = heatLossCoefficient; }

    /// <summary>熱損失係数[kW/K]を取得する</summary>
    /// <param name="tankIndex">水槽番号</param>
    /// <returns>熱損失係数[kW/K]</returns>
    public double GetHeatLossCoefficient(int tankIndex)
    { return heatLossCoefficients[tankIndex]; }

    #endregion

    #region staticメソッド

    /// <summary>水槽温度を計算する</summary>
    /// <param name="temperatures">水槽温度[C]</param>
    /// <param name="wMat">計算領域（3×m）</param>
    /// <param name="timeStep">計算間隔[s]</param>
    /// <param name="waterInletTemperature">流入温度[C]</param>
    /// <param name="waterFlowRate">水量[m3/s]</param>
    /// <param name="heatLossCoefficients">熱損失係数[kW/K]</param>
    /// <param name="ambientTemperature">周囲の温度[C]</param>
    /// <param name="volumes">水槽容量[m3]</param>
    /// <param name="isForwardFlow">順流か</param>
    public static void UpdateTemperature
      (ref IVector temperatures, ref IMatrix wMat, double timeStep, double waterInletTemperature,
      double waterFlowRate, double[] heatLossCoefficients, double ambientTemperature, double[] volumes,
      bool isForwardFlow)
    {
      double tRhoc = timeStep / (WATER_SPECIFIC_HEAT * 1000);
      int num = temperatures.Length;
      double wft = waterFlowRate * timeStep;
      wMat.Initialize(0);

      //対角行列を作成
      for (int i = 0; i < num; i++)
      {
        double s = wft / volumes[i];
        double r = heatLossCoefficients[i] / volumes[i] * tRhoc;

        wMat[1, i] = 1 + s + r;
        temperatures[i] += ambientTemperature * r;
        if (isForwardFlow)
        {
          if (i == 0) temperatures[i] += s * waterInletTemperature;
          else wMat[0, i] = -s;
          wMat[2, i] = 0;
        }
        else
        {
          if (i == num - 1) temperatures[i] += s * waterInletTemperature;
          else wMat[2, i] = -s;
          wMat[0, i] = 0;
        }
      }
      LinearAlgebra.SolveTridiagonalMatrix(wMat, ref temperatures);
    }

    #endregion

  }
}
