﻿/* MultipleStratifiedWaterTank.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.ThermophysicalProperty;
using Popolo.Numerics.MatrixOperation;

namespace Popolo.HVAC.ThermalStorage
{
  /// <summary>温度成層型蓄熱槽</summary>
  [Serializable]
  public class MultipleStratifiedWaterTank: ImmutableMultipleStratifiedWaterTank
  {

    #region 定数宣言

    /// <summary>重力加速度[m/s2]</summary>
    private const double G_FORCES = 9.8;

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>計算領域</summary>
    private IMatrix wMat;
    private IVector wVec1, wVec2;

    /// <summary>水槽内温度分布[C]</summary>
    private IVector temperatures;
    
    /// <summary>タイムステップ[sec]</summary>
    private double timeStep = 60;

    /// <summary>タイムステップ[sec]を設定・取得する</summary>
    public double TimeStep
    {
      get { return timeStep; }
      set { if (0 < value) timeStep = value; }
    }

    /// <summary>水深[m]を取得する</summary>
    public double WaterDepth { get; private set; }

    /// <summary>水平断面積[m2]を取得する</summary>
    public double SectionalArea { get; private set; }

    /// <summary>水量[m3]を取得する</summary>
    public double WaterVolume { get { return WaterDepth * SectionalArea; } }

    /// <summary>流入水温[C]を取得する</summary>
    public double WaterInletTemperature { get; private set; }

    /// <summary>上部流出口水温[C]を取得する</summary>
    public double UpperOutletTemperarture
    { get { return temperatures[temperatures.Length - PipeInstallationLayer - 1]; } }

    /// <summary>下部流出口水温[C]を取得する</summary>
    public double LowerOutletTemperarture
    { get { return temperatures[PipeInstallationLayer]; } }

    /// <summary>流入量[m3/s]を取得する</summary>
    public double WaterFlowRate { get; private set; }

    /// <summary>流出入口直径[m]を取得する</summary>
    public double PipeDiameter { get; private set; }

    /// <summary>流出入口が設置された層番号を取得する</summary>
    public int PipeInstallationLayer { get; private set; }

    /// <summary>熱損失率[kW/K]を設定・取得する</summary>
    public double HeatLossCoefficient { get; set; }

    /// <summary>周囲の温度[C]を設定・取得する</summary>
    public double AmbientTemperature { get; set; } = 20;

    /// <summary>下向き水流か否か</summary>
    public bool IsDownFlow { get; private set; }

    /// <summary>水槽の分割数を取得する</summary>
    public int LayerNumber { get { return temperatures.Length; } }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="waterDepth">水深[m]</param>
    /// <param name="sectionalArea">水平断面積[m2]</param>
    /// <param name="pipeDiameter">流出入口直径[m]</param>
    /// <param name="pipeInstallationHeight">流出入口設置高さ[m]（上部）</param>
    /// <param name="layerNumber">分割数</param>
    public MultipleStratifiedWaterTank
      (double waterDepth, double sectionalArea, double pipeDiameter, double pipeInstallationHeight, int layerNumber)
    {
      this.WaterDepth = waterDepth;
      this.SectionalArea = sectionalArea;
      this.PipeDiameter = pipeDiameter;
      wMat = new Matrix(3, layerNumber);
      wVec1 = new Vector(layerNumber);
      wVec2 = new Vector(layerNumber);
      temperatures = new Vector(layerNumber);
      temperatures.Initialize(20);

      double dz = WaterDepth / LayerNumber;
      PipeInstallationLayer = (int)Math.Floor(pipeInstallationHeight / dz);
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>水槽内温度[C]を初期化する</summary>
    /// <param name="temperature">初期化する温度[C]</param>
    public void InitializeTemperature(double temperature)
    { this.temperatures.Initialize(temperature); }

    /// <summary>水槽内温度[C]を初期化する</summary>
    /// <param name="temperature">初期化する温度[C]</param>
    /// <param name="layerNumber">分割番号</param>
    public void InitializeTemperature(int layerNumber, double temperature)
    { this.temperatures[layerNumber] = temperature; }

    /// <summary>水槽内温度分布を設定する</summary>
    /// <param name="temperatures">水槽内温度分布</param>
    public void InitializeTemperature(double[] temperatures)
    { for (int i = 0; i < temperatures.Length; i++) this.temperatures[i] = temperatures[i]; }

    /// <summary>水槽内温度[C]を取得する</summary>
    /// <param name="layerNumber">分割番号</param>
    /// <returns>水槽内温度[C]</returns>
    public double GetTemperature(int layerNumber) { return temperatures[layerNumber]; }

    /// <summary>水槽内温度分布を取得する</summary>
    /// <param name="temperatures">出力:水槽内温度分布</param>
    public void GetTemperatures(ref double[] temperatures)
    { for (int i = 0; i < temperatures.Length; i++) temperatures[i] = this.temperatures[i]; }

    /// <summary>将来の水温分布を予測する</summary>
    /// <param name="waterInletTemperature">流入水温[C]</param>
    /// <param name="waterFlowRate">流入流量[m3/s]</param>
    /// <param name="isDownFlow">水流の向きは下向きか否か</param>
    public void UpdateState(double waterInletTemperature, double waterFlowRate, bool isDownFlow)
    {
      this.WaterInletTemperature = waterInletTemperature;
      this.WaterFlowRate = waterFlowRate;
      this.IsDownFlow = isDownFlow;
      
      UpdateTemperature(ref temperatures, ref wMat, ref wVec1, ref wVec2, timeStep, 
        WaterInletTemperature, WaterFlowRate, HeatLossCoefficient, AmbientTemperature, WaterDepth,
        PipeDiameter, SectionalArea, PipeInstallationLayer, IsDownFlow);
    }
    
    #endregion

    #region 蓄熱量、蓄放熱流の計算処理

    /// <summary>蓄熱量[MJ]を計算する（温水蓄熱が正、冷水蓄熱が負）</summary>
    /// <param name="referenceTemperature">基準温度[C]</param>
    /// <returns>蓄熱量[MJ]</returns>
    public double GetHeatStorage(double referenceTemperature)
    {
      double dz = WaterDepth / temperatures.Length;  //分割幅
      double sum = 0;
      for (int i = 0; i < temperatures.Length; i++) sum += (temperatures[i] - referenceTemperature);
      double rho = Water.GetLiquidDensity(referenceTemperature);
      return 0.001 * sum * rho * WATER_SPECIFIC_HEAT * dz * SectionalArea;
    }

    /// <summary>蓄熱流[kW]を計算する</summary>
    /// <returns>蓄熱流[kW]</returns>
    public double GetHeatStorageFlow()
    {
      double two;
      if (IsDownFlow) two = LowerOutletTemperarture;
      else two = UpperOutletTemperarture;
      double aveTemp = 0.5 * (WaterInletTemperature + two);
      return (WaterInletTemperature - two) * WaterFlowRate * Water.GetLiquidDensity(aveTemp) * WATER_SPECIFIC_HEAT;
    }

    #endregion

    #region staticメソッド

    /// <summary>槽内温度を計算する</summary>
    /// <param name="temperature">槽内温度分布[C]</param>
    /// <param name="wMat">計算領域（3×層の数）</param>
    /// <param name="wVec1">計算領域(層の数)</param>
    /// <param name="wVec2">計算領域(層の数)</param>
    /// <param name="timeStep">タイムステップ[sec]</param>
    /// <param name="waterInletTemperature">流入水温[C]</param>
    /// <param name="waterFlowRate">流入量[m3/s]</param>
    /// <param name="heatLossCoefficient">熱損失率[kW/K]</param>
    /// <param name="ambientTemperature">周囲の温度[C]</param>
    /// <param name="waterDepth">水槽の水深[m]</param>
    /// <param name="pipeDiameter">流出入口直径[m]</param>
    /// <param name="sectionalArea">断面積[m2]</param>
    /// <param name="pipeInstallationLayer">流入出口設置層番号</param>
    /// <param name="isDownFlow">下向きか否か</param>
    public static void UpdateTemperature
      (ref IVector temperature, ref IMatrix wMat, ref IVector wVec1, ref IVector wVec2, double timeStep,
      double waterInletTemperature, double waterFlowRate, double heatLossCoefficient, double ambientTemperature,
      double waterDepth, double pipeDiameter, double sectionalArea, int pipeInstallationLayer, bool isDownFlow)
    {
      int layerNum = temperature.Length;  //分割数
      double dz = waterDepth / layerNum;  //分割幅

      //平均水温を計算
      double aveTemp = 0;
      for (int i = 0; i < layerNum; i++) aveTemp += temperature[i];
      aveTemp /= layerNum;

      //混合域の噴流配分の計算
      IVector phiN = wVec1;
      IVector uN = wVec2;
      if (waterFlowRate == 0)
      {
        phiN.Initialize(0);
        uN.Initialize(0);
      }
      //温度逆転の場合は混合すると温度が逆転する層まで完全混合とし、噴流は等分割
      else if ((isDownFlow && waterInletTemperature <= temperature[0]) 
        || (!isDownFlow && temperature[layerNum - 1] <= waterInletTemperature))
      {
        //温度逆転範囲を求めて平均温度を計算
        double mixTemp, mixedAve;
        int mixedNum;
        if (isDownFlow)
        {
          mixedAve = temperature[0];
          mixTemp = temperature[0] 
            + timeStep / (sectionalArea * dz) * (waterInletTemperature - temperature[0]) * waterFlowRate;
        }
        else
        {
          mixedAve = temperature[temperature.Length - 1];
          mixTemp = temperature[temperature.Length - 1] 
            + timeStep * (waterInletTemperature - temperature[temperature.Length - 1]) / (sectionalArea * dz);
        }
        
        for (mixedNum = 1; mixedNum < layerNum; mixedNum++)
        {
          if (isDownFlow)
          {
            if (temperature[mixedNum] < mixTemp) break;
            mixTemp = (mixTemp * mixedNum + temperature[mixedNum]) / (mixedNum + 1);
            mixedAve += temperature[mixedNum];
          }
          else
          {
            int tgtLayer = layerNum - (mixedNum + 1);
            if (mixTemp < temperature[tgtLayer]) break;
            mixTemp = 0.5 * (mixTemp + temperature[tgtLayer]);
            mixedAve += temperature[tgtLayer];
          }
        }
        mixedAve /= mixedNum;
        double bf = waterFlowRate / (mixedNum * dz);

        //平均温度と噴流配分を設定
        for (int i = 0; i < layerNum; i++)
        {
          int tgtLayer = i;
          if (!isDownFlow) tgtLayer = layerNum - (i + 1);

          if (i < mixedNum)
          {
            phiN[tgtLayer] = bf;
            temperature[tgtLayer] = mixedAve;
            uN[tgtLayer] = dz / sectionalArea * phiN[tgtLayer];
          }
          else phiN[tgtLayer] = uN[tgtLayer] = 0;
          if (i != 0)
          {
            if (isDownFlow) uN[tgtLayer] += uN[tgtLayer - 1];
            else uN[tgtLayer] += uN[tgtLayer + 1];
          }
        }
      }
      else
      {
        double rhowi = Water.GetLiquidDensity(waterInletTemperature);
        double pipeSA = Math.Pow(pipeDiameter / 2, 2) * Math.PI; //流入口断面積[m2]
        double uwi2 = Math.Pow(waterFlowRate / pipeSA, 2);//流入速度2乗[(m/s)^2]

        //噴流が到達する層を求める
        int lmax;
        double tgt = rhowi * uwi2 / (G_FORCES * dz);
        double rr = 0;
        for (lmax = 0; lmax < layerNum - 1; lmax++)
        {
          if (isDownFlow) rr += (Water.GetLiquidDensity(temperature[lmax]) - rhowi);
          else rr += (rhowi - Water.GetLiquidDensity(temperature[layerNum - lmax - 1]));
          if (tgt < rr) break;
        }
        if (!isDownFlow) lmax = layerNum - lmax - 1;

        //混合域深さの計算************
        double tlMax = temperature[lmax];
        double lm;
        //等温の層まで到達する場合には混合域100%とする
        if (tlMax == waterInletTemperature) lm = waterDepth;
        else
        {          
          double rho0 = Water.GetLiquidDensity(tlMax);
          double ari = pipeDiameter * G_FORCES * Math.Abs(rho0 - rhowi) / (rho0 * uwi2);  //アルキメデス数
          double ndt = 0;
          if (isDownFlow)
            for (int i = 0; i < lmax; i++) ndt += (temperature[i] - tlMax);
          else for (int i = lmax; i < layerNum; i++) ndt += (temperature[i] - tlMax);
          ndt = (ndt * dz) / (waterDepth * (waterInletTemperature - tlMax));  //無次元時間
          lm = waterDepth * 0.8 * Math.Pow(ari, -0.5) * pipeDiameter / waterDepth + 0.5 * ndt;
        }

        double z1 = 0;
        double bf = waterFlowRate / (2 * lm * lm * lm);
        for (int i = 0; i < layerNum; i++)
        {
          int ln = i;
          if (!isDownFlow) ln = layerNum - (i + 1);
          double z2 = z1 + dz;
          if (z2 < lm) phiN[ln] = bf * (3 * lm * lm - dz * dz * (3 * i * (i + 1) + 1));
          else if (lm < z1) phiN[ln] = 0;
          else phiN[ln] = bf * Math.Pow(i * dz - lm, 2) * (i + 2 * lm / dz);

          if (i == 0) uN[ln] = dz / sectionalArea * phiN[ln];
          else if (isDownFlow) uN[ln] = uN[ln - 1] + dz / sectionalArea * phiN[ln];
          else uN[ln] = uN[ln + 1] + dz / sectionalArea * phiN[ln];
          z1 = z2;
        }
      }

      //3重対角行列を解く
      wMat.Initialize(0);
      int outletLayer = pipeInstallationLayer;
      if (!isDownFlow) outletLayer = layerNum - pipeInstallationLayer - 1;
      double s = timeStep * Water.GetLiquidThermalDiffusivity(aveTemp) / (dz * dz);
      double p = heatLossCoefficient * timeStep /
        (Water.GetLiquidDensity(aveTemp) * WATER_SPECIFIC_HEAT * waterDepth * sectionalArea);
      for (int i = 0; i < layerNum; i++)
      {

        double r1 = 0;
        double r2 = 0;
        if (isDownFlow && i != 0 && i <= outletLayer) r1 = uN[i - 1] * timeStep / dz;
        if (!isDownFlow && i != layerNum - 1 && outletLayer <= i) r2 = uN[i + 1] * timeStep / dz;

        if (i == layerNum - 1) wMat[0, i] = -(2 * s + r1);
        else if (i != 0) wMat[0, i] = -(s + r1);

        if (i == 0) wMat[2, i] = -(2 * s + r2);
        else if (i != layerNum - 1) wMat[2, i] = -(s + r2);

        wMat[1, i] = 2 * s + r1 + r2 + 1 + p;
        temperature[i] += p * ambientTemperature;
        if (phiN[i] != 0)
        {
          double q = timeStep * phiN[i] / sectionalArea;
          wMat[1, i] += q;
          temperature[i] += q * waterInletTemperature;
        }
      }
      LinearAlgebra.SolveTridiagonalMatrix(wMat, ref temperature);
    }

    #endregion

  }

  #region 読み取り専用インターフェース

  /// <summary>読み取り専用の温度成層型蓄熱槽</summary>
  public interface ImmutableMultipleStratifiedWaterTank
  {
    /// <summary>タイムステップ[sec]を取得する</summary>
    double TimeStep { get; }

    /// <summary>水深[m]を取得する</summary>
    double WaterDepth { get; }

    /// <summary>水平断面積[m2]を取得する</summary>
    double SectionalArea { get; }

    /// <summary>水量[m3]を取得する</summary>
    double WaterVolume { get; }

    /// <summary>流入水温[C]を取得する</summary>
    double WaterInletTemperature { get; }

    /// <summary>上部流出口水温[C]を取得する</summary>
    double UpperOutletTemperarture { get; }

    /// <summary>下部流出口水温[C]を取得する</summary>
    double LowerOutletTemperarture { get; }

    /// <summary>流入量[m3/s]を取得する</summary>
    double WaterFlowRate { get; }

    /// <summary>流出入口直径[m]を取得する</summary>
    double PipeDiameter { get; }

    /// <summary>流出入口が設置された層番号を取得する</summary>
    int PipeInstallationLayer { get; }

    /// <summary>熱損失率[kW/K]を取得する</summary>
    double HeatLossCoefficient { get; }

    /// <summary>周囲の温度[C]を取得する</summary>
    double AmbientTemperature { get; }

    /// <summary>下向き水流か否か</summary>
    bool IsDownFlow { get; }

    /// <summary>水槽の分割数を取得する</summary>
    int LayerNumber { get; }

    /// <summary>水槽内温度[C]を取得する</summary>
    /// <param name="layerNumber">分割番号</param>
    /// <returns>水槽内温度[C]</returns>
    double GetTemperature(int layerNumber);

    /// <summary>蓄熱量[MJ]を計算する（温水蓄熱が正、冷水蓄熱が負）</summary>
    /// <param name="referenceTemperature">基準温度[C]</param>
    /// <returns>蓄熱量[MJ]</returns>
    double GetHeatStorage(double referenceTemperature);

    /// <summary>蓄熱流[kW]を計算する</summary>
    /// <returns>蓄熱流[kW]</returns>
    double GetHeatStorageFlow();

  }

  #endregion

}
