﻿/* MultiNodeModel.cs
 * 
 * Copyright (C) 2014 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;
using System.Collections.Generic;

using Popolo.ThermophysicalProperty;
using Popolo.Numerics.MatrixOperation;

namespace Popolo.HumanBody
{
  /// <summary>多質点人体モデル</summary>
  [Serializable]
  public partial class MultiNodeModel : ImmutableMultiNodeModel
  {

    #region 定数宣言

    /// <summary>血液の体積比熱[J/(mLK)]</summary>
    public const double BLD_SPECIFICHEAT = 3.842;

    /// <summary>代謝量換算係数[(W/m2)/met]</summary>
    private const double CONVERT_MET_TO_W = 58.2;

    /// <summary>標準体躯の表面積[m2]</summary>
    private const double STANDARD_SURFACE_AREA = 1.87;

    /// <summary>標準体躯の血流量[mL/s]</summary>
    private const double STANDARD_BLOOD_FLOW = 290.004 / 3.6;

    /// <summary>標準体躯の重量[kg]</summary>
    private const double STANDARD_WEIGHT = 74.43;

    /// <summary>大気圧=101.325[kPa]</summary>
    private const double ATMOSPHERIC_PRESSURE = 101.325;

    /// <summary>骨の定圧比熱[J/(kgK)]</summary>
    private const double SPECIFIC_HEAT_BORN = 2092;

    /// <summary>脂肪の定圧比熱[J/(kgK)]</summary>
    private const double SPECIFIC_HEAT_FAT = 2510;

    /// <summary>その他の人体定圧比熱[J/(kgK)]</summary>
    private const double SPECIFIC_HEAT_ELSE = 3766;

    /// <summary>黒体の放射定数[W/(m2K4)]</summary>
    private const double BLACK_CONSTANT = 5.67e-8;

    /// <summary>絶対温度と摂氏との変換定数</summary>
    private const double CONVERT_C_TO_K = 273.15;

    /// <summary>衣服の透湿係数[K/kPa]</summary>
    private const double I_CLS = 0.45;

    /// <summary>四肢末端部位</summary>
    private const Node TERMINAL_NODE =
      Node.LeftHand | Node.RightHand | Node.LeftFoot | Node.RightFoot;

    /// <summary>四肢部位</summary>
    private const Node LIMBS =
      TERMINAL_NODE | Node.LeftShoulder | Node.RightShoulder | Node.LeftArm | Node.RightArm |
      Node.LeftThigh | Node.RightThigh | Node.LeftLeg | Node.RightLeg;

    #endregion

    #region 列挙型定義

    /// <summary>体の部位</summary>
    [Flags]
    public enum Node
    {
      /// <summary>頭</summary>
      Head = 1,
      /// <summary>首</summary>
      Neck = 2,
      /// <summary>胸</summary>
      Chest = 4,
      /// <summary>背中</summary>
      Back = 8,
      /// <summary>下腹部</summary>
      Pelvis = 16,
      /// <summary>左肩</summary>
      LeftShoulder = 32,
      /// <summary>左腕</summary>
      LeftArm = 64,
      /// <summary>左手</summary>
      LeftHand = 128,
      /// <summary>右肩</summary>
      RightShoulder = 256,
      /// <summary>右腕</summary>
      RightArm = 512,
      /// <summary>右手</summary>
      RightHand = 1024,
      /// <summary>左大腿部</summary>
      LeftThigh = 2048,
      /// <summary>左下腿部</summary>
      LeftLeg = 4096,
      /// <summary>左脚</summary>
      LeftFoot = 8192,
      /// <summary>右大腿部</summary>
      RightThigh = 16384,
      /// <summary>右下腿部</summary>
      RightLeg = 32768,
      /// <summary>右脚</summary>
      RightFoot = 65536,
    }

    /// <summary>組織</summary>
    public enum Layer
    {
      /// <summary>核層</summary>
      Core = 0,
      /// <summary>筋肉層</summary>
      Muscle = 1,
      /// <summary>脂肪層</summary>
      Fat = 2,
      /// <summary>皮膚層</summary>
      Skin = 4,
      /// <summary>動脈</summary>
      Artery = 8,
      /// <summary>表在静脈</summary>
      SuperficialVein = 16,
      /// <summary>深部動脈</summary>
      DeepVein = 32,
      /// <summary>AVA</summary>
      AVA = 64
    }

    #endregion

    #region インスタンス変数

    /// <summary>体の部位</summary>
    private Dictionary<Node, bodyPart> parts = new Dictionary<Node, bodyPart>();

    /// <summary>筋肉の重量比[-]</summary>
    private Dictionary<Node, double> rMuscle = new Dictionary<Node, double>();

    /// <summary>温度ベクトル</summary>
    private IVector tVector = new Vector(115);

    /// <summary>核の平均セットポイント[C]</summary>
    private double averageCoreSetPoint;

    /// <summary>皮膚の平均セットポイント[C]</summary>
    private double averageSkinSetPoint;

    #endregion

    #region プロパティ

    /// <summary>体重[kg]を取得する</summary>
    public double Weight { get; private set; }

    /// <summary>身長[m]を取得する</summary>
    public double Height { get; private set; }

    /// <summary>年齢[-]を取得する</summary>
    public double Age { get; private set; }

    /// <summary>男性か否か</summary>
    public bool IsMale { get; private set; }

    /// <summary>立位か否か</summary>
    public bool IsStanding { get; private set; }

    /// <summary>体脂肪率[-]を取得する</summary>
    public double FatPercentage { get; private set; }

    /// <summary>体表面積[m2]を取得する</summary>
    public double SurfaceArea { get; private set; }

    /// <summary>中央血液溜まりの温度[C]を取得する</summary>
    public double CentralBloodTemperature { get; private set; }

    /// <summary>呼吸による熱損失[W]を取得する</summary>
    public double HeatLossByBreathing { get; private set; }

    /// <summary>全身の基礎血流[mL/s]を取得する</summary>
    public double BasalBloodFlow { get; private set; }

    /// <summary>全身の血流[mL/s]を取得する</summary>
    public double BloodFlow { get; private set; }

    /// <summary>全身の基礎代謝[W]を取得する</summary>
    public double BasalMetabolicRate { get; private set; }

    /// <summary>全身の代謝量[W]を取得する</summary>
    public double MetabolicRate { get; private set; }

    /// <summary>計算時間間隔[sec]を取得する</summary>
    public double TimeStep { get; private set; }

    #endregion

    #region コンストラクタ・初期化処理

    /// <summary>インスタンスを初期化する</summary>
    /// <remarks>引数が無い場合には標準体躯とする</remarks>
    public MultiNodeModel() : this(74.43, 1.72, 25, true, 0.15, true) { }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="weight">体重[kg]</param>
    /// <param name="height">身長[m]</param>
    /// <param name="age">年齢</param>
    /// <param name="isMale">男性か否か</param>
    /// <param name="fatPercentage">体脂肪率[%]</param>
    /// <param name="isStanding">立位か否か</param>
    public MultiNodeModel
      (double weight, double height, double age, bool isMale, double fatPercentage, bool isStanding)
    {
      //情報保存
      Weight = weight;
      Height = height;
      Age = age;
      IsMale = isMale;
      IsStanding = isStanding;
      FatPercentage = fatPercentage;

      //体表面積[m2]を計算
      SurfaceArea = 0.202 * Math.Pow(Weight, 0.425) * Math.Pow(Height, 0.725);
      //心係数[(mL/s)/m2]を計算
      double ci = 115 + Age * (-2.4 + Age * (0.0167 + Age * (3.56e-4 - 4.29e-6 * Age)));
      //基礎血流[mL/s]を計算
      BasalBloodFlow = ci * SurfaceArea;

      //基礎代謝[W]を計算
      BasalMetabolicRate = (0.1238 + 2.34 * height + 0.0481 * weight - 0.0138 * age);
      if (isMale) BasalMetabolicRate = (BasalMetabolicRate - 0.5473) / 0.0864;
      else BasalMetabolicRate = (BasalMetabolicRate - 0.5473 * 2) / 0.0864;

      //部位を作成
      double muscleSum = 0;
      Node[] nds = (Node[])Enum.GetValues(typeof(Node));
      foreach (Node nd in nds)
      {
        bodyPart bp = new bodyPart
          (this, nd, Weight, Height, SurfaceArea, fatPercentage, BasalMetabolicRate, BasalBloodFlow);
        parts.Add(nd, bp);
        muscleSum += bp.muscleWeight;
      }
      //筋肉の重量比を計算
      foreach (Node nd in parts.Keys) rMuscle[nd] = parts[nd].muscleWeight / muscleSum;

      //部位を接続
      parts[Node.Neck].connect(parts[Node.Head]);
      parts[Node.Pelvis].connect(parts[Node.LeftThigh]);
      parts[Node.Pelvis].connect(parts[Node.RightThigh]);
      parts[Node.LeftThigh].connect(parts[Node.LeftLeg]);
      parts[Node.RightThigh].connect(parts[Node.RightLeg]);
      parts[Node.LeftLeg].connect(parts[Node.LeftFoot]);
      parts[Node.RightLeg].connect(parts[Node.RightFoot]);
      parts[Node.LeftShoulder].connect(parts[Node.LeftArm]);
      parts[Node.RightShoulder].connect(parts[Node.RightArm]);
      parts[Node.LeftArm].connect(parts[Node.LeftHand]);
      parts[Node.RightArm].connect(parts[Node.RightHand]);

      //体温を初期化
      InitializeTemperature(36);

      //セットポイント初期化
      initializeSetPoint();
    }

    /// <summary>セットポイントを初期化する</summary>
    private void initializeSetPoint()
    {
      //制御をOFF
      foreach (Node nd in parts.Keys)
      {
        parts[nd].initializing = true;
        parts[nd].setClothingIndex(0);
      }

      //PMV=0となる境界条件を設定
      UpdateBoundary(0, 28.8, 28.8, 50);
      SetMetabolicRate(1);

      //定常状態まで計算（24時間）
      for (int i = 0; i < 24; i++) Update(3600);

      //セットポイント設定
      foreach (Node nd in parts.Keys)
      {
        parts[nd].initializing = false;
        parts[nd].setPoint_Skin = parts[nd].temperatures[Layer.Skin];
        parts[nd].setPoint_Core = parts[nd].temperatures[Layer.Core];
      }

      //平均セットポイントを作成
      //体中心の核
      Layer core = Layer.Core;
      double capSum = parts[Node.Chest].heatCapacity[core]
        + parts[Node.Pelvis].heatCapacity[core] + parts[Node.Back].heatCapacity[core];
      averageCoreSetPoint =
        (parts[Node.Chest].setPoint_Core * parts[Node.Chest].heatCapacity[core]
        + parts[Node.Pelvis].setPoint_Core * parts[Node.Pelvis].heatCapacity[core]
        + parts[Node.Back].setPoint_Core * parts[Node.Back].heatCapacity[core]) / capSum;

      //全身の皮膚
      averageSkinSetPoint = 0;
      foreach (Node bp in parts.Keys) averageSkinSetPoint += parts[bp].setPoint_Skin * parts[bp].surfaceArea;
      averageSkinSetPoint /= SurfaceArea;
    }

    /// <summary>体温を初期化する</summary>
    /// <param name="temperature">初期化する温度[C]</param>
    public void InitializeTemperature(double temperature)
    {
      //行列に設定
      for (int i = 0; i < tVector.Length; i++) tVector[i] = temperature;
      //各部位への設定処理
      CentralBloodTemperature = temperature;
      foreach (Node nd in parts.Keys) parts[nd].updateTemperature(tVector);
    }

    #endregion

    #region staticメソッド

    /// <summary>上流の部位を取得する</summary>
    /// <param name="node">部位</param>
    /// <returns>上流の部位</returns>
    public static Node GetUpstreamNode(Node node)
    {
      switch (node)
      {
        case Node.Head:
          return Node.Neck;
        case Node.LeftHand:
          return Node.LeftArm;
        case Node.LeftArm:
          return Node.LeftShoulder;
        case Node.RightHand:
          return Node.RightArm;
        case Node.RightArm:
          return Node.RightShoulder;
        case Node.LeftFoot:
          return Node.LeftLeg;
        case Node.LeftLeg:
          return Node.LeftThigh;
        case Node.LeftThigh:
          return Node.Pelvis;
        case Node.RightFoot:
          return Node.RightLeg;
        case Node.RightLeg:
          return Node.RightThigh;
        case Node.RightThigh:
          return Node.Pelvis;
        default:
          return 0;
      }
    }

    /// <summary>下流の部位を取得する</summary>
    /// <param name="node">部位</param>
    /// <returns>下流の部位</returns>
    public static Node[] GetDownStreamNode(Node node)
    {
      switch (node)
      {
        case Node.Neck:
          return new Node[] { Node.Head };
        case Node.LeftArm:
          return new Node[] { Node.LeftHand };
        case Node.LeftShoulder:
          return new Node[] { Node.LeftArm };
        case Node.RightArm:
          return new Node[] { Node.RightHand };
        case Node.RightShoulder:
          return new Node[] { Node.RightArm };
        case Node.LeftLeg:
          return new Node[] { Node.LeftFoot };
        case Node.LeftThigh:
          return new Node[] { Node.LeftLeg };
        case Node.RightLeg:
          return new Node[] { Node.RightFoot };
        case Node.RightThigh:
          return new Node[] { Node.RightLeg };
        case Node.Pelvis:
          return new Node[] { Node.LeftThigh, Node.RightThigh };
        default:
          return new Node[] { };
      }
    }

    #endregion

    #region 更新処理

    /// <summary>状態を更新する</summary>
    /// <param name="timeStep">計算間隔[sec]</param>
    public void Update(double timeStep)
    {
      this.TimeStep = timeStep;
      IVector zVector = new Vector(115);

      //制御を更新
      updateControl();

      //計算用疎行列を用意
      SparseMatrix bMatrix = new SparseMatrix(115, 115);

      double mr = 0;
      foreach (Node nd in parts.Keys)
      {
        //行列に各部位の項を設定
        parts[nd].makeMatrix(bMatrix, zVector);
        //代謝量を積算
        mr += parts[nd].shiveringLoad + parts[nd].externalWork
          + parts[nd].basalMetabolicRate[Layer.Core] + parts[nd].basalMetabolicRate[Layer.Muscle]
          + parts[nd].basalMetabolicRate[Layer.Fat] + parts[nd].basalMetabolicRate[Layer.Skin];
      }

      //胸部の呼吸の項を追加
      bodyPart head = parts[Node.Head];
      HeatLossByBreathing = mr * (0.0014 * (34 - head.drybulbTemperature)
        + 0.017251 * (5.8662 - head.waterVaporPressure));
      zVector[13] -= HeatLossByBreathing;

      //中央血液溜まりの項を追加
      bMatrix[0, 0] = parts[Node.Chest].centralBloodHeatCapacity / timeStep;
      bodyPart[] bps = new bodyPart[] { parts[Node.Neck], parts[Node.LeftShoulder],
        parts[Node.RightShoulder], parts[Node.Pelvis], parts[Node.Chest], parts[Node.Back]};
      for (int i = 0; i < bps.Length; i++)
      {
        double dvf = BLD_SPECIFICHEAT * bps[i].bloodFlow[Layer.DeepVein];
        double svf = BLD_SPECIFICHEAT * bps[i].bloodFlow[Layer.SuperficialVein];
        bMatrix[0, 0] += (dvf + svf);
        bMatrix[0, bodyPart.mOffset[bps[i].node] + 5] = -dvf;
        bMatrix[0, bodyPart.mOffset[bps[i].node] + 6] = -svf;
      }
      zVector[0] = parts[Node.Chest].centralBloodHeatCapacity / timeStep * CentralBloodTemperature;

      //連立代数方程式を解く
      bMatrix.SolveLinearEquation(zVector, ref tVector);

      //設定
      foreach (Node nd in parts.Keys) parts[nd].updateTemperature(tVector);
      CentralBloodTemperature = tVector[0];
    }

    /// <summary>制御信号を計算する</summary>
    private void updateControl()
    {
      //全身の温冷感信号の計算
      double cldSignal, wrmSignal;
      cldSignal = wrmSignal = 0;
      foreach (Node bp in parts.Keys)
      {
        double sig = parts[bp].getSignal();
        if (0 < sig) wrmSignal += sig;
        else cldSignal += sig;
      }
      double signal = wrmSignal + cldSignal;

      //頭部の核の温冷感信号
      Layer core = Layer.Core;
      double sHead = parts[Node.Head].temperatures[core] - parts[Node.Head].setPoint_Core;

      //発汗・ふるえ・血管収縮・血管拡張の信号を計算
      double sfRate = SurfaceArea / STANDARD_SURFACE_AREA;
      double sweatSignal = Math.Max(0, (371.2 * sHead + 33.64 * signal)) * sfRate;
      double shiveringSignal = (-24.36 * Math.Max(0, -sHead) * cldSignal) * sfRate;
      double vasoconstrictionSignal = Math.Max(0, -10.8 * sHead - 10.8 * signal);
      double vasodilatationSignal = Math.Max(0, 32.5 * sHead + 2.08 * signal);
      vasodilatationSignal *= BasalBloodFlow / STANDARD_BLOOD_FLOW;

      //AVA血流開度信号を計算
      //体中心（腰・胸・背中）の平均温度を計算
      double capSum = parts[Node.Chest].heatCapacity[core]
          + parts[Node.Pelvis].heatCapacity[core] + parts[Node.Back].heatCapacity[core];
      double aveCore = 0;
      aveCore += parts[Node.Chest].temperatures[core] * parts[Node.Chest].heatCapacity[core];
      aveCore += parts[Node.Pelvis].temperatures[core] * parts[Node.Pelvis].heatCapacity[core];
      aveCore += parts[Node.Back].temperatures[core] * parts[Node.Back].heatCapacity[core];
      aveCore /= capSum;
      //全身の平均皮膚温を計算
      double atSkin = GetAverageSkinTemperature();
      //AVA開度の計算
      double ovaHand = 0.265 * (atSkin - (averageSkinSetPoint - 0.43))
          + 0.953 * (aveCore - (averageCoreSetPoint - 0.1905)) + 0.9126;
      double ovaFoot = 0.265 * (atSkin - (averageSkinSetPoint - 0.97))
          + 0.953 * (aveCore - (averageCoreSetPoint + 0.0095)) + 0.9126;

      //皮膚血管運動・発汗・ふるえ熱生産・AVA血流を計算
      foreach (Node bp in parts.Keys)
      {
        if ((bp == Node.LeftHand) || (bp == Node.RightHand))
          parts[bp].updateControl
            (signal, sweatSignal, shiveringSignal, vasodilatationSignal, vasoconstrictionSignal, ovaHand);
        else parts[bp].updateControl
            (signal, sweatSignal, shiveringSignal, vasodilatationSignal, vasoconstrictionSignal, ovaFoot);
      }

      //血流更新処理//中央血液溜まり直結部以外はメソッド内で再帰的に呼び出し
      BloodFlow = 0;
      Node[] nds =
        new Node[] { Node.Neck, Node.Chest, Node.Back, Node.Pelvis, Node.LeftShoulder, Node.RightShoulder };
      for (int i = 0; i < nds.Length; i++)
      {
        parts[nds[i]].updateBloodFlow();
        BloodFlow += parts[nds[i]].bloodFlow[Layer.DeepVein] + parts[nds[i]].bloodFlow[Layer.SuperficialVein];
      }

      //全身代謝量[W]を更新
      MetabolicRate = BasalMetabolicRate;
      foreach (Node nd in parts.Keys)
      {
        bodyPart part = parts[nd];
        MetabolicRate += part.externalWork + part.shiveringLoad;
      }
    }

    #endregion

    #region 境界条件設定処理

    /// <summary>代謝量[met]を設定する</summary>
    /// <param name="met">代謝量[met]</param>
    public void SetMetabolicRate(double met)
    {
      met = Math.Max(0, CONVERT_MET_TO_W * met * SurfaceArea - BasalMetabolicRate);
      foreach (Node bp in parts.Keys) parts[bp].externalWork = rMuscle[bp] * met;
    }

    /// <summary>境界条件を更新する</summary>
    /// <param name="velocity">気流速度[m/s]</param>
    /// <param name="meanRadiantTemperature">平均放射温度[C]</param>
    /// <param name="drybulbTemperature">乾球温度[C]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    public void UpdateBoundary
      (double velocity, double meanRadiantTemperature, double drybulbTemperature, double relativeHumidity)
    {
      double hr = MoistAir.GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
        (drybulbTemperature, relativeHumidity, ATMOSPHERIC_PRESSURE);
      double wvp = MoistAir.GetWaterVaporPartialPressureFromHumidityRatio(hr, ATMOSPHERIC_PRESSURE);
      foreach (Node bp in parts.Keys)
        parts[bp].updateBoundary(velocity, meanRadiantTemperature, drybulbTemperature, wvp);
    }

    /// <summary>境界条件を更新する</summary>
    /// <param name="node">更新する部位</param>
    /// <param name="velocity">気流速度[m/s]</param>
    /// <param name="meanRadiantTemperature">平均放射温度[C]</param>
    /// <param name="drybulbTemperature">乾球温度[C]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    public void UpdateBoundary
      (Node node, double velocity, double meanRadiantTemperature, double drybulbTemperature, double relativeHumidity)
    {
      double hr = MoistAir.GetHumidityRatioFromDryBulbTemperatureAndRelativeHumidity
        (drybulbTemperature, relativeHumidity, ATMOSPHERIC_PRESSURE);
      double wvp = MoistAir.GetWaterVaporPartialPressureFromHumidityRatio(hr, ATMOSPHERIC_PRESSURE);
      parts[node].updateBoundary(velocity, meanRadiantTemperature, drybulbTemperature, wvp);
    }

    /// <summary>着衣量[clo]を設定する</summary>
    /// <param name="node">部位</param>
    /// <param name="clo">着衣量[clo]</param>
    public void SetClothingIndex(Node node, double clo)
    { parts[node].setClothingIndex(clo); }

    /// <summary>物体に接触させる</summary>
    /// <param name="node">部位</param>
    /// <param name="temperature">物体の温度[C]</param>
    /// <param name="heatConductance">物体への熱コンダクタンス[W/K]</param>
    /// <param name="contactPortionRate">接触比率[-]</param>
    public void Contact(Node node, double temperature, double heatConductance, double contactPortionRate)
    { parts[node].contact(temperature, heatConductance, contactPortionRate); }

    #endregion

    #region 情報取得処理

    /// <summary>気流速度[m/s]を取得する</summary>
    /// <param name="node">体の部位</param>
    /// <returns>気流速度[m/s]</returns>
    public double GetVelocity(Node node) { return parts[node].velocity; }

    /// <summary>平均放射温度[C]を取得する</summary>
    /// <param name="node">体の部位</param>
    /// <returns>平均放射温度[C]</returns>
    public double GetMeanRadiantTemperature(Node node) { return parts[node].meanRadiantTemperature; }

    /// <summary>乾球温度[C]を取得する</summary>
    /// <param name="node">体の部位</param>
    /// <returns>乾球温度[C]</returns>
    public double GetDrybulbTemperature(Node node) { return parts[node].drybulbTemperature; }

    /// <summary>相当温度[C]を取得する</summary>
    /// <param name="node">体の部位</param>
    /// <returns>相当温度[C]</returns>
    public double GetOperatingTemperature(Node node) { return parts[node].operatingTemperature; }

    /// <summary>相対湿度[%]を取得する</summary>
    /// <param name="node">体の部位</param>
    /// <returns>相対湿度[%]</returns>
    public double GetRelativeHumidity(Node node)
    {
      double hr = MoistAir.GetHumidityRatioFromWaterVaporPartialPressure
        (parts[node].waterVaporPressure, ATMOSPHERIC_PRESSURE);
      return MoistAir.GetRelativeHumidityFromDryBulbTemperatureAndHumidityRatio
        (parts[node].drybulbTemperature, hr, ATMOSPHERIC_PRESSURE);
    }

    /// <summary>着衣量[clo]を取得する</summary>
    /// <param name="node">体の部位</param>
    /// <returns>着衣量[clo]</returns>
    public double GetClothingIndex(Node node) { return parts[node].clothingIndex; }

    /// <summary>組織間の熱コンダクタンス[W/K]を取得する</summary>
    /// <param name="node">部位</param>
    /// <param name="layer1">組織1</param>
    /// <param name="layer2">組織2</param>
    /// <returns>組織間の熱コンダクタンス[W/K]</returns>
    public double GetHeatConductance(Node node, Layer layer1, Layer layer2)
    { return parts[node].getHeatConductance(layer1, layer2); }

    /// <summary>代謝量[W]を取得する</summary>
    /// <returns>代謝量[W]</returns>
    public double GetMetabolicRate(Node node, Layer layer)
    {
      double bm = parts[node].basalMetabolicRate[layer];
      if (layer != Layer.Muscle) return bm;
      else return bm + parts[node].externalWork + parts[node].shiveringLoad;
    }

    /// <summary>体温[C]を取得する</summary>
    /// <param name="node">体の部位</param>
    /// <param name="layer">層</param>
    /// <returns>体温[C]</returns>
    public double GetTemperature(Node node, Layer layer) { return parts[node].temperatures[layer]; }

    /// <summary>血流[mL/s]を取得する</summary>
    /// <param name="node">体の部位</param>
    /// <param name="layer">層</param>
    /// <returns>血流[mL/s]</returns>
    public double GetBloodFlow(Node node, Layer layer) { return parts[node].bloodFlow[layer]; }

    /// <summary>平均皮膚温[C]を取得する</summary>
    /// <returns>平均皮膚温[C]</returns>
    public double GetAverageSkinTemperature()
    {
      double ctSum = 0;
      foreach (Node bp in parts.Keys)
      {
        bodyPart bPart = parts[bp];
        ctSum += bPart.temperatures[Layer.Skin] * bPart.surfaceArea;
      }
      return ctSum / SurfaceArea;
    }

    /// <summary>皮膚表面からの顕熱損失[W]を取得する</summary>
    /// <param name="node">部位</param>
    /// <returns>皮膚表面からの顕熱損失[W]</returns>
    public double GetSensibleHeatLoss(Node node) { return parts[node].getSensibleHeatLoss(); }

    /// <summary>皮膚表面からの潜熱損失[W]を取得する</summary>
    /// <param name="node">部位</param>
    /// <returns>皮膚表面からの潜熱損失[W]</returns>
    public double GetLatentHeatLoss(Node node) { return parts[node].latentHeatLoss; }

    #endregion

  }

  #region 読み取り専用interface

  /// <summary>読み取り専用多質点人体モデル</summary>
  public interface ImmutableMultiNodeModel
  {

    #region プロパティ

    /// <summary>体重[kg]を取得する</summary>
    double Weight { get; }

    /// <summary>身長[m]を取得する</summary>
    double Height { get; }

    /// <summary>年齢[-]を取得する</summary>
    double Age { get; }

    /// <summary>男性か否か</summary>
    bool IsMale { get; }

    /// <summary>立位か否か</summary>
    bool IsStanding { get; }

    /// <summary>体脂肪率[-]を取得する</summary>
    double FatPercentage { get; }

    /// <summary>体表面積[m2]を取得する</summary>
    double SurfaceArea { get; }

    /// <summary>中央血液溜まりの温度[C]を取得する</summary>
    double CentralBloodTemperature { get; }

    /// <summary>呼吸による熱損失[W]を取得する</summary>
    double HeatLossByBreathing { get; }

    /// <summary>全身の基礎血流[mL/s]を取得する</summary>
    double BasalBloodFlow { get; }

    /// <summary>全身の血流[mL/s]を取得する</summary>
    double BloodFlow { get; }

    /// <summary>全身の基礎代謝[W]を取得する</summary>
    double BasalMetabolicRate { get; }

    /// <summary>全身の代謝量[W]を取得する</summary>
    double MetabolicRate { get; }

    /// <summary>計算時間間隔[sec]を取得する</summary>
    double TimeStep { get; }

    #endregion

    #region メソッド

    /// <summary>気流速度[m/s]を取得する</summary>
    /// <param name="node">体の部位</param>
    /// <returns>気流速度[m/s]</returns>
    double GetVelocity(MultiNodeModel.Node node);

    /// <summary>平均放射温度[C]を取得する</summary>
    /// <param name="node">体の部位</param>
    /// <returns>平均放射温度[C]</returns>
    double GetMeanRadiantTemperature(MultiNodeModel.Node node);

    /// <summary>乾球温度[C]を取得する</summary>
    /// <param name="node">体の部位</param>
    /// <returns>乾球温度[C]</returns>
    double GetDrybulbTemperature(MultiNodeModel.Node node);

    /// <summary>相当温度[C]を取得する</summary>
    /// <param name="node">体の部位</param>
    /// <returns>相当温度[C]</returns>
    double GetOperatingTemperature(MultiNodeModel.Node node);

    /// <summary>相対湿度[%]を取得する</summary>
    /// <param name="node">体の部位</param>
    /// <returns>相対湿度[%]</returns>
    double GetRelativeHumidity(MultiNodeModel.Node node);

    /// <summary>着衣量[clo]を取得する</summary>
    /// <param name="node">体の部位</param>
    /// <returns>着衣量[clo]</returns>
    double GetClothingIndex(MultiNodeModel.Node node);

    /// <summary>組織間の熱コンダクタンス[W/K]を取得する</summary>
    /// <param name="node">部位</param>
    /// <param name="layer1">組織1</param>
    /// <param name="layer2">組織2</param>
    /// <returns>組織間の熱コンダクタンス[W/K]</returns>
    double GetHeatConductance
      (MultiNodeModel.Node node, MultiNodeModel.Layer layer1, MultiNodeModel.Layer layer2);

    /// <summary>代謝量[W]を取得する</summary>
    /// <returns>代謝量[W]</returns>
    double GetMetabolicRate(MultiNodeModel.Node node, MultiNodeModel.Layer layer);

    /// <summary>体温[C]を取得する</summary>
    /// <param name="node">体の部位</param>
    /// <param name="layer">層</param>
    /// <returns>体温[C]</returns>
    double GetTemperature(MultiNodeModel.Node node, MultiNodeModel.Layer layer);

    /// <summary>血流[mL/s]を取得する</summary>
    /// <param name="node">体の部位</param>
    /// <param name="layer">層</param>
    /// <returns>血流[mL/s]</returns>
    double GetBloodFlow(MultiNodeModel.Node node, MultiNodeModel.Layer layer);

    /// <summary>平均皮膚温[C]を取得する</summary>
    /// <returns>平均皮膚温[C]</returns>
    double GetAverageSkinTemperature();

    /// <summary>皮膚表面からの顕熱損失[W]を取得する</summary>
    /// <param name="node">部位</param>
    /// <returns>皮膚表面からの顕熱損失[W]</returns>
    double GetSensibleHeatLoss(MultiNodeModel.Node node);

    /// <summary>皮膚表面からの潜熱損失[W]を取得する</summary>
    /// <param name="node">部位</param>
    /// <returns>皮膚表面からの潜熱損失[W]</returns>
    double GetLatentHeatLoss(MultiNodeModel.Node node);

    #endregion

  }

  #endregion

}
