﻿/* ThermalComfort.cs
 * 
 * Copyright (C) 2014 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

using Popolo.ThermophysicalProperty;
using Popolo.Numerics;

namespace Popolo.HumanBody
{
  /// <summary>熱的快適性に関する静的メソッドを提供する</summary>
  public static class ThermalComfort
  {
    
    #region Constants

    /// <summary>代謝量換算係数[(W/m2)/met]</summary>
    private const double CONVERT_MET_TO_W = 58.15;

    /// <summary>絶対温度換算用数値</summary>
    private const double CONVERT_C_TO_K = 273;

    #endregion

    #region 列挙型

    /// <summary>仕事</summary>
    public enum Tasks
    {
      /// <summary>休息：睡眠</summary>
      Resting_Sleeping,
      /// <summary>休息：リクライニング</summary>
      Resting_Reclining,
      /// <summary>休息：座位安静</summary>
      Resting_Seated_Quiet,
      /// <summary>休息：立位安静</summary>
      Resting_Standing_Relaxed,
      /// <summary>歩行：0.9m/s</summary>
      Walking_Slow_09ms,
      /// <summary>歩行：1.2m/s</summary>
      Walking_Normal_12ms,
      /// <summary>歩行：1.8m/s</summary>
      Walking_Fast_18ms,
      /// <summary>オフィス作業：座位読み書き</summary>
      OfficeActivities_Seated_Reading_Writing,
      /// <summary>オフィス作業：タイピング</summary>
      OfficeActivities_Typing,
      /// <summary>オフィス作業：座位ファイル作業</summary>
      OfficeActivities_Filing_Seated,
      /// <summary>オフィス作業：立位ファイル作業</summary>
      OfficeActivities_Filing_Standing,
      /// <summary>オフィス作業：歩行</summary>
      OfficeActivities_Walking,
      /// <summary>オフィス作業：運搬・梱包</summary>
      OfficeActivities_Lifting_Packing,
      /// <summary>運転：自動車</summary>
      Driving_Automobile,
      /// <summary>運転：飛行機</summary>
      Driving_Aircraft_Routine,
      /// <summary>運転：飛行機着陸</summary>
      Driving_Aircraft_Instrument_Landing,
      /// <summary>運転：戦闘機</summary>
      Driving_Aircraft_Combat,
      /// <summary>運転：大型車両</summary>
      Driving_HeavyVehicle,
      /// <summary>その他：調理</summary>
      Other_Occupational_Cooking,
      /// <summary>その他：清掃</summary>
      Other_Occupational_HouseCleaning,
      /// <summary>その他：激しい着席中手作業</summary>
      Other_Occupational_Seated_HeavyLimbMovement,
      /// <summary>その他：のこ引き</summary>
      Other_Occupational_MachineWork_Sawing,
      /// <summary>その他：軽い機械作業</summary>
      Other_Occupational_MachineWork_Light,
      /// <summary>その他：激しい機械作業</summary>
      Other_Occupational_MachineWork_Heavy,
      /// <summary>その他：50kg程度の荷物運搬作業</summary>
      Other_Occupational_Handling50kgBags,
      /// <summary>その他：シャベルによる掘削作業</summary>
      Other_Occupational_PickAndShovelWork,
      /// <summary>その他：ダンス</summary>
      Other_Leisure_Dancing,
      /// <summary>その他：運動</summary>
      Other_Leisure_Exercise,
      /// <summary>その他：テニス</summary>
      Other_Leisure_Tennes,
      /// <summary>その他：バスケットボール</summary>
      Other_Leisure_Basketball,
      /// <summary>その他：レスリング</summary>
      Other_Leisure_Wrestling
    }

    #endregion

    #region publicメソッド

    /// <summary>人体への熱負荷[W/m2]を計算する</summary>
    /// <param name="drybulbTemperature">乾球温度[C]</param>
    /// <param name="meanRadiantTemperature">平均放射温度[C]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <param name="relativeAirVelocity">相対気流速度[m/s]</param>
    /// <param name="clothing">着衣量[clo]</param>
    /// <param name="metabolicRate">代謝量[met]</param>
    /// <param name="externalWork">外部仕事量[met]</param>
    /// <returns>人体への熱負荷[W/m2]</returns>
    public static double GetThermalLoad
      (double drybulbTemperature, double meanRadiantTemperature, double relativeHumidity, 
      double relativeAirVelocity, double clothing, double metabolicRate, double externalWork)
    {
      double dbtA = CONVERT_C_TO_K + drybulbTemperature;
      double mrtA = CONVERT_C_TO_K + meanRadiantTemperature;

      //周囲の水蒸気分圧[kPa]の計算
      double pa = relativeHumidity / 100d * Water.GetSaturationPressure(drybulbTemperature);

      //代謝量[W/m2]の計算
      double m = metabolicRate * CONVERT_MET_TO_W;
      double mw = m - externalWork * CONVERT_MET_TO_W;

      //着衣の熱抵抗[m2K/W]の計算
      double rcl = 0.155 * clothing;
      //着衣面積率[-]の計算
      double fcl;
      if (rcl < 0.078) fcl = 1.0 + 1.29 * rcl;
      else fcl = 1.05 + 0.645 * rcl;

      //強制対流による対流熱伝達率[W/(m2K)]
      double hcf = 12.1 * Math.Sqrt(relativeAirVelocity);

      //着衣表面温度の反復計算
      double tcla = dbtA + (35.5 - drybulbTemperature) / (3.5 * rcl + 0.1);//初期値
      double p1 = rcl * fcl;
      double p2 = p1 * 3.96;
      double p3 = p1 * 100;
      double p4 = p1 * dbtA;
      double p5 = 308.7 - 0.028 * mw + p2 * Math.Pow(mrtA / 100.0, 4);
      double xn = tcla / 100.0;
      double xf = xn;
      double hc; //対流熱伝達率[W/(m2K)]
      int iterNum = 0;
      while (true)
      {
        if (150 < iterNum) throw new Exception("ThermalComofrt Class iteration error");
        xf = (xf + xn) / 2d;
        //対流熱伝達率[W/(m2K)]を更新
        hc = Math.Max(hcf, 2.38 * Math.Pow(Math.Abs(100.0 * xn - dbtA), 0.25));
        //状態値更新
        xn = (p5 + p4 * hc - p2 * Math.Pow(xf, 4)) / (100d + p3 * hc);
        //収束判定
        if (Math.Abs(xn - xf) < 0.00015) break;
        iterNum++;
      }
      //着衣表面温度[C]
      double tcl = 100.0 * xn - CONVERT_C_TO_K;

      //人体熱負荷の計算

      double ediff = 3.05 * (5.733 - 0.00699 * mw - pa);      //皮膚表面からの潜熱損失[W/m2]
      double esw = 0.42 * Math.Max(0, mw - CONVERT_MET_TO_W); //発汗による潜熱損失[W/m2]
      double lres = 0.017 * m * (5.867 - pa);                 //呼吸による潜熱損失[W/m2]
      double dres = 0.0014 * m * (34.0 - drybulbTemperature); //呼吸による顕熱損失[W/m2]
      double r = 3.96 * fcl * (Math.Pow(xn, 4) - Math.Pow(mrtA / 100.0, 4)); //着衣表面からの放射熱損失[W/m2]      
      double c = fcl * hc * (tcl - drybulbTemperature);       //着衣表面からの対流熱損失[W/m2]

      //集計
      return mw - (ediff + esw + lres + dres + r + c);
    }

    /// <summary>PMV値[-]を計算する</summary>
    /// <param name="drybulbTemperature">乾球温度[CDB]</param>
    /// <param name="meanRadiantTemperature">平均放射温度[C]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <param name="relativeAirVelocity">相対気流速度[m/s]</param>
    /// <param name="clothing">着衣量[clo]</param>
    /// <param name="metabolicRate">代謝量[met]</param>
    /// <param name="externalWork">外部仕事量[met]</param>
    /// <returns>PMV値[-]</returns>
    public static double GetPMV
      (double drybulbTemperature, double meanRadiantTemperature, double relativeHumidity, 
      double relativeAirVelocity, double clothing, double metabolicRate, double externalWork)
    {
      double thermalLoad = GetThermalLoad(drybulbTemperature, meanRadiantTemperature, 
        relativeHumidity, relativeAirVelocity, clothing, metabolicRate, externalWork);
      return GetPMV(metabolicRate, thermalLoad);
    }

    /// <summary>PMV値[-]を計算する</summary>
    /// <param name="metabolicRate">代-謝量[met]</param>
    /// <param name="thermalLoad">人体への熱負荷[W/m2]</param>
    /// <returns>PMV値[-]</returns>
    public static double GetPMV(double metabolicRate, double thermalLoad)
    {
      double m = metabolicRate * CONVERT_MET_TO_W;
      return (0.303 * Math.Exp(-0.036 * m) + 0.028) * thermalLoad;
    }

    /// <summary>PPD値[%]を計算する</summary>
    /// <param name="pmv">PMV値</param>
    /// <returns>PPD値[%]</returns>
    public static double GetPPD(double pmv)
    {
      double p2 = pmv * pmv;
      return 100.0 - 95.0 * Math.Exp(-0.03353 * p2 * p2 - 0.2179 * p2);
    }

    /// <summary>PMV値[-]から乾球温度[C]を求める</summary>
    /// <param name="pmv">PMV値[-]</param>
    /// <param name="meanRadiantTemperature">平均放射温度[C]</param>
    /// <param name="relativeAirVelocity">気流速度[m/s]</param>
    /// <param name="relativeHumidity">相対湿度[%]</param>
    /// <param name="clothing">着衣量[clo]</param>
    /// <param name="metabolicRate">代謝量[met]</param>
    /// <param name="externalWork">外部仕事量[met]</param>
    /// <returns>乾球温度[C]</returns>
    public static double GetDrybulbTemperature
      (double pmv, double meanRadiantTemperature, double relativeHumidity, double relativeAirVelocity, 
      double clothing, double metabolicRate, double externalWork)
    {
      //乾球温度上下限値
      double MAXDB = 50;
      double MINDB = -10;

      //上下限値確認
      double pmvh = GetPMV(MAXDB, meanRadiantTemperature, relativeHumidity,
        relativeAirVelocity, clothing, metabolicRate, externalWork);
      if (pmvh < pmv) return MAXDB;
      double pmvl = GetPMV(MINDB, meanRadiantTemperature, relativeHumidity,
        relativeAirVelocity, clothing, metabolicRate, externalWork);
      if (pmv < pmvl) return MINDB;

      //ニュートン・ラフソン法で収束計算
      Roots.ErrorFunction eFnc = delegate (double dbTemp)
      {
        //誤差を評価
        double pmv2 = GetPMV(dbTemp, meanRadiantTemperature, relativeHumidity,
          relativeAirVelocity, clothing, metabolicRate, externalWork);
        return pmv2 - pmv;
      };
      return Roots.Newton(eFnc, 26, 0.0001, 0.01, 0.01, 20);
    }

    /// <summary>PMV値[-]から相対湿度[%]を求める</summary>
    /// <param name="pmv">PMV値[-]</param>
    /// <param name="dryBulbTemperature">乾球温度[CDB]</param>
    /// <param name="meanRadiantTemperature">平均放射温度[C]</param>
    /// <param name="relativeAirVelocity">気流速度[m/s]</param>
    /// <param name="clothing">着衣量[clo]</param>
    /// <param name="metabolicRate">代謝量[met]</param>
    /// <param name="externalWork">外部仕事量[met]</param>
    /// <returns>計算成功の真偽</returns>
    public static double GetRelativeHumidity
      (double pmv, double dryBulbTemperature, double meanRadiantTemperature, double relativeAirVelocity, 
      double clothing, double metabolicRate, double externalWork)
    {
      //上下限値確認
      double pmvl = GetPMV(dryBulbTemperature, meanRadiantTemperature, 0,
        relativeAirVelocity, clothing, metabolicRate, externalWork);
      if (pmv < pmvl) return 0;
      double pmvh = GetPMV(dryBulbTemperature, meanRadiantTemperature, 100,
        relativeAirVelocity, clothing, metabolicRate, externalWork);
      if (pmvh < pmv) return 100;

      //ニュートン・ラフソン法で収束計算
      Roots.ErrorFunction eFnc = delegate (double rHumid)
      {
        //誤差を評価
        double pmv2 = GetPMV(dryBulbTemperature, meanRadiantTemperature, rHumid,
          relativeAirVelocity, clothing, metabolicRate, externalWork);
        return pmv2 - pmv;
      };
      return Roots.Newton(eFnc, 40, 0.0001, 0.01, 0.01, 20);
    }

    /// <summary>仕事に対する代謝量[met]を返す</summary>
    /// <param name="task">仕事種類</param>
    /// <returns>代謝量[met]</returns>
    public static double GetMet(Tasks task)
    {
      switch (task)
      {
        case Tasks.Driving_Aircraft_Combat:
          return 2.4;
        case Tasks.Driving_Aircraft_Instrument_Landing:
          return 1.8;
        case Tasks.Driving_Aircraft_Routine:
          return 1.2;
        case Tasks.Driving_Automobile:
          return 1.5;
        case Tasks.Driving_HeavyVehicle:
          return 3.2;
        case Tasks.OfficeActivities_Filing_Seated:
          return 1.2;
        case Tasks.OfficeActivities_Filing_Standing:
          return 1.4;
        case Tasks.OfficeActivities_Lifting_Packing:
          return 2.1;
        case Tasks.OfficeActivities_Seated_Reading_Writing:
          return 1.0;
        case Tasks.OfficeActivities_Typing:
          return 1.1;
        case Tasks.OfficeActivities_Walking:
          return 1.7;
        case Tasks.Other_Leisure_Basketball:
          return 5.8;
        case Tasks.Other_Leisure_Dancing:
          return 3.4;
        case Tasks.Other_Leisure_Exercise:
          return 3.5;
        case Tasks.Other_Leisure_Tennes:
          return 3.8;
        case Tasks.Other_Leisure_Wrestling:
          return 7.8;
        case Tasks.Other_Occupational_Cooking:
          return 1.8;
        case Tasks.Other_Occupational_Handling50kgBags:
          return 4.0;
        case Tasks.Other_Occupational_HouseCleaning:
          return 2.7;
        case Tasks.Other_Occupational_MachineWork_Heavy:
          return 4.0;
        case Tasks.Other_Occupational_MachineWork_Light:
          return 1.8;
        case Tasks.Other_Occupational_MachineWork_Sawing:
          return 2.2;
        case Tasks.Other_Occupational_PickAndShovelWork:
          return 4.4;
        case Tasks.Other_Occupational_Seated_HeavyLimbMovement:
          return 2.2;
        case Tasks.Resting_Reclining:
          return 0.8;
        case Tasks.Resting_Seated_Quiet:
          return 1.0;
        case Tasks.Resting_Sleeping:
          return 0.7;
        case Tasks.Resting_Standing_Relaxed:
          return 1.2;
        case Tasks.Walking_Fast_18ms:
          return 3.8;
        case Tasks.Walking_Normal_12ms:
          return 2.6;
        case Tasks.Walking_Slow_09ms:
          return 2.0;
        default:
          throw new Exception("Thermal Comfort Class Tasks Error");
      }
    }

    #endregion

  }
}
