﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Popolo.Lighting
{
  /// <summary>形態係数計算クラス</summary>
  /// <remarks>
  /// Z
  /// ^    Y
  /// |   /
  /// |  /
  /// | /
  /// |/
  /// --------->X
  /// </remarks>
  public static class ViewFactor
  {



    #region privateメソッド

    /// <summary>面1から並行面2をみる形態係数を計算する</summary>
    /// <param name="width">面1の幅（Y方向）</param>
    /// <param name="height">面1の高さ（Z方向）</param>
    /// <param name="distance">並行面2までの距離（X方向）</param>
    /// <returns>面1から並行面2をみる形態係数</returns>
    private static double viewParallelRectangle2FromRectangle1(double width, double height, double distance)
    {
      if (distance == 0) return 1.0;
      if (width == 0 || height == 0 || distance == double.PositiveInfinity || distance < 0) return 0;

      double x = width / distance;
      double y = height / distance;
      double rx = Math.Sqrt(1 + x * x);
      double ry = Math.Sqrt(1 + y * y);
      double rxy1 = Math.Sqrt(1 + x * x + y * y);

      return (2 / Math.PI)
        * (ry * Math.Atan(x / ry) / y + rx / x * Math.Atan(y / rx)
        - Math.Atan(x) / y - Math.Atan(y) / x + Math.Log(rx * ry / rxy1) / (x * y));
    }

    /// <summary>面1から垂直面2をみる形態係数を計算する</summary>
    /// <param name="width">面1の幅（Y方向）</param>
    /// <param name="height">面1の高さ（Z方向）</param>
    /// <param name="depth">垂直面2の奥行き（X方向）</param>
    /// <returns>面1から垂直面2をみる形態係数</returns>
    private static double viewVerticalRectangle2FromRectangle1(double width, double height, double depth)
    {
      if (depth == double.PositiveInfinity) return 0.5;
      if (width == 0 || height == 0 || depth <= 0) return 0;

      double x = width / depth;
      double y = height / depth;
      double rx = Math.Sqrt(1 + x * x);
      double ry = Math.Sqrt(1 + y * y);
      double rxy1 = Math.Sqrt(1 + x * x + y * y);
      double rxy2 = Math.Sqrt(x * x + y * y);

      return (Math.Atan(x) / y - ry * Math.Atan(x / ry) / y + Math.Atan(x / y)
        + 0.5 / x / y * Math.Log(rxy1 * y / rxy2 / ry)
        + 0.5 * x / y * Math.Log(rxy2 * rx / rxy1 / x)) / Math.PI;
    }

    private static double viewVerticalRectangle2FromRectangle1(double width, double height, double depth, double deltaZ)
    {
      if (deltaZ < 0) throw new Exception("ViewFactor class: deltaZ < 0");

      double ff1 = viewVerticalRectangle2FromRectangle1(width, height + deltaZ, depth);
      if (deltaZ == 0) return ff1;

      double ff2 = viewVerticalRectangle2FromRectangle1(width, deltaZ, depth);
      return ((width * (height + deltaZ)) * ff1 - (width * deltaZ) * ff2) / (width * height);
    }

    private static double viewVerticalRectangle2FromRectangle1(double width, double height, double depth, double deltaZ, double deltaX)
    {
      if (deltaX < 0) throw new Exception("ViewFactor class: deltaX < 0");

      double ff1 = viewVerticalRectangle2FromRectangle1(width, height, depth, deltaZ);
      if (deltaX == 0) return ff1;

      double ff2 = viewVerticalRectangle2FromRectangle1(width, height, deltaX, deltaZ);
      return ff1 - ff2;
    }



    #endregion

  }
}
