﻿/* MultiLayerPerceptron.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;
using System.Text;

using Popolo.Numerics;

namespace Popolo.MachineLearning
{
  /// <summary>多層パーセプトロン型ニューラルネットワーク</summary>
  /// <remarks>【参考文献】涌井良幸, 涌井貞美:ディープラーニングがわかる数学入門</remarks>
  [Serializable]
  public class MultiLayerPerceptron
  {

    #region インスタンス変数・プロパティ

    /// <summary>素子出力配列[層番号][素子番号]</summary>
    private double[][] outputs;

    /// <summary>閾値配列[層番号][素子番号]</summary>
    private double[][] tHolds;

    /// <summary>ドロップアウト判定</summary>
    private bool[][] droppedOut;

    /// <summary>閾値による誤差微分値</summary>
    private double[][] dCdHs;

    /// <summary>結合荷重[接続先層番号][接続先素子番号][接続元素子番号]</summary>
    /// <remarks>
    /// C#では多段階配列の方が計算速度が早い。
    /// 入力に近い層が「接続元」、出力に近い層が「接続先」。
    /// </remarks>
    private double[][][] weights;

    /// <summary>結合荷重による誤差微分値</summary>
    private double[][][] dCdWs;

    /// <summary>一様乱数製造機</summary>
    private MersenneTwister uRnd;

    /// <summary>ドロップアウト率を設定・取得する</summary>
    public double DropOutRate { get; set; } = 0.5;

    /// <summary>入力（素子）数を取得する</summary>
    public uint InputNumber { get { return (uint)outputs[0].Length; } }

    /// <summary>出力（素子）数を取得する</summary>
    public uint OutputNumber { get { return (uint)outputs[outputs.Length - 1].Length; } }

    /// <summary>中間層の層数を取得する</summary>
    public uint HiddenLayerNumber { get { return (uint)outputs.Length - 2; } }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="inputNumber">入力層素子数</param>
    /// <param name="outputNumber">出力層素子数</param>
    /// <param name="hiddenNumbers">中間層素子数</param>
    public MultiLayerPerceptron(uint inputNumber, uint outputNumber, params uint[] hiddenNumbers)
    {
      //配列を用意
      makeArray(inputNumber, outputNumber, hiddenNumbers);

      //パラメータ初期化
      InitializeParameters((uint)DateTime.Now.Millisecond);
      uRnd = new MersenneTwister((uint)DateTime.Now.Millisecond);
    }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="data">初期化用文字列</param>
    public MultiLayerPerceptron(string data)
    {
      //中間層数と素子数の読み込み
      string[] strs = data.Split(',');
      uint[] hiddenNumbers = new uint[uint.Parse(strs[0])];
      uint inputNumber = uint.Parse(strs[1]);
      for (int i = 0; i < hiddenNumbers.Length; i++)
        hiddenNumbers[i] = uint.Parse(strs[i + 2]);
      uint outputNumber = uint.Parse(strs[hiddenNumbers.Length + 2]);

      //配列を用意
      makeArray(inputNumber, outputNumber, hiddenNumbers);
      int indx = hiddenNumbers.Length + 3;

      //閾値
      for (int i = 0; i < tHolds.Length; i++)
        for (int j = 0; j < tHolds[i].Length; j++)
        {
          tHolds[i][j] = double.Parse(strs[indx]);
          indx++;
        }

      //結合荷重
      for (int i = 0; i < weights.Length; i++)
        for (int j = 0; j < weights[i].Length; j++)
          for (int k = 0; k < weights[i][j].Length; k++)
          {
            weights[i][j][k] = double.Parse(strs[indx]);
            indx++;
          }

      //乱数製造機の作成
      uRnd = new MersenneTwister((uint)DateTime.Now.Millisecond);
    }

    /// <summary>配列を用意する</summary>
    /// <param name="inputNumber">入力層素子数</param>
    /// <param name="outputNumber">出力層素子数</param>
    /// <param name="hiddenNumbers">中間層素子数</param>
    private void makeArray(uint inputNumber, uint outputNumber, params uint[] hiddenNumbers)
    {
      //出力・閾値配列を用意
      outputs = new double[hiddenNumbers.Length + 2][];
      tHolds = new double[hiddenNumbers.Length + 1][];
      droppedOut = new bool[hiddenNumbers.Length][];
      dCdHs = new double[hiddenNumbers.Length + 1][];
      outputs[0] = new double[inputNumber];
      for (int i = 0; i < hiddenNumbers.Length; i++)
      {
        outputs[1 + i] = new double[hiddenNumbers[i]];
        tHolds[i] = new double[hiddenNumbers[i]];
        droppedOut[i] = new bool[hiddenNumbers[i]];
        dCdHs[i] = new double[hiddenNumbers[i]];
      }
      outputs[hiddenNumbers.Length + 1] = new double[outputNumber];
      tHolds[hiddenNumbers.Length] = new double[outputNumber];
      dCdHs[hiddenNumbers.Length] = new double[outputNumber];

      //結合荷重配列を用意
      weights = new double[outputs.Length - 1][][];
      dCdWs = new double[outputs.Length - 1][][];
      for (int i = 0; i < weights.Length; i++)
      {
        weights[i] = new double[outputs[i + 1].Length][];
        dCdWs[i] = new double[outputs[i + 1].Length][];
        for (int j = 0; j < weights[i].Length; j++)
        {
          weights[i][j] = new double[outputs[i].Length];
          dCdWs[i][j] = new double[outputs[i].Length];
        }
      }
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>出力を取得する</summary>
    /// <param name="number">出力番号</param>
    /// <returns>出力</returns>
    public double GetOutput(int number) { return outputs[outputs.Length - 1][number]; }

    /// <summary>中間層素子数を取得する</summary>
    /// <param name="hiddenLayerIndex">中間層の層番号</param>
    /// <returns>中間層素子数</returns>
    public uint GetHiddenNeuronNumber(int hiddenLayerIndex)
    { return (uint)outputs[hiddenLayerIndex + 1].Length; }

    /// <summary>学習を開始する</summary>
    /// <param name="inputs">入力[教師データ番号][入力番号]</param>
    /// <param name="answers">出力[教師データ番号][出力番号]</param>
    /// <param name="learnCoef">学習係数[-]</param>
    /// <param name="maxIteration">最大反復回数[回]</param>
    /// <param name="errTol">許容絶対誤差[-]（0で最大反復回数まで強制学習）</param>
    public double StartLearning(double[][] inputs, double[][] answers, double learnCoef, uint maxIteration, double errTol)
    {
      //誤差微分積算用配列を用意
      double[][] dCdHSum = new double[dCdHs.Length][];
      for (int i = 0; i < dCdHs.Length; i++) dCdHSum[i] = new double[dCdHs[i].Length];

      double aTOL = errTol * inputs.Length;
      double errorSum = 0;
      for (int iter = 0; iter < maxIteration; iter++)
      {
        errorSum = 0;

        //ドロップアウトリスト更新
        makeDropoutList();

        //パラメータ更新+誤差微分配列初期化
        for (int i = 0; i < dCdHSum.Length; i++)
        {
          for (int j = 0; j < dCdHSum[i].Length; j++)
          {
            //閾値
            if (iter != 0) tHolds[i][j] -= learnCoef * dCdHSum[i][j];
            dCdHSum[i][j] = 0;
            //結合荷重
            for (int k = 0; k < dCdWs[i][j].Length; k++)
            {
              if(iter != 0) weights[i][j][k] -= learnCoef * dCdWs[i][j][k];
              dCdWs[i][j][k] = 0;
            }
          }
        }

        for (int i = 0; i < inputs.Length; i++)
        {
          //前進伝播処理
          ForwardPropagate(inputs[i]);

          //誤差を積算
          for (int j = 0; j < outputs[outputs.Length - 1].Length; j++)
            errorSum += Math.Abs(outputs[outputs.Length - 1][j] - answers[i][j]);

          //逆伝播
          backPropagate(answers[i]);
          for (int j = 0; j < dCdHSum.Length; j++)
            for (int k = 0; k < dCdHSum[j].Length; k++)
              dCdHSum[j][k] += dCdHs[j][k];
        }
        //収束判定
        if (errorSum < aTOL) break;
      }

      //ドロップアウトリストを消去
      for (int i = 0; i < droppedOut.Length; i++)
        for (int j = 0; j < droppedOut[i].Length; j++)
          droppedOut[i][j] = false;
      return errorSum;
    }

    /// <summary>パラメータを初期化する</summary>
    /// <param name="seed">初期化用乱数シード</param>
    public void InitializeParameters(uint seed)
    {
      NormalRandom nRnd = new NormalRandom(seed);

      //閾値初期化
      for (int i = 0; i < tHolds.Length; i++)
        for (int j = 0; j < tHolds[i].Length; j++)
          tHolds[i][j] = nRnd.NextDouble_Standard();

      //結合荷重初期化
      for (int i = 0; i < weights.Length; i++)
        for (int j = 0; j < weights[i].Length; j++)
          for (int k = 0; k < weights[i][j].Length; k++)
            weights[i][j][k] = nRnd.NextDouble_Standard();
    }

    /// <summary>前進伝播させる</summary>
    /// <param name="inputs">入力値配列</param>
    public void ForwardPropagate(double[] inputs)
    {
      for (int i = 0; i < outputs[0].Length; i++) outputs[0][i] = inputs[i];

      for (int i = 0; i < weights.Length; i++)
      {
        for (int j = 0; j < weights[i].Length; j++)
        {
          double sum = 0;
          for (int k = 0; k < weights[i][j].Length; k++) sum += outputs[i][k] * weights[i][j][k];
          if (i == weights.Length - 1 || !droppedOut[i][j])
            outputs[i + 1][j] = 1 / (1 + Math.Exp(-(sum + tHolds[i][j])));  //活性化関数はシグモイド関数
          else outputs[i + 1][j] = 0;
        }
      }
    }

    /// <summary>このインスタンスの値をstringに変換する</summary>
    /// <returns>このインスタンスの値</returns>
    public override string ToString()
    {
      //中間層の層数と各層の素子数
      StringBuilder sBuilder = new StringBuilder(HiddenLayerNumber.ToString());
      sBuilder.Append("," + InputNumber);
      for (int i = 0; i < HiddenLayerNumber; i++)
        sBuilder.Append("," + GetHiddenNeuronNumber(i));
      sBuilder.Append("," + OutputNumber);

      //閾値
      for (int i = 0; i < tHolds.Length; i++)
        for (int j = 0; j < tHolds[i].Length; j++)
          sBuilder.Append("," + tHolds[i][j]);

      //結合荷重
      for (int i = 0; i < weights.Length; i++)
        for (int j = 0; j < weights[i].Length; j++)
          for (int k = 0; k < weights[i][j].Length; k++)
            sBuilder.Append("," + weights[i][j][k]);

      return sBuilder.ToString();
    }

    /// <summary>誤差を逆伝播させる</summary>
    /// <param name="answers">正解</param>
    private void backPropagate(double[] answers)
    {
      for (int i = outputs.Length - 1; 0 < i; i--)
      {
        for (int j = 0; j < outputs[i].Length; j++)
        {
          double erBf;
          //出力層
          if (i == outputs.Length - 1)
            erBf = outputs[i][j] - answers[j];
          //中間層
          else
          {
            erBf = 0;
            for (int k = 0; k < outputs[i + 1].Length; k++)
              erBf += dCdHs[i][k] * weights[i][k][j];
          }
          double bf;
          if (i != outputs.Length - 1 && droppedOut[i - 1][j]) bf = dCdHs[i - 1][j] = 0;
          else bf = dCdHs[i - 1][j] = erBf * (outputs[i][j] * (1 - outputs[i][j]));

          for (int k = 0; k < outputs[i - 1].Length; k++)
            dCdWs[i - 1][j][k] += bf * outputs[i - 1][k];
        }
      }
    }

    /// <summary>ドロップアウトリストを作成</summary>
    private void makeDropoutList()
    {
      for (int i = 0; i < droppedOut.Length; i++)
      {
        bool noNeuron = true;
        for (int j = 0; j < droppedOut[i].Length; j++)
        {
          droppedOut[i][j] = uRnd.NextDouble() <= DropOutRate;
          if (droppedOut[i][j]) noNeuron = false;
        }
        if (noNeuron) droppedOut[i][0] = false;
      }       
    }

    #endregion

  }
}
