﻿using System;

namespace Popolo.Numerics
{
  /// <summary>微分方程式ソルバ</summary>
  public static class ODESolver
  {

    #region delegate定義

    /// <summary>微分方程式</summary>
    /// <param name="t">時点</param>
    /// <param name="yt">現在値</param>
    /// <returns>微分値</returns>
    public delegate double DifferentialEquation(double t, double yt);

    /// <summary>強制終了判定関数</summary>
    /// <param name="t">時点</param>
    /// <param name="yt">現在値</param>
    /// <returns>強制終了する場合にはtrueを出力させる</returns>
    public delegate bool TerminateProcess(double t, double yt);

    #endregion

    #region static変数とコンストラクタ

    /// <summary>Runge-Kutta-Fehlberg法の係数</summary>
    private static readonly double[][] RKF45;

    /// <summary>静的コンストラクタ</summary>
    static ODESolver()
    {
      RKF45 = new double[6][];
      RKF45[0] = new double[] { 0 };
      RKF45[1] = new double[] { 1 / 4d, 1 / 4d };
      RKF45[2] = new double[] { 3 / 8d, 3 / 32d, 9 / 32d };
      RKF45[3] = new double[] { 12 / 13d, 1932 / 2197d, -7200 / 2197d, 7296 / 2197d };
      RKF45[4] = new double[] { 1, 439 / 216d, -8, 3680 / 513d, -845 / 4104d };
      RKF45[5] = new double[] { 0.5, -8 / 27d, 2, -3544 / 2565d, 1859 / 4104d, -11 / 40d };
    }

    #endregion

    #region Runge-Kutta法

    /// <summary>四次のRunge-Kutta法でdt秒後のy(t)を求める</summary>
    /// <param name="dEqn">微分方程式</param>
    /// <param name="dt">タイムステップ</param>
    /// <param name="t">時点</param>
    /// <param name="yt">現在値</param>
    /// <returns>タイムステップ経過後のy(t+dt)</returns>
    public static double SolveRK4(DifferentialEquation dEqn, double dt, double t, double yt)
    {
      double dt2 = 0.5 * dt;
      double k1 = dt * dEqn(t, yt);
      double k2 = dt * dEqn(t + dt2, yt + 0.5 * k1);
      double k3 = dt * dEqn(t + dt2, yt + 0.5 * k2);
      double k4 = dt * dEqn(t + dt, yt + k3);
      return yt + (k1 + 2 * (k2 + k3) + k4) / 6d;
    }

    /// <summary>Runge-Kutta-Fehlberg法で解を求める</summary>
    /// <param name="dEqn">微分方程式</param>
    /// <param name="tFnc">強制終了判定関数</param>
    /// <param name="dtMax">最大タイムステップ</param>
    /// <param name="t">開始時点</param>
    /// <param name="tend">終了時点</param>
    /// <param name="yt">初期値</param>
    /// <param name="errTol">許容誤差</param>
    /// <returns>タイムステップ終了時点のy(tend)</returns>
    /// <remarks>http://slpr.sakura.ne.jp/qp/runge-kutta-ex</remarks>
    public static double SolveRKF45(DifferentialEquation dEqn, TerminateProcess tFnc, double dtMax, double t, double tend, double yt, double errTol)
    {
      double dt = dtMax;
      double[] k = new double[6];

      while (true)
      {
        for (int i = 0; i < k.Length; i++)
        {
          double sm = yt;
          for (int j = 1; j < RKF45[i].Length; j++) sm += RKF45[i][j] * k[j - 1];
          k[i] = dt * dEqn(t + dt * RKF45[i][0], sm);
        }

        //誤差を評価して時点と状態値を更新
        double r = Math.Abs(1 / 360d * k[0] - 128 / 4275d * k[2] - 2197 / 75240d * k[3] + 1 / 50d * k[4] + 2 / 55d * k[5]) / dt;
        if (r < errTol)
        {
          yt += 25 / 216d * k[0] + 1408 / 2565d * k[2] + 2197 / 4104d * k[3] - 1 / 5d * k[4];
          t += dt;
        }
        if (tend <= t || tFnc(t, yt)) return yt;

        //タイムステップの調整
        double delta = Math.Pow(errTol / (2 * r), 0.25);
        if (delta <= 0.1) dt = 0.1 * dt;
        else if (4 <= delta) dt = 4 * dt;
        else dt = delta * dt;
        dt = Math.Min(dtMax, dt);
        if (tend < t + dt) dt = tend - t;
      }
    }

    #endregion
    
  }
}
