﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Popolo.ThermophysicalProperty;

namespace Popolo.ThermalLoad
{

  /// <summary>天井裏または床下のエアチャンバ（対流熱伝達の詳細計算タイプ、Fは上側、Bは下側）</summary>
  /// <remarks>無限水平流体層の対流熱伝達計算式を使う</remarks>
  public class HorizontalAirChamber : WallLayer
  {

    #region 定数宣言

    /// <summary>物性の再計算を行う温度差[K]</summary>
    private const double RECALC_TMP = 0.1;

    /// <summary>チャンバ空気の絶対湿度[kg/kg]</summary>
    private const double HUMIDITY_RATIO = 0.010;

    /// <summary>大気圧[kPa]</summary>
    private const double ATM = 101.325;

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>前回物性更新時の表面温度[C]</summary>
    private double lstUpTmp, lstDwnTmp;

    /// <summary>放射熱伝達率[W/(m2K)]を取得する</summary>
    public double RadiativeHeatTransferCoefficient { private set; get; }

    /// <summary>対流熱伝達率[W/(m2K)]を取得する</summary>
    public double ConvectiveHeatTransferCoefficient { private set; get; }

    /// <summary>上側の放射率[-]を取得する</summary>
    public double UpperEmissivity { private set; get; } = 0.9;

    /// <summary>下側の放射率[-]を取得する</summary>
    public double LowerEmissivity { private set; get; } = 0.9;

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="name">名称</param>
    /// <param name="thickness">厚み[m]</param>
    /// <param name="upperEmissivity">上側放射率[-]</param>
    /// <param name="lowerEmissivity">下側放射率[-]</param>
    public HorizontalAirChamber(string name, double thickness, double upperEmissivity, double lowerEmissivity)
    {
      IsVariableProperties = true;
      Name = name;
      Thickness = thickness;
      UpperEmissivity = upperEmissivity;
      LowerEmissivity = lowerEmissivity;

      VolSpecificHeat = MoistAir.GetSpecificHeat(HUMIDITY_RATIO) * 1.2;
      HeatCapacity_B = HeatCapacity_F = 0.5 * VolSpecificHeat * Thickness * 1000;

      lstDwnTmp = lstUpTmp = 0;
      UpdateState(26, 26);
    }

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="name">名称</param>
    /// <param name="thickness">厚み[m]</param>
    public HorizontalAirChamber(string name, double thickness) 
      : this(name, thickness, 0.9, 0.9) { }

    #endregion

    /// <summary>壁層端部の温度に応じて状態値を更新する</summary>
    /// <param name="temperatureF">F側温度</param>
    /// <param name="temperatureB">B側温度</param>
    /// <returns>状態値変更の有無</returns>
    public override bool UpdateState(double temperatureF, double temperatureB)
    {
      const double RAY_LMT = 1708;  //限界レイリー数

      //前回の物性計算時と温度差がRECALC_TMP未満ならば更新しない（計算速度確保のため）
      if (Math.Abs(temperatureF - lstUpTmp) < RECALC_TMP && Math.Abs(temperatureB - lstDwnTmp) < RECALC_TMP)
        return false;
      lstUpTmp = temperatureF;
      lstDwnTmp = temperatureB;

      //平均空気温度
      double aveTemp = 0.5 * (temperatureF + temperatureB);
      ThermalConductivity = MoistAir.GetThermalConductivity(aveTemp);
      RadiativeHeatTransferCoefficient = 4 * UpperEmissivity * LowerEmissivity * Math.Pow(273.15 + aveTemp, 3) * 5.67e-8;  //放射熱伝達の線形近似式

      //温度が逆転している場合には対流が生じない
      if (temperatureB <= temperatureF)
        ConvectiveHeatTransferCoefficient = ThermalConductivity / Thickness;
      else
      {
        //無次元数の計算        
        double nu = MoistAir.GetDynamicViscosity(aveTemp, HUMIDITY_RATIO, ATM); //動粘性係数[m2/s]
        double alpha = MoistAir.GetThermalDiffusivity(aveTemp, HUMIDITY_RATIO, ATM);  //熱拡散係数[m2/s]
        double plandtl = nu / alpha;  //プラントル数[-]
        double beta = MoistAir.GetExpansionCoefficient(aveTemp);  //体積膨張率[1/K]
        double grashof = 9.8 * Math.Pow(Thickness, 3) * beta * (temperatureB - temperatureF) / Math.Pow(nu, 2);  //グラスホフ数[-]
        double rayleigh = plandtl * grashof;  //レイリー数[-]

        //対流が生じない場合
        if (rayleigh < RAY_LMT)
          ConvectiveHeatTransferCoefficient = ThermalConductivity / Thickness;
        else
        {
          //平均ヌセルト数から熱伝達率[W/m2K]を計算
          double fpr = Math.Pow(1 + Math.Pow(0.5 / plandtl, 9d / 16d), -16d / 9d);
          double nud = Math.Pow(Math.Pow(1d + 1.466 * (1.0 - RAY_LMT / rayleigh), 15) + Math.Pow(rayleigh * fpr / 1420d, 5), 1d / 15d);
          ConvectiveHeatTransferCoefficient = nud * ThermalConductivity / Thickness;
        }
      }

      HeatConductance = ConvectiveHeatTransferCoefficient + RadiativeHeatTransferCoefficient;
      return true;
    }

  }
}
