﻿/* MultiRooms.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;
using System.Collections.Generic;

using Popolo.Numerics.MatrixOperation;
using Popolo.Numerics;
using Popolo.Weather;

namespace Popolo.ThermalLoad
{
  /// <summary>多数室</summary>
  [Serializable]
  public class MultiRooms : ImmutableMultiRooms
  {

    #region 定数宣言

    /// <summary>湿り空気の定圧比熱[J/(kgK)]</summary>
    /// <remarks>24C, 50%, 9.5g/kg</remarks>
    private const double AIR_SPECIFICHEAT = 1005 + 1846 * 0.0095;

    /// <summary>放射熱伝達率計算用係数[W/(m2K)]</summary>
    private const double RAD_COEF = 4.0 * 5.67e-8 * 297.15 * 297.15 * 297.15;

    #endregion

    #region インスタンス変数

    /// <summary>ゾーンリスト</summary>
    private Zone[] zones;

    /// <summary>壁リスト</summary>
    private Wall[] walls;

    /// <summary>窓リスト</summary>
    private Window[] windows;

    /// <summary>形態係数更新の要否</summary>
    private bool needInitialize = true;

    /// <summary>ゾーン間換気[kg/s]</summary>
    private double[,] zoneVent;

    /// <summary>各室の形態係数</summary>
    private IMatrix[] formFactor;

    /// <summary>ゲブハルトの吸収率行列（長波長は規準化）</summary>
    /// <remarks>surfacesと同サイズの行列数</remarks>
    private double[,] gMatL, gMatS;

    /// <summary>室内に面する壁・窓表面リスト</summary>
    private wallWindowSurface[] surfaces;

    /// <summary>窓面からの短波長を優先的に割り振る床表面リスト</summary>
    private Dictionary<Window, wallWindowSurface> swDistFloor;

    /// <summary>窓面からの短波長を床に割り振る割合</summary>
    private Dictionary<Window, double> swDistRate;

    /// <summary>壁・窓表面裏側が境界条件か否か</summary>
    private bool[] isSFboundary;

    /// <summary>室内に面する壁・窓表面に入射する短波長放射[W/m2]リスト</summary>
    /// <remarks>surfacesと同サイズ。窓に関しては吸収日射熱取得の値を保持</remarks>
    private double[] radToSurf_S;

    /// <summary>各室の壁表面に入射する長波長放射[W/m2]リスト</summary>
    /// <remarks>RoomNumberと同サイズ</remarks>
    private double[] radToSurf_L;

    /// <summary>境界条件表面リスト</summary>
    private List<wallWindowSurface> bndSurfaces = new List<wallWindowSurface>();

    /// <summary>ガラスの短波長放射率リスト</summary>
    /// <remarks>windowsと同サイズ</remarks>
    private double[] wSWEmissivity;

    /// <summary>各室に属するゾーンの番号</summary>
    private List<int>[] rZones;

    /// <summary>壁・窓の通し番号。[室][ゾーン][壁・窓]</summary>
    private int[][][] wsNumber;

    /// <summary>顕熱平衡の予測計算中か否か</summary>
    private bool forecastingHeatTransfer = false;

    /// <summary>水分平衡の予測計算中か否か</summary>
    private bool forecastingWaterTransfer = false;

    /// <summary>ゾーン温湿度リスト（一時保存）</summary>
    private double[] zoneTemp, zoneHumid;

    /// <summary>顕熱平衡計算用行列</summary>
    private IMatrix matA, matAInv, matB, matD, matF, matI, matK, matKInv, matBf;
    private IVector vecC, vecEJ, vecTH, vecTWS;

    /// <summary>水分平衡計算用行列</summary>
    private IMatrix matAW, matCW, matCWInv;
    private IVector vecWH;

    #endregion

    #region プロパティ

    /// <summary>計算時間間隔[sec]</summary>
    private double timeStep = 3600;

    /// <summary>計算時間間隔[sec]を設定・取得する</summary>
    public double TimeStep
    {
      get { return timeStep; }
      set
      {
        for (int i = 0; i < walls.Length; i++) walls[i].TimeStep = value;
        timeStep = value;
      }
    }

    /// <summary>熱・水分同時移動モデルか否か</summary>
    public bool SolveMoistureTransferSimultaneously { get; private set; }

    /// <summary>現在の日時を設定する</summary>
    public DateTime CurrentDateTime { get; private set; }

    /// <summary>太陽を設定・取得する</summary>
    public ImmutableSun Sun { get; set; }

    /// <summary>室数を取得する</summary>
    public int RoomNumber { get; private set; }

    /// <summary>ゾーン数を取得する</summary>
    public int ZoneNumber { get; private set; }

    /// <summary>ゾーンを取得する</summary>
    public ImmutableZone[] Zones { get { return zones; } }

    /// <summary>壁を取得する</summary>
    public ImmutableWall[] Walls { get { return walls; } }

    /// <summary>窓を取得する</summary>
    public ImmutableWindow[] Windows { get { return windows; } }

    /// <summary>外気乾球温度[C]を取得する</summary>
    public double OutdoorTemperature { get; private set; }

    /// <summary>外気絶対湿度[kg/kg]を取得する</summary>
    public double OutdoorHumidityRatio { get; private set; }

    /// <summary>夜間放射[W/m2]を取得する</summary>
    public double NocturnalRadiation { get; private set; }

    /// <summary>地表面反射率[-]を設定・取得する</summary>
    public double Albedo { get; set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="rmNumber">部屋の数</param>
    /// <param name="zones">ゾーンリスト</param>
    /// <param name="walls">壁リスト</param>
    /// <param name="windows">窓リスト</param>
    public MultiRooms(int rmNumber, Zone[] zones, Wall[] walls, Window[] windows)
    {
      RoomNumber = rmNumber;
      ZoneNumber = zones.Length;
      this.walls = walls;
      this.windows = windows;
      this.zones = zones;
      formFactor = new IMatrix[RoomNumber];
      rZones = new List<int>[RoomNumber];
      radToSurf_L = new double[RoomNumber];
      wsNumber = new int[RoomNumber][][];
      zoneVent = new double[ZoneNumber, ZoneNumber];
      wSWEmissivity = new double[windows.Length];
      zoneTemp = new double[ZoneNumber];
      zoneHumid = new double[ZoneNumber];
      swDistFloor = new Dictionary<Window, wallWindowSurface>();
      swDistRate = new Dictionary<Window, double>();
      for (int i = 0; i < RoomNumber; i++) rZones[i] = new List<int>();
      for (int i = 0; i < ZoneNumber; i++)
      {
        AddZone(0, i);
        zones[i].MultiRoom = this;
        zones[i].Index = i;
      }

      //熱水分同時移動判定//水分透過壁が1つでもあれば
      foreach (Wall wl in walls)
      {
        if (wl.ComputeMoistureTransfer)
        {
          SolveMoistureTransferSimultaneously = true;
          break;
        }
      }
      int num = ZoneNumber;
      if (SolveMoistureTransferSimultaneously) num *= 2;
      matD = new Matrix(num, num);
      vecEJ = new Vector(num);
      matI = new Matrix(num, num);
      matK = new Matrix(num, num);
      matKInv = new Matrix(num, num);
      vecTH = new Vector(num);
      matAW = new Matrix(num, num);
      matCW = new Matrix(num, num);
      matCWInv = new Matrix(num, num);
      vecWH = new Vector(num);
    }

    #endregion

    #region 熱平衡（熱水分同時移動対応）に関する処理

    /// <summary>将来の熱平衡状態を予測する</summary>
    /// <remarks>
    /// 繰り返し呼び出すことが可能。
    /// 最後にFixHeatTransferを呼び出して確定させる
    /// </remarks>
    internal void ForecastHeatTransfer()
    {
      //準備計算実行
      prepareForHeatTransfer();

      //IJ行列作成処理
      makeIJMatrix();

      bool[] ctrl = new bool[matK.Columns];
      for (int i = 0; i < ctrl.Length; i++)
      {
        if (i < ZoneNumber) ctrl[i] = zones[i].TemperatureControlled;
        else ctrl[i] = zones[i - ZoneNumber].HumidityControlled;
      }

      matK.Initialize(0);
      for (int i = 0; i < matI.Rows; i++)
      {
        //K行列を作成
        for (int j = 0; j < matI.Columns; j++)
        {
          if (!ctrl[j]) matK[i, j] = matI[i, j];
          else if (ctrl[j] && i == j) matK[i, j] = -1;
        }

        //Lベクトルを作成
        vecTH[i] = vecEJ[i];
        for (int j = 0; j < matI.Columns; j++)
        {
          if (ctrl[j])
          {
            if (j < ZoneNumber) vecTH[i] -= matI[i, j] * zones[j].TemperatureSetpoint;
            else vecTH[i] -= matI[i, j] * zones[j - ZoneNumber].HumidityRatioSetpoint;
          }
        }
        if (!ctrl[i])
        {
          if (i < ZoneNumber) vecTH[i] += zones[i].HeatSupply;
          else vecTH[i] += zones[i - ZoneNumber].WaterSupply;
        }
      }

      //連立方程式を解く（温湿度を計算）
      LinearAlgebra.SolveLinearEquations(ref matK, ref vecTH);
      for (int i = 0; i < ZoneNumber; i++)
      {
        if (zones[i].TemperatureControlled)
        {
          zones[i].HeatSupply = vecTH[i];
          zones[i].Temperature = zones[i].TemperatureSetpoint;
        }
        else zones[i].Temperature = vecTH[i];
        if (SolveMoistureTransferSimultaneously)
        {
          if (zones[i].HumidityControlled)
          {
            zones[i].WaterSupply = vecTH[i + ZoneNumber];
            zones[i].HumidityRatio = zones[i].HumidityRatioSetpoint;
          }
          else zones[i].HumidityRatio = vecTH[i + ZoneNumber];
        }
      }
    }

    /// <summary>熱平衡状態を確定する</summary>
    internal void FixHeatTransfer()
    {
      forecastingHeatTransfer = false;

      //壁表面温度を計算
      for (int i = 0; i < ZoneNumber; i++)
      {
        vecTH[i] = zones[i].Temperature;
        if (SolveMoistureTransferSimultaneously) vecTH[i + ZoneNumber] = zones[i].HumidityRatio;
      }
      LinearAlgebra.Multiplicate(matB, vecTH, ref vecC, 1, 1);
      LinearAlgebra.Multiplicate(matAInv, vecC, ref vecTWS, 1, 0);

      //温湿度条件を設定
      for (int i = 0; i < RoomNumber; i++)
      {
        for (int j = 0; j < wsNumber[i].Length; j++)
        {
          for (int k = 0; k < wsNumber[i][j].Length; k++)
          {
            //相当温度
            wallWindowSurface ws1 = surfaces[wsNumber[i][j][k]];
            ws1.SolAirTemperature = 0;
            for (int m = 0; m < wsNumber[i].Length; m++)
            {
              for (int n = 0; n < wsNumber[i][m].Length; n++)
              {
                int s2 = surfaces[wsNumber[i][m][n]].Index;
                ws1.SolAirTemperature += gMatL[ws1.Index, s2] * vecTWS[s2];
              }
            }
            ws1.SolAirTemperature *= ws1.RadiativeCoefficient;
            ws1.SolAirTemperature += zones[rZones[i][j]].Temperature * ws1.ConvectiveCoefficient;
            ws1.SolAirTemperature += radToSurf_L[i];
            if (ws1.IsWall) ws1.SolAirTemperature += radToSurf_S[ws1.Index];
            ws1.SolAirTemperature /= ws1.FilmCoefficient;

            //絶対湿度
            if (SolveMoistureTransferSimultaneously) ws1.HumidityRatio = zones[rZones[i][j]].HumidityRatio;
          }
        }
      }
    }

    /// <summary>熱平衡状態を更新確定する</summary>
    internal void UpdateHeatTransfer()
    {
      ForecastHeatTransfer();
      FixHeatTransfer();
    }

    /// <summary>熱平衡計算用A,B行列を計算する</summary>
    private void makeABMatrix()
    {
      //AB行列の再計算の要否を確認
      bool needUpdateAB = false;
      for (int i = 0; i < walls.Length; i++)
      {
        if (walls[i].invMatrixUpdated)
        {
          needUpdateAB = true;
          break;
        }
      }
      if (!needUpdateAB) return;

      int nS = surfaces.Length;
      matA.Initialize(0);
      matB.Initialize(0);
      for (int i = 0; i < RoomNumber; i++)
      {
        for (int j = 0; j < wsNumber[i].Length; j++)
        {
          for (int k = 0; k < wsNumber[i][j].Length; k++)
          {
            int s1 = wsNumber[i][j][k];
            wallWindowSurface ws1 = surfaces[s1];
            wallWindowSurface ws1R = ws1.ReverseSideSurface;

            //行列Aを作成
            //同一室
            for (int m = 0; m < wsNumber[i].Length; m++)
            {
              for (int n = 0; n < wsNumber[i][m].Length; n++)
              {
                int s2 = wsNumber[i][m][n];
                if (s1 == s2)
                {
                  matA[s1, s2] += 1;
                  if (SolveMoistureTransferSimultaneously) matA[s1 + nS, s2 + nS] += 1;
                }
                else
                {
                  double bf = ws1.RadiativeRate * gMatL[s1, s2];
                  matA[s1, s2] += -ws1.FFS2 * bf;
                  if (SolveMoistureTransferSimultaneously) matA[s1 + nS, s2] += -ws1.FFS3 * bf;
                }
              }
            }
            //裏側室
            if (!isSFboundary[s1])
            {
              int rvRm = zones[ws1R.ZoneIndex].RoomIndex;
              for (int m = 0; m < wsNumber[rvRm].Length; m++)
              {
                for (int n = 0; n < wsNumber[rvRm][m].Length; n++)
                {
                  int s2 = wsNumber[rvRm][m][n];
                  if (ws1R.Index != s2)
                  {
                    double bf = ws1R.RadiativeRate * gMatL[ws1R.Index, s2];
                    matA[s1, s2] += -ws1.BFS2 * bf;
                    if (SolveMoistureTransferSimultaneously) matA[s1 + nS, s2] += -ws1.BFS3 * bf;
                  }
                }
              }
            }

            //行列Bを作成
            for (int q = 0; q < zones.Length; q++)
            {
              int nQ = ZoneNumber;
              if (rZones[i][j] == q)
              {
                matB[s1, q] += ws1.FFS2 * ws1.ConvectiveRate;
                if (SolveMoistureTransferSimultaneously)
                {
                  matB[s1, q + nQ] += ws1.FFL2;
                  matB[s1 + nS, q] += ws1.FFS3 * ws1.ConvectiveRate;
                  matB[s1 + nS, q + nQ] += ws1.FFL3;
                }
              }
              if (ws1R.ZoneIndex == q && !isSFboundary[s1])
              {
                matB[s1, q] += ws1.BFS2 * ws1R.ConvectiveRate;
                if (SolveMoistureTransferSimultaneously)
                {
                  matB[s1, q + nQ] += ws1.BFL2;
                  matB[s1 + nS, q] += ws1.BFS3 * ws1R.ConvectiveRate;
                  matB[s1 + nS, q + nQ] += ws1.BFL3;
                }
              }
            }
          }
        }
      }

      //Aの逆行列計算
      LinearAlgebra.GetInverse(ref matA, ref matAInv);
    }

    /// <summary>熱平衡計算用Cベクトルを計算する</summary>
    private void makeCVector()
    {
      int nS = surfaces.Length;
      for (int i = 0; i < RoomNumber; i++)
      {
        for (int j = 0; j < wsNumber[i].Length; j++)
        {
          for (int k = 0; k < wsNumber[i][j].Length; k++)
          {
            int s1 = wsNumber[i][j][k];
            wallWindowSurface ws1 = surfaces[s1];
            wallWindowSurface ws1R = ws1.ReverseSideSurface;

            //ベクトルCを作成
            double rdsl;
            if (ws1.IsWall) rdsl = (radToSurf_L[i] + radToSurf_S[s1]) / ws1.FilmCoefficient;
            else rdsl = radToSurf_L[i] / ws1.FilmCoefficient;
            vecC[s1] = ws1.IF2 + ws1.FFS2 * rdsl;
            if (SolveMoistureTransferSimultaneously) vecC[s1 + nS] = ws1.IF3 + ws1.FFS3 * rdsl;
            if (!isSFboundary[s1])
            {
              rdsl = (radToSurf_L[zones[ws1R.ZoneIndex].RoomIndex] + radToSurf_S[ws1R.Index]) / ws1R.FilmCoefficient;
              vecC[s1] += ws1.BFS2 * rdsl;
              if (SolveMoistureTransferSimultaneously) vecC[s1 + nS] += ws1.BFS3 * rdsl;
            }
            else
            {
              vecC[s1] += ws1.BFS2 * ws1R.SolAirTemperature + ws1.BFL2 * ws1R.HumidityRatio;
              if (SolveMoistureTransferSimultaneously)
                vecC[s1 + nS] += ws1.BFS3 * ws1R.SolAirTemperature + ws1.BFL3 * ws1R.HumidityRatio;
            }
          }
        }
      }
    }

    /// <summary>熱平衡計算用IJ行列を計算する</summary>
    private void makeIJMatrix()
    {
      int nS = surfaces.Length;
      int nQ = zones.Length;
      matD.Initialize(0);
      matF.Initialize(0);
      for (int q1 = 0; q1 < ZoneNumber; q1++)
      {
        Zone zq1 = zones[q1];
        double capSZN = (zq1.HeatCapacity + zq1.AirMass * AIR_SPECIFICHEAT) / TimeStep;
        double capLZN = (zq1.WaterCapacity + zq1.AirMass) / TimeStep;

        //行列Dを作成
        for (int q2 = 0; q2 < ZoneNumber; q2++)
        {
          if (q1 == q2)
          {
            double bf = zq1.VentilationRate + zq1.SupplyAirFlowRate + zq1.supplyAirFlowRate2;
            for (int q3 = 0; q3 < ZoneNumber; q3++)
              if (zoneVent[q1, q3] != 0) bf += zoneVent[q1, q3];
            matD[q1, q1] = bf * AIR_SPECIFICHEAT + capSZN;
            if (SolveMoistureTransferSimultaneously) matD[q1 + nQ, q1 + nQ] = bf + capLZN;
            for (int k = 0; k < zq1.Surfaces.Count; k++)
            {
              wallWindowSurface ws = zq1.Surfaces[k];
              matD[q1, q1] += ws.Area * ws.ConvectiveCoefficient;
              if (SolveMoistureTransferSimultaneously) matD[q1 + nQ, q1 + nQ] += ws.Area * ws.MoistureCoefficient;
            }
          }
          else
          {
            if (zoneVent[q1, q2] != 0)
            {
              matD[q1, q2] = -zoneVent[q1, q2] * AIR_SPECIFICHEAT;
              if (SolveMoistureTransferSimultaneously) matD[q1 + nQ, q2 + nQ] = -zoneVent[q1, q2];
            }
          }
        }

        //ベクトルEを作成
        vecEJ[q1] = capSZN * zq1.Temperature + AIR_SPECIFICHEAT
          * (zq1.VentilationRate * OutdoorTemperature + zq1.SupplyAirFlowRate * zq1.SupplyAirTemperature
          + zq1.supplyAirFlowRate2 * zq1.supplyAirTemperature2) + zq1.IntegrateConvectiveHeatgains();
        if (SolveMoistureTransferSimultaneously)
          vecEJ[q1 + nQ] = capLZN * zq1.HumidityRatio
            + zq1.VentilationRate * OutdoorHumidityRatio + zq1.SupplyAirFlowRate * zq1.SupplyAirHumidityRatio
            + zq1.supplyAirFlowRate2 * zq1.supplyAirHumidityRatio2 + zq1.IntegrateWaterGains();

        //行列Fを作成
        for (int k = 0; k < zq1.Surfaces.Count; k++)
        {
          wallWindowSurface ws = zq1.Surfaces[k];
          matF[q1, ws.Index] = ws.Area * ws.ConvectiveCoefficient;
          if (SolveMoistureTransferSimultaneously) matF[q1 + nQ, ws.Index + nS] = ws.Area * ws.MoistureCoefficient;
        }
      }

      //IJ行列作成処理
      LinearAlgebra.Multiplicate(matF, matAInv, ref matBf);
      LinearAlgebra.Multiplicate(matBf, matB, ref matI);
      LinearAlgebra.Add(matD, ref matI, 1, -1);
      LinearAlgebra.Multiplicate(matBf, vecC, ref vecEJ, 1, 1);
    }

    /// <summary>外気条件を設定する</summary>
    /// <remarks>窓面吸収日射計算後に呼び出す必要あり</remarks>
    private void setOutdoorAirState()
    {
      foreach (Window win in windows)
      {
        wallWindowSurface ws = win.OutsideSurface;
        double fs = win.OutsideIncline.ConfigurationFactorToSky;
        ws.SolAirTemperature = OutdoorTemperature
          + radToSurf_S[win.InsideSurface.Index] * win.GetResistance()
          - ws.LongWaveEmissivity * fs * NocturnalRadiation / ws.FilmCoefficient;
      }

      foreach (wallWindowSurface ws in bndSurfaces)
      {
        if (!ws.IsGroundWall)
        {
          if (ws.AdjacentSpaceFactor < 0)
          {
            double fs = ws.Incline.ConfigurationFactorToSky;
            ws.SolAirTemperature = OutdoorTemperature
              + (ws.ShortWaveEmissivity * ws.Incline.GetSolarIrradiance(Sun, Albedo)
              - ws.LongWaveEmissivity * fs * NocturnalRadiation) / ws.FilmCoefficient;
            ws.HumidityRatio = OutdoorHumidityRatio;
          }
          else
          {
            //隣室温度差係数を用いる場合
            double ftd = ws.AdjacentSpaceFactor;
            double tmp = zones[ws.ReverseSideSurface.ZoneIndex].Temperature;
            ws.SolAirTemperature = (1 - ftd) * tmp + ftd * OutdoorTemperature;
            ws.HumidityRatio = (1 - ftd) * tmp + ftd * OutdoorHumidityRatio;
          }
        }
      }
    }

    /// <summary>熱平衡予測のための準備計算を行う</summary>
    private void prepareForHeatTransfer()
    {
      //温湿度を復旧
      if (forecastingHeatTransfer) ResetAirState();
      else
      {
        //温湿度を一時保存
        for (int i = 0; i < ZoneNumber; i++)
        {
          zoneTemp[i] = zones[i].Temperature;
          zoneHumid[i] = zones[i].HumidityRatio;
        }

        //ゾーン・表面通し番号付与、形態係数行列初期化
        initialize();

        //窓の光学特性を更新、ゲブハルト吸収率行列初期化
        updateWindowOpticalProperties();

        //長波長・短波長放射を分配
        for (int i = 0; i < radToSurf_S.Length; i++) radToSurf_S[i] = 0;
        for (int i = 0; i < RoomNumber; i++) radToSurf_L[i] = 0;
        distributeShortwaveRad();
        distributeLongwaveRad();

        //屋外側相当温度を設定
        setOutdoorAirState();

        //ABC行列を計算
        makeABMatrix();
        makeCVector();
        forecastingHeatTransfer = true;
      }
    }

    /// <summary>ゾーンの温湿度を将来予測値から現在値に戻す</summary>
    internal void ResetAirState()
    {
      if (forecastingHeatTransfer)
      {
        for (int i = 0; i < ZoneNumber; i++)
        {
          zones[i].Temperature = zoneTemp[i];
          zones[i].HumidityRatio = zoneHumid[i];
        }
      }
    }

    #endregion

    #region 水分平衡に関する処理

    /// <summary>将来の水分衡状態を予測する</summary>
    /// <remarks>
    /// 繰り返し呼び出すことが可能。
    /// 最後にFixWaterTransferを呼び出して確定させる
    /// </remarks>
    internal void ForecastWaterTransfer()
    {
      if (forecastingWaterTransfer)
        for (int i = 0; i < ZoneNumber; i++) zones[i].HumidityRatio = zoneHumid[i];
      else
      {
        for (int i = 0; i < ZoneNumber; i++) zoneHumid[i] = zones[i].HumidityRatio;
        forecastingWaterTransfer = true;
      }

      //熱平衡を計算
      for (int q1 = 0; q1 < ZoneNumber; q1++)
      {
        Zone zq1 = zones[q1];
        double capLZN = (zq1.WaterCapacity + zq1.AirMass) / TimeStep;

        //行列AWを作成
        for (int q2 = 0; q2 < ZoneNumber; q2++)
        {
          if (q1 == q2)
          {
            matAW[q1, q1] = zq1.VentilationRate + zq1.SupplyAirFlowRate + zq1.supplyAirFlowRate2 + capLZN;
            for (int q3 = 0; q3 < ZoneNumber; q3++) matAW[q1, q1] += zoneVent[q1, q3];
          }
          else matAW[q1, q2] = -zoneVent[q1, q2];
        }
        //ベクトルBWを作成
        vecWH[q1] = capLZN * zq1.HumidityRatio
          + zq1.VentilationRate * OutdoorHumidityRatio + zq1.SupplyAirFlowRate * zq1.SupplyAirHumidityRatio
          + zq1.supplyAirFlowRate2 * zq1.supplyAirHumidityRatio2 + zq1.IntegrateWaterGains();
      }

      //潜熱平衡を計算
      matCW.Initialize(0);
      for (int i = 0; i < ZoneNumber; i++)
      {
        //CW行列を作成
        for (int j = 0; j < ZoneNumber; j++)
        {
          if (!zones[j].HumidityControlled) matCW[i, j] = matAW[i, j];
          else if (zones[j].HumidityControlled && (i == j)) matCW[i, j] = -1;
        }

        //DWベクトルを作成
        for (int j = 0; j < ZoneNumber; j++)
          if (zones[j].HumidityControlled) vecWH[i] -= matAW[i, j] * zones[j].HumidityRatioSetpoint;
        if (!zones[i].HumidityControlled) vecWH[i] += zones[i].WaterSupply;
      }

      //連立方程式を解く（湿度を計算）
      LinearAlgebra.SolveLinearEquations(ref matCW, ref vecWH);
      for (int i = 0; i < ZoneNumber; i++)
      {
        if (zones[i].HumidityControlled)
        {
          zones[i].WaterSupply = vecWH[i];
          zones[i].HumidityRatio = zones[i].HumidityRatioSetpoint;
        }
        else zones[i].HumidityRatio = vecWH[i];
      }
    }

    /// <summary>水分平衡状態を確定する</summary>
    internal void FixWaterTransfer()
    { forecastingWaterTransfer = false; }

    /// <summary>水分平衡状態を更新確定する</summary>
    internal void UpdateWaterTransfer()
    {
      ForecastWaterTransfer();
      FixWaterTransfer();
    }

    #endregion

    #region 初期化処理

    /// <summary>初期化処理</summary>
    private void initialize()
    {
      if (needInitialize)
      {
        makeSerialNumber(); //通し番号付与

        //形態係数の計算（自動推定）
        for (int i = 0; i < RoomNumber; i++)
        {
          List<double> area = new List<double>();
          for (int j = 0; j < wsNumber[i].Length; j++)
            for (int k = 0; k < wsNumber[i][j].Length; k++)
              area.Add(surfaces[wsNumber[i][j][k]].Area);

          formFactor[i] = computeFormFactor(area.ToArray());
          int wsn = area.Count;
        }
        //ゲブハルトの吸収率行列更新
        computeGebhartMatrix();
        needInitialize = false;
      }
    }

    /// <summary>壁・窓表面に通し番号を振る</summary>
    private void makeSerialNumber()
    {
      List<wallWindowSurface> sfs = new List<wallWindowSurface>();
      int wsIndex = 0;
      for (int i = 0; i < RoomNumber; i++)
      {
        wsNumber[i] = new int[rZones[i].Count][];
        for (int j = 0; j < wsNumber[i].Length; j++)
        {
          Zone zn = zones[rZones[i][j]];
          wsNumber[i][j] = new int[zn.Surfaces.Count];
          for (int k = 0; k < wsNumber[i][j].Length; k++)
          {
            sfs.Add(zn.Surfaces[k]);
            zn.Surfaces[k].Index = wsNumber[i][j][k] = wsIndex;
            wsIndex++;
          }
        }
      }
      surfaces = sfs.ToArray();
      int sNum = surfaces.Length;
      int zNum = ZoneNumber;
      isSFboundary = new bool[sNum];
      for (int i = 0; i < sfs.Count; i++) isSFboundary[i] = !sfs.Contains(sfs[i].ReverseSideSurface);
      radToSurf_S = new double[sNum];
      gMatL = new double[sNum, sNum];
      gMatS = new double[sNum, sNum];
      if (SolveMoistureTransferSimultaneously)
      {
        sNum *= 2;
        zNum *= 2;
      }
      matA = new Matrix(sNum, sNum);
      matAInv = new Matrix(sNum, sNum);
      matB = new Matrix(sNum, zNum);
      vecC = new Vector(sNum);
      matF = new Matrix(zNum, sNum);
      vecTWS = new Vector(sNum);
      matBf = new Matrix(zNum, sNum);
    }

    /// <summary>表面積から形態係数を推定する</summary>
    /// <param name="area">表面積リスト</param>
    /// <returns>形態係数行列</returns>
    private static IMatrix computeFormFactor(double[] area)
    {
      if (area.Length == 2)
      {
        IMatrix mat = new Matrix(2, 2);
        mat[0, 0] = mat[1, 1] = 0.0;
        mat[0, 1] = mat[1, 0] = 1.0;
        return mat;
      }

      //昇順に並べ替え
      int[] index = new int[area.Length];
      for (int i = 0; i < index.Length; i++) index[i] = i;
      Array.Sort(area, index);

      double[,] ff = new double[area.Length, area.Length];
      double sA = 0;
      int last = area.Length - 1;
      for (int i = 0; i < last; i++) sA += area[i];
      if (sA < area[last])
      {
        //最も大きい面積が他の合算よりも大きい場合
        ff[last, last] = 1.0;
        for (int i = 0; i < last; i++)
        {
          ff[i, last] = 1.0;
          ff[last, i] = area[i] / area[last];
          ff[last, last] -= ff[last, i];
        }
      }
      else
      {
        //仮想的な半径を収束計算
        double[] alpha = new double[area.Length];
        Roots.ErrorFunction eFnc = delegate (double r)
        {
          double sum = 0;
          for (int i = 0; i < area.Length; i++)
          {
            double bf = (area[i] / r);
            alpha[i] = Math.Acos(Math.Max(-1, 1 - 0.5 * bf * bf));
            sum += alpha[i];
          }
          return sum - 2 * Math.PI;
        };
        //面積和=円周と見立てて収束計算の初期値を決定
        double sumA = 0;
        for (int i = 0; i < area.Length; i++) sumA += area[i];
        double rad = Roots.NewtonBisection         (eFnc, 0.5 * (sumA / Math.PI), 0.0001, 0.01, 0.01, 10);
        rad = Roots.Newton(eFnc, rad, 0.00001, 0.00001, 0.00001, 10);

        //形態係数を計算
        for (int i = 0; i < area.Length - 1; i++)
        {
          ff[i, i] = 0;
          for (int j = i + 1; j < area.Length; j++)
          {
            double a1 = alpha[i];
            double a2 = alpha[j];
            double a3 = 0;
            for (int k = i + 1; k < j; k++) a3 += alpha[k];
            double d1 = 2 * rad * Math.Sin(a3 / 2);
            double d2 = 2 * rad * Math.Sin(Math.PI - (a1 + a2 + a3) / 2);
            double d3 = 2 * rad * Math.Sin((a2 + a3) / 2);
            double d4 = 2 * rad * Math.Sin((a1 + a3) / 2);
            ff[i, j] = (d3 + d4 - d1 - d2) / (2 * area[i]);
            ff[j, i] = ff[i, j] * area[i] / area[j];
          }
        }
      }

      //順序を戻す
      IMatrix ff2 = new Matrix(area.Length, area.Length);
      for (int i = 0; i < area.Length; i++)
        for (int j = 0; j < area.Length; j++)
          ff2[index[i], index[j]] = ff[i, j];

      return ff2;
    }

    /// <summary>ゲブハルト吸収率行列を計算する</summary>
    private void computeGebhartMatrix()
    {
      for (int i = 0; i < RoomNumber; i++)
      {
        //壁番号を保存
        List<int> wInd = new List<int>();
        for (int j = 0; j < rZones[i].Count; j++)
          for (int k = 0; k < wsNumber[i][j].Length; k++)
            wInd.Add(wsNumber[i][j][k]);

        int wsn = wInd.Count;
        IMatrix ffRhoL = new Matrix(wsn, wsn);
        IMatrix ffRhoS = new Matrix(wsn, wsn);
        IMatrix ffRhoInv = new Matrix(wsn, wsn);
        for (int j = 0; j < wsn; j++)
        {
          wallWindowSurface ws = surfaces[wInd[j]];
          for (int k = 0; k < wsn; k++)
          {
            ffRhoL[j, k] = -(1 - ws.LongWaveEmissivity) * formFactor[i][j, k];
            ffRhoS[j, k] = -(1 - ws.ShortWaveEmissivity) * formFactor[i][j, k];
            if (j == k)
            {
              ffRhoL[j, k]++;
              ffRhoS[j, k]++;
            }
          }
        }
        LinearAlgebra.GetInverse(ref ffRhoL, ref ffRhoInv);
        LinearAlgebra.Multiplicate(formFactor[i], ffRhoInv, ref ffRhoL);
        LinearAlgebra.GetInverse(ref ffRhoS, ref ffRhoInv);
        LinearAlgebra.Multiplicate(formFactor[i], ffRhoInv, ref ffRhoS);
        for (int j = 0; j < wsn; j++)
        {
          wallWindowSurface ws1 = surfaces[wInd[j]];
          for (int k = 0; k < wsn; k++)
          {
            wallWindowSurface ws2 = surfaces[wInd[k]];
            gMatL[wInd[j], wInd[k]] = ffRhoL[j, k] * ws2.LongWaveEmissivity;
            gMatS[wInd[j], wInd[k]] = ffRhoS[j, k] * ws2.ShortWaveEmissivity;
          }
          //長波長は基準化して放射熱伝達率を計算
          double bf = 1 - gMatL[wInd[j], wInd[j]];
          ws1.RadiativeCoefficient = bf * ws1.LongWaveEmissivity * RAD_COEF;
          for (int k = 0; k < wsn; k++)
          {
            if (j != k) gMatL[wInd[j], wInd[k]] /= bf;
            else gMatL[wInd[j], wInd[k]] = 0;
          }
        }
      }
    }

    #endregion

    #region 放射分配処理

    /// <summary>長波長放射を分配する</summary>
    private void distributeLongwaveRad()
    {
      for (int i = 0; i < RoomNumber; i++)
      {
        radToSurf_L[i] = 0;
        double saSum = 0;
        for (int j = 0; j < rZones[i].Count; j++)
        {
          radToSurf_L[i] += zones[rZones[i][j]].IntegrateRadiativeHeatGains();
          for (int k = 0; k < wsNumber[i][j].Length; k++)  saSum += surfaces[wsNumber[i][j][k]].Area;
        }
        radToSurf_L[i] /= saSum;
      }
    }

    /// <summary>窓の光学特性を更新</summary>
    private void updateWindowOpticalProperties()
    {
      bool needUpdateGebhartMatrix = false;
      for (int i = 0; i < windows.Length; i++)
      {
        windows[i].UpdateOpticalProperties(Sun);
        if (wSWEmissivity[i] != windows[i].ShortWaveEmissivityB)
        {
          wSWEmissivity[i] = windows[i].ShortWaveEmissivityB;
          needUpdateGebhartMatrix = true;
        }
      }
      if (needUpdateGebhartMatrix) computeGebhartMatrix();
    }

    /// <summary>短波長放射を分配する</summary>
    private void distributeShortwaveRad()
    {
      //窓面からの日射を各表面に分配
      for (int i = 0; i < RoomNumber; i++)
      {
        for (int j = 0; j < wsNumber[i].Length; j++)
        {
          for (int k = 0; k < wsNumber[i][j].Length; k++)
          {
            wallWindowSurface ws1 = surfaces[wsNumber[i][j][k]];
            if (!ws1.IsWall)
            {
              int indx1 = wsNumber[i][j][k];
              //窓からの透過・吸収日射を計算
              Window win = ws1.Window;
              ImmutableIncline inc = win.OutsideIncline;
              double dir = inc.GetDirectSolarIrradiance(Sun) * (1 - win.SunShade.GetShadowRate(Sun));
              double dif = inc.GetDiffuseSolarIrradiance(Sun, Albedo);
              radToSurf_S[indx1] += 
                dir * win.DirectSolarIncidentAbsorptance + dif * win.DiffuseSolarIncidentAbsorptance;
              //床に優先配分される短波長を計算
              wallWindowSurface flr = null;
              double flRate = 0.0;
              double dir2 = dir * win.DirectSolarIncidentTransmittance * win.Area;
              double radFromFloor = 0.0;
              if (swDistFloor.ContainsKey(win))
              {
                flr = swDistFloor[win];
                flRate = swDistRate[win];
                radToSurf_S[flr.Index] += dir2 * flRate * flr.ShortWaveEmissivity / flr.Area;
                radFromFloor = dir2 * flRate * (1.0 - flr.ShortWaveEmissivity);
                dir2 *= (1 - flRate);
              }
              double rad = dir2 + dif * win.DiffuseSolarIncidentTransmittance * win.Area;

              //同じ室に属する壁表面に分配
              if (0 < rad)
              {
                for (int j2 = 0; j2 < wsNumber[i].Length; j2++)
                {
                  for (int k2 = 0; k2 < wsNumber[i][j2].Length; k2++)
                  {
                    int indx2 = wsNumber[i][j2][k2];
                    wallWindowSurface wsf2 = surfaces[indx2];
                    double ibsw = gMatS[indx1, indx2] * rad;
                    //床面からの放射を加算
                    if (flr != null) ibsw += gMatS[flr.Index, indx2] * radFromFloor;
                    ibsw /= wsf2.Area;
                    if (!wsf2.IsWall)
                      ibsw *= wsf2.Window.DiffuseSolarIncidentAbsorptance
                        / (1 - wsf2.Window.DiffuseSolarIncidentReflectance);
                    radToSurf_S[indx2] += ibsw;
                  }
                }
              }
            }
          }
        }
      }
    }

    #endregion

    #region 境界条件設定処理

    /// <summary>外気条件を更新する</summary>
    /// <param name="dTime">現在の日時</param>
    /// <param name="sun">太陽</param>
    /// <param name="temperature">外気乾球温度[C]</param>
    /// <param name="humidityRatio">外気絶対湿度[kg/kg]</param>
    /// <param name="nocRadiation">夜間放射[W/m2]</param>
    internal void UpdateOutdoorCondition
      (DateTime dTime, ImmutableSun sun, double temperature, double humidityRatio, double nocRadiation)
    {
      CurrentDateTime = dTime;
      Sun = sun;
      OutdoorTemperature = temperature;
      OutdoorHumidityRatio = humidityRatio;
      NocturnalRadiation = nocRadiation;
    }

    /// <summary>地中温度[C]を設定する</summary>
    /// <param name="wallIndex">壁番号</param>
    /// <param name="isSideF">F側か否か</param>
    /// <param name="groundTemperature">地中温度[C]</param>
    public void SetGroundTemperature(int wallIndex, bool isSideF, double groundTemperature)
    {
      wallWindowSurface ws;
      if (isSideF) ws = walls[wallIndex].SurfaceF;
      else ws = walls[wallIndex].SurfaceB;
      if (bndSurfaces.Contains(ws)) ws.SolAirTemperature = groundTemperature;
    }

    /// <summary>ゾーン間換気量[kg/s]を設定する</summary>
    /// <param name="zoneIndex1">ゾーン番号1</param>
    /// <param name="zoneIndex2">ゾーン番号2</param>
    /// <param name="airMassFlowRate">ゾーン間換気量[kg/s]</param>
    public void SetCrossVentilation(int zoneIndex1, int zoneIndex2, double airMassFlowRate)
    { zoneVent[zoneIndex1, zoneIndex2] = zoneVent[zoneIndex2, zoneIndex1] = airMassFlowRate; }

    /// <summary>ゾーン間換気量[kg/s]を設定する</summary>
    /// <param name="zone1">ゾーン1</param>
    /// <param name="zone2">ゾーン2</param>
    /// <param name="airMassFlowRate">ゾーン間換気量[kg/s]</param>
    public void SetCrossVentilation(ImmutableZone zone1, ImmutableZone zone2, double airMassFlowRate)
    { SetCrossVentilation(Array.IndexOf(zones, zone1), Array.IndexOf(zones, zone2), airMassFlowRate); }

    /// <summary>ゾーン1からゾーン2へ向かう換気量[kg/s]を設定する</summary>
    /// <param name="zoneIndex1">ゾーン番号1</param>
    /// <param name="zoneIndex2">ゾーン番号2</param>
    /// <param name="airMassFlowRate">換気量[kg/s]</param>
    /// <remarks>他のゾーンを含め風量収支を合わせる必要あり</remarks>
    public void SetAirFlow(int zoneIndex1, int zoneIndex2, double airMassFlowRate)
    { zoneVent[zoneIndex1, zoneIndex2] = airMassFlowRate; }

    /// <summary>ゾーン1からゾーン2へ向かう換気量[kg/s]を設定する</summary>
    /// <param name="zone1">ゾーン1</param>
    /// <param name="zone2">ゾーン2</param>
    /// <param name="airMassFlowRate">換気量[kg/s]</param>
    /// <remarks>他のゾーンを含め風量収支を合わせる必要あり</remarks>
    public void SetAirFlow(ImmutableZone zone1, ImmutableZone zone2, double airMassFlowRate)
    { SetAirFlow(Array.IndexOf(zones, zone1), Array.IndexOf(zones, zone2), airMassFlowRate); }

    /// <summary>換気量[kg/s]を設定する</summary>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="ventilationRate">換気量[kg/s]</param>
    public void SetVentilationRate(int zoneIndex, double ventilationRate)
    { zones[zoneIndex].VentilationRate = ventilationRate; }

    /// <summary>換気量[kg/s]を設定する</summary>
    /// <param name="zone">ゾーン</param>
    /// <param name="ventilationRate">換気量[kg/s]</param>
    public void SetVentilationRate(ImmutableZone zone, double ventilationRate)
    { SetVentilationRate(Array.IndexOf(zones, zone), ventilationRate); }

    /// <summary>窓からの短波長（直達）を優先的に割り振る床を設定する</summary>
    /// <param name="windowIndex">窓番号</param>
    /// <param name="wallIndex">壁（床）番号</param>
    /// <param name="isSideF">F側か否か</param>
    /// <param name="distRate">短波長（直達）分配比率[-]</param>
    public void SetSWDistributionRateToFloor(int windowIndex, int wallIndex, bool isSideF, double distRate)
    {
      wallWindowSurface ws;
      if (isSideF) ws = walls[wallIndex].SurfaceF;
      else ws = walls[wallIndex].SurfaceB;
      swDistFloor[windows[windowIndex]] = ws;
      swDistRate[windows[windowIndex]] = distRate;
    }

    /// <summary>窓からの短波長（直達）を優先的に割り振る床を設定する</summary>
    /// <param name="window">窓</param>
    /// <param name="wall">壁（床）</param>
    /// <param name="isSideF">F側か否か</param>
    /// <param name="distRate">短波長（直達）分配比率[-]</param>
    public void SetSWDistributionRateToFloor(ImmutableWindow window, ImmutableWall wall, bool isSideF, double distRate)
    { SetSWDistributionRateToFloor(Array.IndexOf(windows, window), Array.IndexOf(walls, wall), isSideF, distRate); }

    /// <summary>送水条件を設定する</summary>
    /// <param name="wallIndex">壁（床）番号</param>
    /// <param name="mNumber">質点番号</param>
    /// <param name="flowRate">通水量[kg/s]</param>
    /// <param name="temperature">水温[C]</param>
    public void SetBuriedPipeWaterState(int wallIndex, int mNumber, double flowRate, double temperature)
    { walls[wallIndex].SetInletWater(mNumber, flowRate, temperature); }

    #endregion

    #region モデル作成処理

    /// <summary>壁とゾーンの温湿度を初期化する</summary>
    /// <param name="temperature">乾球温度[C]</param>
    /// <param name="humidityRatio">絶対湿度[kg/kg]</param>
    public void InitializeAirState(double temperature, double humidityRatio)
    {
      foreach (Zone zn in zones) zn.InitializeAirState(temperature, humidityRatio);
      foreach (Wall wl in walls) wl.Initialize(temperature);
    }

    /// <summary>室にゾーンを追加する</summary>
    /// <param name="roomIndex">室番号</param>
    /// <param name="zoneIndex">ゾーン番号</param>
    public void AddZone(int roomIndex, int zoneIndex)
    {
      needInitialize = true;
      for (int i = 0; i < RoomNumber; i++) rZones[i].Remove(zoneIndex);
      rZones[roomIndex].Add(zoneIndex);
      zones[zoneIndex].RoomIndex = roomIndex;
    }

    /// <summary>室にゾーンを追加する</summary>
    /// <param name="roomIndex">室番号</param>
    /// <param name="zone">ゾーン</param>
    public void AddZone(int roomIndex, ImmutableZone zone)
    { AddZone(roomIndex, Array.IndexOf(zones, zone)); }

    /// <summary>ゾーンに壁を追加する</summary>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="wallIndex">壁番号</param>
    /// <param name="isSideF">F側の壁表面か否か</param>
    public void AddWall(int zoneIndex, int wallIndex, bool isSideF)
    {
      needInitialize = true;
      wallWindowSurface sf;
      if (isSideF) sf = walls[wallIndex].SurfaceF;
      else sf = walls[wallIndex].SurfaceB;

      bndSurfaces.Remove(sf);
      for (int i = 0; i < ZoneNumber; i++) zones[i].Surfaces.Remove(sf);
      zones[zoneIndex].Surfaces.Add(sf);
      sf.ZoneIndex = zoneIndex;
    }

    /// <summary>ゾーンに壁を追加する</summary>
    /// <param name="zone">ゾーン</param>
    /// <param name="wall">壁</param>
    /// <param name="isSideF">F側の壁表面か否か</param>
    public void AddWall(ImmutableZone zone, ImmutableWall wall, bool isSideF)
    { AddWall(Array.IndexOf(zones, zone), Array.IndexOf(walls, wall), isSideF); }

    /// <summary>ゾーンに壁を追加する</summary>
    /// <param name="zoneFIndex">F側ゾーン番号</param>
    /// <param name="zoneBIndex">B側ゾーン番号</param>
    /// <param name="wallIndex">壁番号</param>
    public void AddWall(int zoneFIndex, int zoneBIndex, int wallIndex)
    {
      AddWall(zoneFIndex, wallIndex, true);
      AddWall(zoneBIndex, wallIndex, false);
    }

    /// <summary>ゾーンに壁を追加する</summary>
    /// <param name="zoneF">F側ゾーン</param>
    /// <param name="zoneB">B側ゾーン</param>
    /// <param name="wall">壁</param>
    public void AddWall(ImmutableZone zoneF, ImmutableZone zoneB, ImmutableWall wall)
    { AddWall(Array.IndexOf(zones, zoneF), Array.IndexOf(zones, zoneB), Array.IndexOf(walls, wall)); }

    /// <summary>屋外に壁を追加する</summary>
    /// <param name="wallIndex">壁番号</param>
    /// <param name="isSideF">F側か否か</param>
    /// <param name="incline">外表面傾斜面</param>
    public void SetOutsideWall(int wallIndex, bool isSideF, ImmutableIncline incline)
    {
      needInitialize = true;
      wallWindowSurface ws;
      if (isSideF) ws = walls[wallIndex].SurfaceF;
      else ws = walls[wallIndex].SurfaceB;

      ws.AdjacentSpaceFactor = -1.0;
      ws.Incline = incline;
      ws.ZoneIndex = -1;
      ws.IsGroundWall = false;
      for (int i = 0; i < ZoneNumber; i++) zones[i].Surfaces.Remove(ws);
      bndSurfaces.Add(ws);
    }

    /// <summary>屋外に壁を追加する</summary>
    /// <param name="wall">壁</param>
    /// <param name="isSideF">F側か否か</param>
    /// <param name="incline">外表面傾斜面</param>
    public void SetOutsideWall(ImmutableWall wall, bool isSideF, ImmutableIncline incline)
    { SetOutsideWall(Array.IndexOf(walls, wall), isSideF, incline); }

    /// <summary>地中壁を追加する</summary>
    /// <param name="wallIndex">壁番号</param>
    /// <param name="isSideF">F側か否か</param>
    /// <param name="groundWallConductance">土壌-壁表面間の熱コンダクタンス[W/(m2K)]</param>
    /// <remarks>地中壁表面を境界条件とする場合には熱コンダクタンスを大きくする</remarks>
    public void SetGroundWall(int wallIndex, bool isSideF, double groundWallConductance)
    {
      needInitialize = true;
      wallWindowSurface ws;
      if (isSideF) ws = walls[wallIndex].SurfaceF;
      else ws = walls[wallIndex].SurfaceB;

      ws.ZoneIndex = -1;
      ws.IsGroundWall = true;
      ws.ConvectiveCoefficient = groundWallConductance;
      ws.RadiativeCoefficient = 0;
      for (int i = 0; i < ZoneNumber; i++) zones[i].Surfaces.Remove(ws);
      bndSurfaces.Add(ws);
    }

    /// <summary>地中壁を追加する</summary>
    /// <param name="wall">壁</param>
    /// <param name="isSideF">F側か否か</param>
    /// <param name="groundWallConductance">土壌-壁表面間の熱コンダクタンス[W/(m2K)]</param>
    /// <remarks>地中壁表面を境界条件とする場合には熱コンダクタンスを大きくする</remarks>
    public void SetGroundWall(ImmutableWall wall, bool isSideF, double groundWallConductance)
    { SetGroundWall(Array.IndexOf(walls, wall), isSideF, groundWallConductance); }

    /// <summary>隣室温度差係数[-]を設定する</summary>
    /// <param name="wallIndex">壁番号</param>
    /// <param name="isSideF">F側か否か</param>
    /// <param name="adjacentSpaceFactor">隣室温度差係数[-]</param>
    public void UseAdjacentSpaceFactor(int wallIndex, bool isSideF, double adjacentSpaceFactor)
    {
      needInitialize = true;
      wallWindowSurface ws;
      if (isSideF) ws = walls[wallIndex].SurfaceF;
      else ws = walls[wallIndex].SurfaceB;

      ws.AdjacentSpaceFactor = adjacentSpaceFactor;
      ws.ZoneIndex = -1;
      ws.IsGroundWall = false;
      for (int i = 0; i < ZoneNumber; i++) zones[i].Surfaces.Remove(ws);
      bndSurfaces.Add(ws);
    }

    /// <summary>隣室温度差係数[-]を設定する</summary>
    /// <param name="wall">壁</param>
    /// <param name="isSideF">F側か否か</param>
    /// <param name="adjacentSpaceFactor">隣室温度差係数[-]</param>
    public void UseAdjacentSpaceFactor(ImmutableWall wall, bool isSideF, double adjacentSpaceFactor)
    { UseAdjacentSpaceFactor(Array.IndexOf(walls, wall), isSideF, adjacentSpaceFactor); }

    /// <summary>ゾーンに窓を追加する</summary>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="windowIndex">窓番号</param>
    public void AddWindow(int zoneIndex, int windowIndex)
    {
      needInitialize = true;
      Window win = windows[windowIndex];
      for (int i = 0; i < ZoneNumber; i++) zones[i].Surfaces.Remove(win.InsideSurface);
      zones[zoneIndex].Surfaces.Add(win.InsideSurface);
      win.InsideSurface.ZoneIndex = zoneIndex;
    }

    /// <summary>ゾーンに窓を追加する</summary>
    /// <param name="zone">ゾーン</param>
    /// <param name="window">窓</param>
    public void AddWindow(ImmutableZone zone, ImmutableWindow window)
    { AddWindow(Array.IndexOf(zones, zone), Array.IndexOf(windows, window)); }

    /// <summary>発熱要素を追加する</summary>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="heatGain">発熱要素</param>
    public void AddHeatGain(int zoneIndex, IHeatGain heatGain)
    { zones[zoneIndex].AddHeatGain(heatGain); }

    /// <summary>発熱要素を追加する</summary>
    /// <param name="zone">ゾーン</param>
    /// <param name="heatGain">発熱要素</param>
    public void AddHeatGain(ImmutableZone zone, IHeatGain heatGain)
    { AddHeatGain(Array.IndexOf(zones, zone), heatGain); }

    /// <summary>発熱要素を削除する</summary>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="heatGain">発熱要素</param>
    public void RemoveHeatGain(int zoneIndex, IHeatGain heatGain)
    { zones[zoneIndex].RemoveHeatGain(heatGain); }

    /// <summary>発熱要素を削除する</summary>
    /// <param name="zone">ゾーン</param>
    /// <param name="heatGain">発熱要素</param>
    public void RemoveHeatGain(ImmutableZone zone, IHeatGain heatGain)
    { RemoveHeatGain(Array.IndexOf(zones, zone), heatGain); }

    /// <summary>屋外側対流熱伝達率[W/(m2K)]を設定する</summary>
    /// <param name="convectiveCoefficient">屋外側対流熱伝達率[W/(m2K)]</param>
    public void SetOutsideConvectiveCoefficient(double convectiveCoefficient)
    {
      needInitialize = true;
      foreach (wallWindowSurface sf in bndSurfaces) sf.ConvectiveCoefficient = convectiveCoefficient;
    }

    /// <summary>対流熱伝達率[W/(m2K)]を設定する</summary>
    /// <param name="wallIndex">壁番号</param>
    /// <param name="isSideF">F側か否か</param>
    /// <param name="convectiveCoefficient">対流熱伝達率[W/(m2K)]</param>
    public void SetConvectiveCoefficient(int wallIndex, bool isSideF, double convectiveCoefficient)
    {
      if (isSideF) walls[wallIndex].ConvectiveCoefficientF = convectiveCoefficient;
      else walls[wallIndex].ConvectiveCoefficientB = convectiveCoefficient;
    }

    #endregion

    #region 制御関連の処理

    /// <summary>乾球温度[C]を制御する</summary>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="setpoint">乾球温度設定値[C]</param>
    public void ControlDrybulbTemperature(int zoneIndex, double setpoint)
    { zones[zoneIndex].ControlDrybulbTemperature(setpoint); }

    /// <summary>顕熱供給[W]を制御する</summary>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="heatSupply">顕熱供給[W]</param>
    public void ControlHeatSupply(int zoneIndex, double heatSupply)
    { zones[zoneIndex].ControlHeatSupply(heatSupply); }

    /// <summary>絶対湿度[kg/kg]を制御する</summary>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="setpoint">絶対湿度設定値[kg/kg]</param>
    public void ControlHumidityRatio(int zoneIndex, double setpoint)
    { zones[zoneIndex].ControlHumidityRatio(setpoint); }

    /// <summary>水分供給[kg/s]を制御する</summary>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="waterSupply">水分供給[kg/s]</param>
    public void ControlWaterSupply(int zoneIndex, double waterSupply)
    { zones[zoneIndex].ControlWaterSupply(waterSupply); }

    /// <summary>給気条件を設定する</summary>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="saTemperature">給気温度[C]</param>
    /// <param name="saHumidityRatio">給気絶対湿度[kg/kg]</param>
    /// <param name="saFlowRate">給気量[kg/s]</param>
    public void SetSupplyAir(int zoneIndex, double saTemperature, double saHumidityRatio, double saFlowRate)
    {
      zones[zoneIndex].SupplyAirTemperature = saTemperature;
      zones[zoneIndex].SupplyAirHumidityRatio = saHumidityRatio;
      zones[zoneIndex].SupplyAirFlowRate = saFlowRate;
    }

    /// <summary>給気条件2を設定する</summary>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="saTemperature">給気温度[C]</param>
    /// <param name="saHumidityRatio">給気絶対湿度[kg/kg]</param>
    /// <param name="saFlowRate">給気量[kg/s]</param>
    internal void setSupplyAir(int zoneIndex, double saTemperature, double saHumidityRatio, double saFlowRate)
    {
      zones[zoneIndex].supplyAirTemperature2 = saTemperature;
      zones[zoneIndex].supplyAirHumidityRatio2 = saHumidityRatio;
      zones[zoneIndex].supplyAirFlowRate2 = saFlowRate;
    }

    #endregion

    #region 空調能力設定関連の処理

    /// <summary>加熱能力[W]を設定する</summary>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="heatingCapacity">加熱能力[W]</param>
    public void SetHeatingCapacity(int zoneIndex, double heatingCapacity)
    { zones[zoneIndex].HeatingCapacity = heatingCapacity; }

    /// <summary>冷却能力[W]を設定する</summary>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="coolingCapacity">冷却能力[W]</param>
    public void SetCoolingCapacity(int zoneIndex, double coolingCapacity)
    { zones[zoneIndex].CoolingCapacity = coolingCapacity; }

    /// <summary>加湿能力[kg/s]を設定する</summary>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="humidifyingCapacity">加湿能力[kg/s]</param>
    public void SetHumidifyingCapacity(int zoneIndex, double humidifyingCapacity)
    { zones[zoneIndex].HumidifyingCapacity = humidifyingCapacity; }

    /// <summary>除湿能力[kg/s]を設定する</summary>
    /// <param name="zoneIndex">ゾーン番号</param>
    /// <param name="dehumidifyingCapacity">除湿能力[kg/s]</param>
    public void SetDehumidifyingCapacity(int zoneIndex, double dehumidifyingCapacity)
    { zones[zoneIndex].DehumidifyingCapacity = dehumidifyingCapacity; }

    #endregion

  }

  #region 読み取り専用多数室

  /// <summary>読み取り専用の多数室</summary>
  public interface ImmutableMultiRooms
  {
    /// <summary>計算時間間隔[sec]を取得する</summary>
    double TimeStep { get; }

    /// <summary>熱・水分同時移動モデルか否か</summary>
    bool SolveMoistureTransferSimultaneously { get; }

    /// <summary>現在の日時を取得する</summary>
    DateTime CurrentDateTime { get; }

    /// <summary>太陽を取得する</summary>
    ImmutableSun Sun { get; }

    /// <summary>室数を取得する</summary>
    int RoomNumber { get; }

    /// <summary>ゾーン数を取得する</summary>
    int ZoneNumber { get; }

    /// <summary>ゾーンを取得する</summary>
    ImmutableZone[] Zones { get; }

    /// <summary>壁を取得する</summary>
    ImmutableWall[] Walls { get; }

    /// <summary>窓を取得する</summary>
    ImmutableWindow[] Windows { get; }

    /// <summary>外気乾球温度[C]を取得する</summary>
    double OutdoorTemperature { get; }

    /// <summary>外気絶対湿度[kg/kg]を取得する</summary>
    double OutdoorHumidityRatio { get; }

    /// <summary>夜間放射[W/m2]を取得する</summary>
    double NocturnalRadiation { get; }

    /// <summary>地表面反射率[-]を取得する</summary>
    double Albedo { get; }

  }

  #endregion

}
