﻿/* VenetianBlind.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;

using Popolo.Numerics.MatrixOperation;

namespace Popolo.ThermalLoad
{
  /// <summary>ベネシャンブラインド</summary>
  /// <remarks>ISO 15099による</remarks>
  [Serializable]
  public class VenetianBlind: IShadingDevice
  {

    #region 定数宣言

    /// <summary>スラット分割数</summary>
    private const int SP_N = 5;

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>有効か否か</summary>
    private bool pullDowned = false;

    /// <summary>直達日射に対する特性更新が必要か否か</summary>
    private bool needUpdateDirectIrradianceProperties = true;

    /// <summary>逆行列の更新が必要か否か</summary>
    private bool needUpdateInverseMatrix = true;

    /// <summary>F側開口部からの拡散日射に対する光学特性（上向き・下向き）</summary>
    private double difTauF_U, difTauF_L, difRhoF_U, difRhoF_L;

    /// <summary>B側開口部からの拡散日射に対する光学特性（上向き・下向き）</summary>
    private double difTauB_U, difTauB_L, difRhoB_U, difRhoB_L;

    /// <summary>F側開口部からの直達日射に対する光学特性（上向き・下向き・直達）</summary>
    private double dirTauF_U, dirTauF_L, dirRhoF_U, dirRhoF_L, dirTauDir;

    /// <summary>スラットからの形態係数行列</summary>
    private double[,] vFacSlt = new double[SP_N + 2, SP_N + 2];

    /// <summary>F側開口部からの形態係数ベクトル</summary>
    private double[] vFacOF = new double[SP_N * 2 + 1];

    /// <summary>B側開口部からの形態係数ベクトル</summary>
    private double[] vFacOB = new double[SP_N * 2 + 1];

    /// <summary>行列</summary>
    private IMatrix iaMatrix = new Matrix(SP_N * 2, SP_N * 2);

    /// <summary>逆行列</summary>
    private IMatrix iaMatrixINV = new Matrix(SP_N * 2, SP_N * 2);

    /// <summary>基準化したスパン長さ[-]</summary>
    private double dRspan;

    /// <summary>見かけの太陽高度[radian]</summary>
    private double pAngle;

    /// <summary>スラット角[radian]</summary>
    private double sAngle;

    /// <summary>有効か否かを設定・取得する</summary>
    public bool Pulldowned
    {
      get { return pullDowned; }
      set
      {
        HasPropertyChanged = (pullDowned != value);
        pullDowned = value;
      }
    }

    /// <summary>光学特性に変化があるか否か</summary>
    public bool HasPropertyChanged { get; private set; }

    /// <summary>見かけの太陽高度[radian]を設定・取得する</summary>
    public double ProfileAngle
    {
      get { return pAngle; }
      set
      {
        if (pAngle == value) return;
        pAngle = value;
        needUpdateDirectIrradianceProperties = true;
        HasPropertyChanged = true;
      }
    }

    /// <summary>スラット角[radian]を設定・取得する（0で水平）</summary>
    public double SlatAngle
    {
      get { return sAngle; }
      set
      {
        if (sAngle == value) return;
        sAngle = value;
        needUpdateInverseMatrix = true;
        HasPropertyChanged = true;
      }
    }

    /// <summary>上側透過率[-]を取得する</summary>
    public double UpsideTransmittance { get; private set; }

    /// <summary>下側透過率[-]を取得する</summary>
    public double DownsideTransmittance { get; private set; }

    /// <summary>上側反射率[-]を取得する</summary>
    public double UpsideReflectance { get; private set; }

    /// <summary>下側反射率[-]を取得する</summary>
    public double DownsideReflectance { get; private set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="slatWidth">スラット幅</param>
    /// <param name="slatSpan">スラット間隔</param>
    /// <param name="upsideTransmittance">スラット上側表面透過率[-]</param>
    /// <param name="downsideTransmittance">スラット下側表面透過率[-]</param>
    /// <param name="upsideReflectance">スラット上側表面反射率[-]</param>
    /// <param name="downsideReflectance">スラット下側表面反射率[-]</param>
    /// <remarks>slatWidthとslatSpanの単位は同じであれば任意</remarks>
    public VenetianBlind(double slatWidth, double slatSpan, double upsideTransmittance,
      double downsideTransmittance, double upsideReflectance, double downsideReflectance)
    {
      dRspan = SP_N * slatSpan / slatWidth;
      UpsideTransmittance = upsideTransmittance;
      DownsideTransmittance = downsideTransmittance;
      UpsideReflectance = upsideReflectance;
      DownsideReflectance = downsideReflectance;
    }

    #endregion

    #region 光学特性計算関連の処理

    /// <summary>光学特性を計算する</summary>
    /// <param name="isDiffuseIrradianceProperties">拡散日射か否か</param>
    /// <param name="irradianceFromSideF">F側開口からの日射か否か</param>
    /// <param name="transmittance">透過率[-]</param>
    /// <param name="reflectance">反射率[-]</param>
    public void ComputeOpticalProperties(bool isDiffuseIrradianceProperties, bool irradianceFromSideF,
      out double transmittance, out double reflectance)
    {
      if (!Pulldowned)
      {
        transmittance = 1.0;
        reflectance = 0.0;
        return;
      }

      //必要に応じて逆行列を更新
      if (needUpdateInverseMatrix)
      {
        updateInverseMatrix();
        updateDirectIrradianceProperties();
        updateDiffuseIrradianceProperties();
        HasPropertyChanged = false;
      }

      //拡散日射に対する光学特性
      if (isDiffuseIrradianceProperties)
      {
        if (irradianceFromSideF)
        {
          transmittance = difTauF_U + difTauF_L;
          reflectance = difRhoF_U + difRhoF_L;
        }
        else
        {
          transmittance = difTauB_U + difTauB_L;
          reflectance = difRhoB_U + difRhoB_L;
        }
      }
      //直達日射に対する光学特性
      else
      {
        if (needUpdateDirectIrradianceProperties)
        {
          updateDirectIrradianceProperties();
          HasPropertyChanged = false;
        }

        if (irradianceFromSideF)
        {
          transmittance = dirTauF_U + dirTauF_L + dirTauDir;
          reflectance = dirRhoF_U + dirRhoF_L;
        }
        else
        {
          //F側と同じ値とする
          transmittance = dirTauF_U + dirTauF_L + dirTauDir;
          reflectance = dirRhoF_U + dirRhoF_L;
        }
      }
    }

    /// <summary>スラットから開口部への入射を計算する</summary>
    /// <param name="eVec">各スラットに入射する日射</param>
    /// <param name="sideF_U">F側開口への上向き入射</param>
    /// <param name="sideF_L">F側開口への下向き入射</param>
    /// <param name="sideB_U">B側開口への上向き入射</param>
    /// <param name="sideB_L">B側開口への下向き入射</param>
    private void computeRateToOpenings
      (IVector eVec, out double sideF_U, out double sideF_L, out double sideB_U, out double sideB_L)
    {
      double td1, td2, td3, td4, rd1, rd2, rd3, rd4;
      td1 = td2 = td3 = td4 = rd1 = rd2 = rd3 = rd4 = 0;
      for (int i = 0; i < SP_N; i++)
      {
        td1 += eVec[i] * vFacSlt[SP_N + 1, i];
        td2 += eVec[i + SP_N] * vFacSlt[SP_N + 1, i];
        td3 += eVec[i + SP_N] * vFacSlt[i, SP_N + 1];
        td4 += eVec[i] * vFacSlt[i, SP_N + 1];
        rd1 += eVec[i] * vFacSlt[SP_N, i];
        rd2 += eVec[i + SP_N] * vFacSlt[SP_N, i];
        rd3 += eVec[i + SP_N] * vFacSlt[i, SP_N];
        rd4 += eVec[i] * vFacSlt[i, SP_N];
      }
      sideB_U = td1 * UpsideTransmittance + td2 * DownsideReflectance;
      sideB_L = td3 * DownsideTransmittance + td4 * UpsideReflectance;
      sideF_U = rd1 * UpsideTransmittance + rd2 * DownsideReflectance;
      sideF_L = rd3 * DownsideTransmittance + rd4 * UpsideReflectance;
    }

    /// <summary>拡散日射に対する光学特性を計算する</summary>
    private void updateDiffuseIrradianceProperties()
    {
      IVector bVec = new Vector(SP_N * 2);
      IVector eVec = new Vector(SP_N * 2);

      //F側開口からの拡散日射に対する透過率・反射率
      for (int i = 0; i < bVec.Length; i++) bVec[i] = vFacOF[i];
      LinearAlgebra.Multiplicate(iaMatrixINV, bVec, ref eVec, 1, 0);
      computeRateToOpenings(eVec, out difRhoF_U, out difRhoF_L, out difTauF_U, out difTauF_L);
      difTauF_L += vFacOF[SP_N * 2];

      //B側開口からの拡散日射に対する透過率・反射率
      for (int i = 0; i < bVec.Length; i++) bVec[i] = vFacOB[i];
      LinearAlgebra.Multiplicate(iaMatrixINV, bVec, ref eVec, 1, 0);
      computeRateToOpenings(eVec, out difTauB_U, out difTauB_L, out difRhoB_U, out difRhoB_L);
      difTauB_L += vFacOB[SP_N * 2];
    }

    /// <summary>直達日射に対する光学特性を計算する</summary>
    private void updateDirectIrradianceProperties()
    {
      //スラットに直接入射する日射を計算
      IVector bVec = new Vector(SP_N * 2);
      IVector eVec = new Vector(SP_N * 2);

      if (-SlatAngle == ProfileAngle)
      {
        dirTauDir = 1.0;
        dirTauF_U = dirTauF_L = dirRhoF_U = dirRhoF_L = 0;
        return;
      }

      double dRrad = dRspan / (Math.Cos(SlatAngle) * Math.Abs(Math.Tan(SlatAngle) + Math.Tan(ProfileAngle)));
      int bf = 0;
      if (ProfileAngle < -SlatAngle) bf = SP_N;
      for (int i = 0; i < SP_N; i++) bVec[i + bf] = Math.Min(1, Math.Max(0, dRrad - i)) / dRrad;
      LinearAlgebra.Multiplicate(iaMatrixINV, bVec, ref eVec, 1, 0);

      //開口部へ向かう日射を計算
      computeRateToOpenings(eVec, out dirRhoF_U, out dirRhoF_L, out dirTauF_U, out dirTauF_L);
      dirTauDir = Math.Max(0, dRrad - SP_N) / dRrad;

      needUpdateDirectIrradianceProperties = false;
    }

    #endregion

    #region 逆行列関連の処理

    /// <summary>形態係数を更新する</summary>
    private void updateViewFactor()
    {
      //スラット間距離を更新
      double[] dn = new double[SP_N * 2 + 1];
      double sinPsi = Math.Sin(SlatAngle);
      for (int i = 0; i < dn.Length; i++)
      {
        int bf = (i - SP_N);
        dn[i] = Math.Sqrt(dRspan * (2 * bf * sinPsi + dRspan) + bf * bf);
      }

      //スラットからみた形態係数を更新
      double sum;
      double[] ff = new double[2 * SP_N - 1];
      for (int i = 0; i < ff.Length; i++) ff[i] = 0.5 * (dn[i] + dn[i + 2]) - dn[i + 1];
      for (int i = 0; i < SP_N; i++)
      {
        sum = 0;
        for (int j = 0; j < SP_N; j++)
        {
          vFacSlt[i, j] = ff[j - i + SP_N - 1];
          sum += vFacSlt[i, j];
        }
        vFacSlt[i, SP_N] = vFacSlt[SP_N + 1, SP_N - 1 - i] = 0.5 * (dn[SP_N - i] - dn[SP_N - 1 - i] + 1);
        vFacSlt[i, SP_N + 1] = vFacSlt[SP_N, SP_N - 1 - i] = 1 - (sum + vFacSlt[i, SP_N]);
      }

      //開口からみた形態係数を更新
      sum = 0;
      for (int i = 0; i < SP_N; i++)
      {
        vFacOF[i] = vFacOB[vFacOB.Length - 2 - i] = vFacSlt[i, SP_N] / dRspan;
        vFacOF[i + SP_N] = vFacOB[SP_N - 1 - i] = vFacSlt[SP_N, i] / dRspan;
        sum += vFacOF[i] + vFacOF[i + SP_N];
      }
      vFacOF[SP_N * 2] = vFacOB[SP_N * 2] = 1 - sum;
    }

    /// <summary>逆行列を更新する</summary>
    private void updateInverseMatrix()
    {
      //形態係数を更新
      updateViewFactor();

      for (int i = 0; i < SP_N; i++)
      {
        for (int j = 0; j < SP_N; j++)
        {
          double dlt = 0;
          if (i == j) dlt = 1;
          iaMatrix[i, j] = dlt - (UpsideTransmittance * vFacSlt[i, j]);
          iaMatrix[i, j + SP_N] = -(DownsideReflectance * vFacSlt[i, j]);
          iaMatrix[i + SP_N, j + SP_N] = dlt - (DownsideTransmittance * vFacSlt[j, i]);
          iaMatrix[i + SP_N, j] = -(UpsideReflectance * vFacSlt[j, i]);
        }
      }
      LinearAlgebra.GetInverse(ref iaMatrix, ref iaMatrixINV);

      needUpdateInverseMatrix = false;
    }

    #endregion

    #region 日照関連の処理

    /// <summary>光学特性を計算する</summary>
    /// <param name="diffuseDiffuseTransmittance_U">拡散光の上側透過率[-]</param>
    /// <param name="diffuseDiffuseTransmittance_L">拡散光の下側透過率[-]</param>
    /// <param name="directDiffuseTransmittance_U">直達光の上側拡散透過率[-]</param>
    /// <param name="directDiffuseTransmittance_L">直達光の下側拡散透過率[-]</param>
    /// <param name="directDirectTransmittanse">直達光の直達透過率[-]</param>
    public void ComputeOpticalProperties
      (out double diffuseDiffuseTransmittance_U, out double diffuseDiffuseTransmittance_L,
      out double directDiffuseTransmittance_U, out double directDiffuseTransmittance_L, out double directDirectTransmittanse)
    {
      //必要に応じて逆行列を更新
      if (needUpdateInverseMatrix)
      {
        updateInverseMatrix();
        updateDirectIrradianceProperties();
        updateDiffuseIrradianceProperties();
        HasPropertyChanged = false;
      }

      if (!Pulldowned)
      {
        diffuseDiffuseTransmittance_U = 0.3;
        diffuseDiffuseTransmittance_L = 0.7;
        directDiffuseTransmittance_U = directDiffuseTransmittance_L = 0;
        directDirectTransmittanse = 1.0;
      }
      else
      {
        diffuseDiffuseTransmittance_U = difTauF_U;
        diffuseDiffuseTransmittance_L = difTauF_L;
        directDiffuseTransmittance_U = dirTauF_U;
        directDiffuseTransmittance_L = dirTauF_L;
        directDirectTransmittanse = dirTauDir;
      }
    }

    #endregion

  }
}
