﻿/* Incline.cs
 * 
 * Copyright (C) 2008 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

namespace Popolo.Weather
{
  /// <summary>傾斜面</summary>
  [Serializable]
  public class Incline : ImmutableIncline
  {

    #region 列挙型定義

    /// <summary>16方位</summary>
    public enum Orientation
    {
      /// <summary>北北東</summary>
      NNE = -7,
      /// <summary>北東</summary>
      NE = -6,
      /// <summary>東北東</summary>
      ENE = -5,
      /// <summary>東</summary>
      E = -4,
      /// <summary>東南東</summary>
      ESE = -3,
      /// <summary>南東</summary>
      SE = -2,
      /// <summary>南南東</summary>
      SSE = -1,
      /// <summary>南</summary>
      S = 0,
      /// <summary>南南西</summary>
      SSW = 1,
      /// <summary>南西</summary>
      SW = 2,
      /// <summary>西南西</summary>
      WSW = 3,
      /// <summary>西</summary>
      W = 4,
      /// <summary>西北西</summary>
      WNW = 5,
      /// <summary>北西</summary>
      NW = 6,
      /// <summary>北北西</summary>
      NNW = 7,
      /// <summary>北</summary>
      N = 8
    }

    #endregion

    #region インスタンス変数
    
    /// <summary>正弦、余弦</summary>
    private double sinBeta, cosBeta, sinAlpha, cosAlpha;

    #endregion

    #region プロパティ

    /// <summary>方位角[radian]（南を0、東を負、西を正とする）を取得する</summary>
    public double HorizontalAngle { private set; get; }

    /// <summary>傾斜角[radian]（水平面を0、垂直面を1/2πとする）を取得する</summary>
    public double VerticalAngle { private set; get; }

    /// <summary>天空への形態係数[-]を取得する</summary>
    public double ConfigurationFactorToSky { private set; get; }

    /// <summary>地表面への形態係数[-]を取得する</summary>
    public double ConfigurationFactorToGround
    {
      get { return 1 - ConfigurationFactorToSky; }
    }

    #endregion

    #region コンストラクタ

    /// <summary>コンストラクタ</summary>
    /// <param name="horizontalAngle">方位角[radian]（南：0、東：負、西：正）</param>
    /// <param name="verticalAngle">傾斜角[radian]（水平面：0、垂直面：1/2π）</param>
    public Incline(double horizontalAngle, double verticalAngle)
    {
      //方位角は-180~180
      double pi2 = Math.PI * 2;
      horizontalAngle = horizontalAngle % pi2;
      if (Math.PI < horizontalAngle) HorizontalAngle = horizontalAngle - pi2;
      else if (horizontalAngle < -Math.PI) HorizontalAngle = horizontalAngle + pi2;
      else HorizontalAngle = horizontalAngle;

      //傾斜角は0~180
      VerticalAngle = verticalAngle % Math.PI;
      if (Math.PI < VerticalAngle) VerticalAngle = Math.PI - VerticalAngle;

      //天空への形態係数を計算
      ConfigurationFactorToSky = GetConfigurationFactorToSky(verticalAngle);

      //傾斜面固有の正接・余弦を計算しておく
      sinBeta = Math.Sin(VerticalAngle);
      cosBeta = Math.Cos(VerticalAngle);
      sinAlpha = Math.Sin(HorizontalAngle);
      cosAlpha = Math.Cos(HorizontalAngle);
    }

    /// <summary>コピーコンストラクタ</summary>
    /// <param name="incline">コピーする傾斜面情報</param>
    public Incline(ImmutableIncline incline) : 
      this(incline.HorizontalAngle, incline.VerticalAngle) { }

    /// <summary>新しいインスタンスを初期化する</summary>
    /// <param name="orientation">16方位</param>
    /// <param name="verticalAngle">傾斜角[radian]（上側水平面：0、垂直面：1/2π、下側水平面：π）</param>
    public Incline(Orientation orientation, double verticalAngle)
      : this(Math.PI / 8 * (int)orientation, verticalAngle) { }

    /// <summary>逆向きの傾斜面を作成する</summary>
    /// <returns>逆向きの傾斜面</returns>
    public Incline MakeReveseIncline()
    { return new Incline(HorizontalAngle + Math.PI, VerticalAngle + Math.PI); }

    #endregion

    #region 傾斜面日射量関連の処理

    /// <summary>傾斜面の法線に対する太陽光線入射角の余弦cosθ[-]を計算する</summary>
    /// <param name="altitude">太陽高度[radian]</param>
    /// <param name="orientation">太陽方位角[radian]</param>
    /// <returns>傾斜面の法線に対する太陽光線入射角の余弦cosθ[-]</returns>
    public double GetDirectSolarRadiationRate(double altitude, double orientation)
    {
      double sh = Math.Sin(altitude);
      double ch = Math.Cos(altitude);
      double caa = Math.Cos(orientation - HorizontalAngle);
      return Math.Max(0, sh * cosBeta + ch * sinBeta * caa);
    }

    /// <summary>傾斜面の法線に対する太陽光線入射角の余弦cosθ[-]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <returns>傾斜面の法線に対する太陽光線入射角の余弦cosθ[-]</returns>
    public double GetDirectSolarRadiationRate(ImmutableSun sun)
    {
      return GetDirectSolarRadiationRate(sun.Altitude, sun.Orientation);
    }

    /// <summary>傾斜面に入射する直達日射[W/m2]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <returns>傾斜面に入射する直達日射[W/m2]</returns>
    public double GetDirectSolarIrradiance(ImmutableSun sun)
    {
      return GetDirectSolarRadiationRate(sun) * sun.DirectNormalRadiation;
    }

    /// <summary>傾斜面に入射する拡散日射[W/m2]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <param name="albedo">地表面反射率[-]</param>
    /// <returns>傾斜面に入射する拡散日射[W/m2]</returns>
    public double GetDiffuseSolarIrradiance(ImmutableSun sun, double albedo)
    {
      return ConfigurationFactorToSky * sun.DiffuseHorizontalRadiation 
        + albedo * ConfigurationFactorToGround * sun.GlobalHorizontalRadiation;
    }

    /// <summary>傾斜面に入射する日射[W/m2]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <param name="albedo">地表面反射率[-]</param>
    /// <returns>傾斜面に入射する日射[W/m2]</returns>
    public double GetSolarIrradiance(ImmutableSun sun, double albedo)
    {
      return GetDirectSolarIrradiance(sun) + GetDiffuseSolarIrradiance(sun, albedo);
    }

    #endregion

    #region 傾斜面日照関連の処理

    /// <summary>傾斜面に入射する直達照度[lx]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <returns>傾斜面に入射する直達照度[lx]</returns>
    public double GetDirectSolarIlluminance(ImmutableSun sun)
    {
      return GetDirectSolarRadiationRate(sun) * sun.DirectNormalIlluminance;
    }

    /// <summary>傾斜面に入射する拡散照度[lx]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <param name="albedo">地表面反射率[-]</param>
    /// <returns>傾斜面に入射する拡散照度[lx]</returns>
    public double GetDiffuseSolarIlluminance(ImmutableSun sun, double albedo)
    {
      return ConfigurationFactorToSky * sun.DiffuseIlluminance
        + albedo * ConfigurationFactorToGround * sun.GlobalHorizontalIlluminance;
    }

    #endregion

    #region 見かけの太陽高度関連の処理

    /// <summary>見かけの太陽高度の正接[-]を計算する</summary>
    /// <param name="altitude">太陽高度[radian]</param>
    /// <param name="orientation">太陽方位角[radian]</param>
    /// <returns>見かけの太陽高度の正接[-]</returns>
    public double GetTangentProfileAngle(double altitude, double orientation)
    {
      double cosTheta = GetDirectSolarRadiationRate(altitude, orientation);
      if (cosTheta <= 0) return -Math.PI;
      double tanGam = (Math.Sin(altitude) * sinBeta - Math.Cos(altitude) * cosBeta
        * (Math.Sin(orientation) * sinAlpha - Math.Cos(orientation) * cosAlpha))
        / cosTheta;
      return tanGam;
    }

    /// <summary>見かけの太陽高度の正接[-]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <returns>見かけの太陽高度の正接[-]</returns>
    public double GetTangentProfileAngle(ImmutableSun sun)
    {
      return GetTangentProfileAngle(sun.Altitude, sun.Orientation);
    }

    /// <summary>見かけの太陽高度[radian]を計算する</summary>
    /// <param name="altitude">太陽高度[radian]</param>
    /// <param name="orientation">太陽方位角[radian]</param>
    /// <returns>見かけの太陽高度[radian]</returns>
    public double GetProfileAngle(double altitude, double orientation)
    {
      return Math.Atan(GetTangentProfileAngle(altitude, orientation));
    }

    /// <summary>見かけの太陽高度[radian]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <returns>見かけの太陽高度[radian]</returns>
    public double GetProfileAngle(ImmutableSun sun)
    {
      return GetProfileAngle(sun.Altitude, sun.Orientation);
    }

    /// <summary>傾斜面の情報をコピーする</summary>
    /// <param name="incline">コピー対象のInclineオブジェクト</param>
    public void Copy(ImmutableIncline incline)
    {
      HorizontalAngle = incline.HorizontalAngle;
      VerticalAngle = incline.VerticalAngle;
      ConfigurationFactorToSky = incline.ConfigurationFactorToSky;
      sinBeta = Math.Sin(VerticalAngle);
      cosBeta = Math.Cos(VerticalAngle);
      sinAlpha = Math.Sin(HorizontalAngle);
      cosAlpha = Math.Cos(HorizontalAngle);
    }

    #endregion

    #region public staticメソッド

    /// <summary>天空に対する傾斜面の形態係数[-]を計算する</summary>
    /// <param name="verticalAngle">傾斜面の傾斜角[radian]（水平面を0とする）</param>
    /// <returns>天空に対する傾斜面の形態係数[-]</returns>
    public static double GetConfigurationFactorToSky(double verticalAngle)
    {
      return (1d + Math.Cos(verticalAngle)) / 2d;
    }

    /// <summary>傾斜面の法線に対する太陽光線入射角の余弦cosθ[-]を計算する</summary>
    /// <param name="horizontalAngle">傾斜面方位角[radian]（南:0、東:負、西:正）</param>
    /// <param name="verticalAngle">傾斜面傾斜角[radian]（水平面:0、垂直面:1/2π）</param>
    /// <param name="altitude">太陽高度[radian]</param>
    /// <param name="orientation">太陽方位角[radian]</param>
    /// <returns>傾斜面に対する太陽光線入射角の余弦cosθ[-]</returns>
    public static double GetDirectSolarRadiationRateToIncline
      (double horizontalAngle, double verticalAngle, double altitude, double orientation)
    {
      double sh = Math.Sin(altitude);
      double cb = Math.Cos(verticalAngle);
      double ch = Math.Cos(altitude);
      double sb = Math.Cos(verticalAngle);
      double caa = Math.Cos(orientation - horizontalAngle);
      return sh * cb + ch * sb * caa;
    }

    #endregion

  }

  #region 読み取り専用傾斜面

  /// <summary>読み取り専用傾斜面</summary>
  public interface ImmutableIncline
  {
    /// <summary>方位角[radian]（南を0、東を負、西を正とする）を取得する</summary>
    double HorizontalAngle { get; }

    /// <summary>傾斜角[radian]（水平面を0、垂直面を1/2πとする）を取得する</summary>
    double VerticalAngle { get; }

    /// <summary>天空への形態係数[-]を取得する</summary>
    double ConfigurationFactorToSky { get; }

    /// <summary>地表面への形態係数[-]を取得する</summary>
    double ConfigurationFactorToGround { get; }

    /// <summary>逆向きの傾斜面を作成する</summary>
    /// <returns>逆向きの傾斜面</returns>
    Incline MakeReveseIncline();

    /// <summary>傾斜面の法線に対する太陽光線入射角の余弦cosθ[-]を計算する</summary>
    /// <param name="altitude">太陽高度[radian]</param>
    /// <param name="orientation">太陽方位角[radian]</param>
    /// <returns>傾斜面の法線に対する太陽光線入射角の余弦cosθ[-]</returns>
    double GetDirectSolarRadiationRate(double altitude, double orientation);

    /// <summary>傾斜面の法線に対する太陽光線入射角の余弦cosθ[-]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <returns>傾斜面の法線に対する太陽光線入射角の余弦cosθ[-]</returns>
    double GetDirectSolarRadiationRate(ImmutableSun sun);

    /// <summary>傾斜面に入射する直達日射[W/m2]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <returns>傾斜面に入射する直達日射[W/m2]</returns>
    double GetDirectSolarIrradiance(ImmutableSun sun);

    /// <summary>傾斜面に入射する拡散日射[W/m2]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <param name="albedo">地表面反射率[-]</param>
    /// <returns>傾斜面に入射する拡散日射[W/m2]</returns>
    double GetDiffuseSolarIrradiance(ImmutableSun sun, double albedo);

    /// <summary>傾斜面に入射する日射[W/m2]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <param name="albedo">地表面反射率[-]</param>
    /// <returns>傾斜面に入射する日射[W/m2]</returns>
    double GetSolarIrradiance(ImmutableSun sun, double albedo);

    /// <summary>傾斜面に入射する直達照度[lx]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <returns>傾斜面に入射する直達照度[lx]</returns>
    double GetDirectSolarIlluminance(ImmutableSun sun);

    /// <summary>傾斜面に入射する拡散照度[lx]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <param name="albedo">地表面反射率[-]</param>
    /// <returns>傾斜面に入射する拡散照度[lx]</returns>
    double GetDiffuseSolarIlluminance(ImmutableSun sun, double albedo);

    /// <summary>見かけの太陽高度の正接[-]を計算する</summary>
    /// <param name="altitude">太陽高度[radian]</param>
    /// <param name="orientation">太陽方位角[radian]</param>
    /// <returns>見かけの太陽高度の正接[-]</returns>
    double GetTangentProfileAngle(double altitude, double orientation);

    /// <summary>見かけの太陽高度の正接[-]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <returns>見かけの太陽高度の正接[-]</returns>
    double GetTangentProfileAngle(ImmutableSun sun);

    /// <summary>見かけの太陽高度[radian]を計算する</summary>
    /// <param name="altitude">太陽高度[radian]</param>
    /// <param name="orientation">太陽方位角[radian]</param>
    /// <returns>見かけの太陽高度[radian]</returns>
    double GetProfileAngle(double altitude, double orientation);

    /// <summary>見かけの太陽高度[radian]を計算する</summary>
    /// <param name="sun">太陽</param>
    /// <returns>見かけの太陽高度[radian]</returns>
    double GetProfileAngle(ImmutableSun sun);
  }

  #endregion

}
