﻿/* Converter.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */
 
using System;
using System.Text;

namespace Popolo.Weather
{
  /// <summary>気象データ変換クラス</summary>
  public static class WeatherConverter
  {

    #region HAS形式データ変換

    /// <summary>HASP形式のデータをCSV形式データに変換する</summary>
    /// <param name="haspData">HASP形式データ</param>
    /// <returns>CSV形式データ</returns>
    public static string HASPtoCSV(string haspData)
    {
      StringBuilder sBuilder = new StringBuilder();
      sBuilder.AppendLine
        ("月,日,時,乾球温度[C],絶対湿度[g/kg],法線面直達日射量[MJ/(㎡h)],"
        + "水平面天空日射量[MJ/(㎡h)],夜間放射量[MJ/(㎡h)],風向,風速[m/s]");
      string[] lines = haspData.Split
        (new[] { "\r\n" }, StringSplitOptions.RemoveEmptyEntries);

      //365日の繰り返し処理
      DateTime dt = new DateTime(1999, 1, 1, 0, 0, 0);
      for (int i = 0; i < 365; i++)
      {
        string dbt = lines[i * 7];
        string hrt = lines[i * 7 + 1];
        string dnr = lines[i * 7 + 2];
        string dhr = lines[i * 7 + 3];
        string ncr = lines[i * 7 + 4];
        string wdr = lines[i * 7 + 5];
        string wsp = lines[i * 7 + 6];

        //24時間の繰り返し処理
        for (int j = 0; j < 24; j++)
        {
          //日時
          sBuilder.Append(dt.Month + "," + dt.Day + "," + dt.Hour);
          //乾球温度
          sBuilder.Append("," + ((double.Parse(dbt.Substring(j * 3, 3)) - 500) * 0.1));
          //絶対湿度
          sBuilder.Append("," + (double.Parse(hrt.Substring(j * 3, 3)) * 0.1));
          //法線面直達日射量
          sBuilder.Append("," + (double.Parse(dnr.Substring(j * 3, 3)) * 0.01));
          //水平面天空日射量
          sBuilder.Append("," + (double.Parse(dhr.Substring(j * 3, 3)) * 0.01));
          //夜間放射量
          sBuilder.Append("," + (double.Parse(ncr.Substring(j * 3, 3)) * 0.01));
          //風向
          sBuilder.Append("," + (int.Parse(wdr.Substring(j * 3, 3))));
          //風速
          sBuilder.AppendLine("," + (double.Parse(wsp.Substring(j * 3, 3)) * 0.1));
          dt = dt.AddHours(1);
        }
      }
      return sBuilder.ToString();
    }

    #endregion

    #region 拡張アメダス形式データ変換

    /// <summary>拡張アメダス形式のデータをCSV形式データに変換する</summary>
    /// <param name="exaData">拡張アメダス形式データ</param>
    /// <param name="locationNumber">地点番号</param>
    /// <param name="latitude">緯度</param>
    /// <param name="longitude">経度</param>
    /// <returns>CSV形式データ</returns>
    public static string EXAtoCSV
      (string exaData, out int locationNumber, out double latitude, out double longitude)
    {
      StringBuilder sBuilder = new StringBuilder();
      sBuilder.AppendLine
        ("月,日,時,乾球温度[C],絶対湿度[g/kg],水平面全天日射量[MJ/(㎡h)],"
        + "大気放射量[MJ/(㎡h)],風向,風速[m/s],降水量[mm],日照時間[hour]");
      string[] lines = exaData.Split
        (new[] { "\r\n" }, StringSplitOptions.RemoveEmptyEntries);

      //地点情報を読み取り
      string l0 = lines[0];
      int sL = l0.Length;
      locationNumber = int.Parse(l0.Substring(0, 3));
      latitude = double.Parse(l0.Substring(sL - 23, 2)) + double.Parse(l0.Substring(sL - 21, 3)) / 600d;
      longitude = double.Parse(l0.Substring(sL - 17, 3)) + double.Parse(l0.Substring(sL - 14, 3)) / 600d;

      //365日の繰り返し処理
      DateTime dt = new DateTime(1999, 1, 1, 0, 0, 0);
      for (int i = 0; i < 365; i++)
      {
        string dbt = lines[i * 8 + 1];
        string hrt = lines[i * 8 + 2];
        string rad = lines[i * 8 + 3];
        string ncr = lines[i * 8 + 4];
        string wdr = lines[i * 8 + 5];
        string wsp = lines[i * 8 + 6];
        string rin = lines[i * 8 + 7];
        string shr = lines[i * 8 + 8];

        //24時間の繰り返し処理
        for (int j = 0; j < 24; j++)
        {
          double buff;
          //日時
          sBuilder.Append(dt.Month + "," + dt.Day + "," + dt.Hour);
          //乾球温度
          if (double.TryParse(dbt.Substring(j * 5, 4), out buff)) sBuilder.Append("," + (buff * 0.1));
          else sBuilder.Append(",0");
          //絶対湿度
          if (double.TryParse(hrt.Substring(j * 5, 4), out buff)) sBuilder.Append("," + (buff * 0.1));
          else sBuilder.Append(",0");
          //水平面全天日射量
          if (double.TryParse(rad.Substring(j * 5, 4), out buff)) sBuilder.Append("," + (buff * 0.01));
          else sBuilder.Append(",0");
          //大気放射量
          sBuilder.Append("," + (double.Parse(ncr.Substring(j * 5, 4)) * 0.01));
          //風向
          if (double.TryParse(wdr.Substring(j * 5, 4), out buff)) sBuilder.Append("," + buff);
          else sBuilder.Append(",0");
          //風速
          if (double.TryParse(wsp.Substring(j * 5, 4), out buff)) sBuilder.Append("," + (buff * 0.1));
          else sBuilder.Append(",0");
          //降水量
          if (double.TryParse(rin.Substring(j * 5, 4), out buff)) sBuilder.Append("," + buff);
          else sBuilder.Append(",0");
          //日照時間
          if (double.TryParse(shr.Substring(j * 5, 4), out buff)) sBuilder.AppendLine("," + (buff * 0.1));
          else sBuilder.AppendLine(",0");
          dt = dt.AddHours(1);
        }
      }
      return sBuilder.ToString();
    }

    #endregion

    #region TMY1形式データ変換

    /// <summary>TMY1形式データをCSV形式データに変換する</summary>
    /// <param name="tmy1Data">TMY1形式データ</param>
    /// <returns>CSV形式データ</returns>
    public static string TMY1toCSV(string tmy1Data)
    {
      StringBuilder sBuilder = new StringBuilder();
      sBuilder.AppendLine
        ("年,月,日,時,分,直達日射[W/m2],水平面天空日射[W/m2],水平面全天日射[W/m2],"
        + "雲高さ[m],視認距離[km],気圧[kPa],外気乾球温度[C],露点温度[C],風向,風速[m/s],雲量,雲量2");
      string[] sl = tmy1Data.Split
        (new[] { "\r\n" }, StringSplitOptions.RemoveEmptyEntries);

      for (int i = 0; i < sl.Length; i++)
      {
        int year = int.Parse(sl[i].Substring(5, 2));
        if (year < 20) year += 2000;
        else year += 1900;
        sBuilder.Append(year);  //年
        sBuilder.Append("," + sl[i].Substring(7, 2)); //月
        sBuilder.Append("," + sl[i].Substring(9, 2)); //日
        sBuilder.Append("," + (int.Parse(sl[i].Substring(11, 2)) - 1)); //時
        sBuilder.Append("," + sl[i].Substring(13, 2)); //分

        if (sl[i].Substring(23, 1) == "9") sBuilder.Append(",");
        else sBuilder.Append("," + double.Parse(sl[i].Substring(24, 4)) / 3.6d); //直達日射[W/m2]
        if (sl[i].Substring(28, 1) == "9") sBuilder.Append(",");
        else sBuilder.Append("," + double.Parse(sl[i].Substring(29, 4)) / 3.6d); //水平面天空日射[W/m2]
        if (sl[i].Substring(53, 1) == "9") sBuilder.Append(",");
        sBuilder.Append("," + double.Parse(sl[i].Substring(54, 4)) / 3.6d); //水平面全天日射[W/m2]
        sBuilder.Append("," + double.Parse(sl[i].Substring(72, 4)) * 10); //雲高さ[m]
        sBuilder.Append("," + double.Parse(sl[i].Substring(81, 4)) * 10); //視認距離[km]
        sBuilder.Append("," + double.Parse(sl[i].Substring(98, 5)) / 100); //気圧[kPa]
        sBuilder.Append("," + double.Parse(sl[i].Substring(103, 4)) / 10); //外気乾球温度[C]
        sBuilder.Append("," + double.Parse(sl[i].Substring(107, 4)) / 10); //露点温度[C]
        sBuilder.Append("," + double.Parse(sl[i].Substring(111, 3))); //風向
        sBuilder.Append("," + double.Parse(sl[i].Substring(111, 4)) / 10); //風速[m/s]
        sBuilder.Append("," + double.Parse(sl[i].Substring(118, 2)) / 10); //雲量
        sBuilder.Append("," + double.Parse(sl[i].Substring(120, 2)) / 10); //雲量2
        sBuilder.AppendLine();
      }
      return sBuilder.ToString();
    }

    #endregion

  }
}
